/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2001 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    and Holger Zimmermann (http://pi3web.sourceforge.net/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Base/FSEntity.h,v $
 * $Date: 2001/04/07 13:30:17 $
 *
 Description:
\*____________________________________________________________________________*/
//$HeaderTop:$

#ifndef FSENTITY_H_
#define FSENTITY_H_

#include <assert.h>
#include <time.h>

#include "PICompat.h"
#include "FSType.h"

/*____________________________________________________________________________*\
 *
 Declarations:
\*____________________________________________________________________________*/

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class FileSystemEntity
{	
private:
	const FileSysType &tFSType;
	const char *pName;
	const char *pExtension;
	PIString sPath;
	PIString sPathRoot;

protected:
	/* --- static protected methods --- */
	void InternalPathRootFromPath( PIString &sRoot,
		const char *pPath, int iPathLen );

	inline const FileSysType &GetFSType()	{ return tFSType; };

	inline bool InternalIsDirectorySeperatorCharacter( char c ) const	
		{ return tFSType.IsDirectorySeperatorCharacter( c ); };
	inline char InternalDirectorySeperatorCharacter() const	
		{ return tFSType.DirectorySeperatorCharacter(); };
	void NewFileInSamePathRoot( const char *pFileName );
	virtual void InternalExpireFileInformation()=0;

	/* --- protected constructor, some derived instances of this class
	use this constructor to get a partially constructed file entity given
	the entity of the directory which contained it. Other methods are
	then called to complete the construction. 
	--- */
	FileSystemEntity(
		const PIString &sTheRootPath,
		const FileSysType &tTheFSType
		);

public:
	FileSystemEntity( const FileSystemEntity &tFSE );
	FileSystemEntity( const FileSysType &tTheFSType,
		const PIString &sFullPath);
	virtual ~FileSystemEntity()	{};

	/* --- modify paths --- */
	bool ConcatenateFileToPath( const PIString &sFileName );
	void SetPath( const PIString &sNewPath );
	
	/* --- directory path information --- */
	const char *GetName();
	virtual const char *GetAlternateName()	{ return GetName(); };
	const char *GetExtension();
	inline const PIString &GetPath() const 	{ return sPath; };
	inline const PIString &GetPathRoot() const	{ return sPathRoot; };

	/* --- stat information --- */
	virtual bool Exists()=0;
	virtual bool IsRegularFile()=0;
	virtual bool IsDirectory()=0;
	virtual bool IsLink()=0;
	virtual bool IsReadable()=0;
	virtual bool IsWritable()=0;
	virtual bool IsExecutable()=0;
	virtual unsigned long GetFileSize()=0;
	virtual time_t GetLastModified()=0;
	virtual FileSystemEntity *FirstFileInDirectory( int iFlags )=0;
	virtual bool NextSiblingFile()=0;

	inline void ExpireFileInformation()	{ InternalExpireFileInformation(); };
};

#endif	/* FSENTITY_H_ */

