/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2001 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    and Holger Zimmermann (http://pi3web.sourceforge.net/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/DB/ParsConf.cpp,v $
 * $Date: 2001/04/07 13:30:18 $
 *
 Description:
	Mechanisms to parse configuration files.
\*____________________________________________________________________________*/
//$SourceTop:$

#include <ctype.h>

#include "ParsConf.h"
#include "_PIDBKey.h"
#include "PISymTab.h"
#include "Utils.h"
#include "PIStrStr.h"
#include "PIConfDB.h"
#include "_PIDBUtl.h"

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#define PIKW_INCLUDE	"include"
#if 0	/* Debug? */
#	include <iostream.h>
#	define D { cerr << __FILE__ << ", " << __LINE__ << ": " << endl; }
#endif

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
ConfigDBParser::ConfigDBParser( 
		const char *pTheFileName,
		_PIDB *pTheDB )
:	ConfigurationFile( pTheFileName, os ),
	pDB( pTheDB ),
	bOK ( true )
{
	pFileName = PISymbolTable_LookupOrInsert( pTheFileName, 1 );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
bool ConfigDBParser::Parse()
{
	bOK = LexConfFile() && bOK;
	if ( !tStack.IsEmpty() )
		{
		Error() << "Some configuration blocks have not been terminated."
			<< endl;
		bOK = false;
		GetOs() << "The blocks are listed in the order in which they \
should have been terminated." << endl;
		const char *pName;
		for(int i=1; !tStack.IsEmpty(); i++)
			{
			pName = tStack.Pop();
			SetDB( (_PIDB *)GetDB()->GetParent() );
			assert( pName && GetDB() );
			return false;
			GetOs() << '\t' << i << "..." << pName << endl;
			}; 
		};

	/* ---
	Add all diagnostics into the configuration database
	--- */
	os << ends;
	if ( *( os.str() ) )
		{
		_PIDBUtil::AddLogMessage( GetDB(), (PIConfigDB *)GetDB(),
			PILOG_ERROR, os.str() );
		};

	return bOK;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void ConfigDBParser::Text()
{
	/* --- sanity check --- */
	assert( GetDB() );

	/* --- get the variable and value --- */
	const char *pszBuf=GetPtr();
	int iL=GetLen();
	PIString tTmp( pszBuf, iL );		
	const char *pTmpBuf=tTmp;
	int i=0;
	while( pTmpBuf[i] && !(isspace(pTmpBuf[i])) ) { i++; };
	PIString sVariable( pTmpBuf, i );
	while( pTmpBuf[i] && (isspace(pTmpBuf[i])) ) { i++; };
	const char *pValue = &( pTmpBuf[i] );
	for( i = strlen( pValue ); i>0 && (isspace(pValue[i-1])); i-- );
	PIString sValue( pValue, i );

	/* ---
	determine if this is an include file
	--- */
	if ( !strcmp( sVariable, PIKW_INCLUDE ) )
		{
		bOK = ((PIConfigDB *)GetDB())->Load( sValue );
		return;
		};

	/* ---
	add the variable/value pair to the configuration tree
	--- */
	GetDB()->Add(
		PIDBTYPE_USER,
		sVariable,
		PI_NEW( UserDBType(
			PIConfigDB::tConfigLineTypeFrame,
			PI_NEW( PIConfigDBString(
				sValue,
				GetX(),
				GetY(),
				GetFileName()
				) )
			) ),
		0 );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void ConfigDBParser::Markup()
{
	/* --- sanity check --- */
	assert( GetDB() );

	/* --- get the variable and value --- */
	const char *pszBuf=GetPtr();
	int iL=GetLen();
	PIString sName( pszBuf, iL );		

	/* --- 
	Check that the markup name is allowed in pi3 files
	--- */
	if (	PIUtil_stricmp( sName, "Library" ) &&
			PIUtil_stricmp( sName, "Object" ) &&
			PIUtil_stricmp( sName, "Class" ) )	
		{
		/* --- not a recognised type --- */
		Error() << "'" << sName << "' is not a recognised pi3 entity type." <<
			endl << "\tValid entity types are 'Library', 'Class' and 'Object'."
			<< endl << "\tHint: Missing quotation (\")?." << endl;
		bOK = false;
		return;
		};

	/* --- add the variable/value pair to the configuration tree --- */
	_PIDB *pSubTree = PI_NEW( PIConfigDB( GetDB(), sName ) );
	const char *pKey = PISymbolTable_Lookup( sName, 1 );
	assert( pKey );
	if ( !pKey )
		{
		assert( 0 );
		return;
		};
	tStack.Push( pKey );
	SetDB( pSubTree );
	
	/* --- add a note --- */
	PIOStrStream ostr;
	ostr << GetFileName() << " " << GetY() << ", " << GetX() << ends;
	GetDB()->Add( PIDBTYPE_STRING, PIDBKEY_NOTE, (void *)ostr.str(), 0 );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void ConfigDBParser::CommentMarkup()
{
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void ConfigDBParser::Markdown()
{
	const char *pszBuf=GetPtr();
	int iL=GetLen();
	PIString sName( pszBuf, iL );
	const char *pszName = sName;
	if ( tStack.IsEmpty() )
		{
		Error() << "Unexpected directive: '" << sName << "'" << endl;
		bOK = false;
		return;
		};
	const char *pLastMarkup = tStack.PeekAtHead();
	_PIDB *pParent = (_PIDB *)GetDB()->GetParent();
	assert( pParent && pLastMarkup );
	if ( !pParent || !pLastMarkup )
		{
		return;
		};
	if ( pi_stricmp( &(pszName[1]), pLastMarkup ) )
		{
		Error() << "Expecting </" << pLastMarkup << ">" << endl;
		bOK = false;
		return;
		};
	tStack.Pop();
	SetDB( pParent );
}

