/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2001 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    and Holger Zimmermann (http://pi3web.sourceforge.net/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/DB/_PIDB.h,v $
 * $Date: 2001/04/07 13:30:18 $
 *
 Description:
	Base class and definitions for any database tree.
\*____________________________________________________________________________*/
/* $HeaderTop:$ */

#ifndef _PIDB_H_
#define _PIDB_H_

#include "PICompat.h"		/* for bool */
#include "Pi2API.h"
#include "ProgMain.h"
#include "PISymTab.h"

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
struct _TypeFrame;

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class _PIDB
{
private:
	_PIDB *pParent;

public:
	_PIDB( _PIDB *pTheParent );
	virtual ~_PIDB()	{};
	
	inline _PIDB *GetParent()		{ return pParent; };
	inline void SetParent( _PIDB *pNewParent )
									{ pParent = pNewParent; };
	virtual int Add( int iType, const char *pKey, void *pValue,
		int iFlags ) = 0;
	virtual int Replace( int iType, const char *pKey, void *pValue,
		int iFlags ) = 0;
	virtual _PIDBIterator *GetIterator( int iType, const char *pKey,
		 int iFlags ) = 0;
	virtual void *Lookup( int iType, const char *pKey, int iFlags ) = 0;
	virtual void Render( ostream &os ) const = 0;

	/* --- implementation of non-trivial API functions --- */
	int Delete();
	_PIDB *GetRoot();
	bool Detach();
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
	This structure contains various functions used to manipulate
	_PIDB types.
\*____________________________________________________________________________*/
struct _TypeFrame
{
	const char *pName;						/* Name of the type */
	const char *pDescription;				/* Description of the type */
	void *(* fnAdd)( void * );				/* Add a new object to a _PIDB */
	void (* fnRender)(ostream &os, const void *pV, int iLevel );
											/* Display the value of the type */
	void (* fnDelete)( void * );			/* Delete the object at Value */
};
typedef struct _TypeFrame TypeFrame;

/*____________________________________________________________________________*\
 *
 Class:
 Description:
	This abstract base class serves only to force a consistent interface
	to iterators
\*____________________________________________________________________________*/
class _PIDBIterator
{
public:
	virtual ~_PIDBIterator()		{};
	virtual bool AtValidElement()=0;
	virtual void *Current( const char **ppKey )=0;
	virtual void *DetachCurrent()=0;
	virtual void ReplaceCurrent( void *pNew )=0;
	virtual void RemoveCurrent()=0;
	virtual void Next()=0;
};

#endif /* _PIDB_H_ */

