/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2001 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    and Holger Zimmermann (http://pi3web.sourceforge.net/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Fcgi/FastCGI.cpp,v $
 * $Date: 2002/01/26 20:11:56 $
 *
 Description:
	Simple Fast CGI as per Open Market specifications.

\*____________________________________________________________________________*/
//$SourceTop:$

#include <iostream.h>
#include <stdio.h>
#include <ctype.h>
#include <fstream.h>

#include "DblList.h"
#include "HandBase.h"
#include "HTTPCore.h"
#include "HTTPUtil.h"
#include "PIStrStr.h"

#define D { cerr << __FILE__ << ": " << __LINE__ << endl; };
// #define D

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#if 0
	/*
	** HTML documentation for this handler
	*/
/*___+++HTMLDOC_BEGIN+++___*/
Name:
	FastCGI

Description:
	Perform simple fastcgi request forwarding.

Options:
<H5>Overview</H5>
<TABLE BORDER=1>
<TH>Option
<TH>Default
<TH>Values
<TH>Short Description
<TH>Example(s)

<TR>
<TD>IOObject
<TD>+
<TD>&lt;A Pi3 Object&gt;
<TD>Prototype IO object for fast cgi application server
<TD>IOObject "FastCGIIO"

<TR>
<TD>File
<TD>-
<TD>&lt;a filename&gt;
<TD>Filename to match
<TD>File "App1.fcgi"

<TR>
<TD>FCGIHost
<TD>localhost
<TD>&lt;A hostname or IP Address&gt;
<TD>Host to connect to application server
<TD>FCGIHost "localhost"

<TR>
<TD>FCGIPort
<TD>3333
<TD>&lt;A TCP port&gt;
<TD>Port to connect to application server
<TD>FCGIPort 2323

<TR>
<TD>Variable
<TD>-
<TD>&lt;Pi3Expression&gt;
<TD>A variable,value parameter definition
<TD>Variable "REMOTE_ADDR=$A"

</TABLE>
<STRONG>-</STRONG> in the <IT>default</IT> indicates no default<BR>
<STRONG>+</STRONG> in the <IT>default</IT> indicates the field is mandatory<BR>

<H4>Description of Options</H4>
<H5>IOObject</H5>
Sepcifies an IOObject which will be the prototype of IO objects used to
communicate with the fastcgi application server process.
This IOObject may directly implement an IPC mechanism (shared memory,
TCP/IP or named pipes) or its may be a filter (SSL, monitor) applied
to one of the IPC mechanisms.

<H5>File</H5>
Specifies a filename to match against to activate this handler with
the specified host and port. If this field is not present the will
attempt to get the host and port pair from the resource associated
with the request (the file should contain the hostname followed by
whitespace followed by the port number).

<H5>FCGIHost</H5>
Specifies the hostname that will be passed as argument 1 to the prototype
IO object when creating the new IO object using PIIObject_copy(), for 
TCPIP IO objects this will be used as the hostname on which to connect 
to the fast cgi application server.

<H5>FCGIPort</H5>
Specifies the port that will be passed as argument 2 to the prototype
IO object when creating the new IO object using PIIObject_copy(), for 
TCPIP IO objects this will be used as the port on which to connect 
to the fast cgi application server.

<H5>Variable</H5>
Specifies a parameter expression that will be evaluated and sent to the
fast cgi application server as part of the parameter definition. This
is equivalent to specifying a CGI environment variable.

Phase:
	HANDLE

Returns:
	PIAPI_COMPLETED on success.
	PIAPI_CONTINUE is this handler ignored the request.
	If this handler was invoked for a phase other than HANDLE then PIAPI_ERROR
	is returned.

Note:

Example:
	<PRE>
	&lt;Object&gt;
		Name FastCGI
		Class FastCGIClass
	&lt;/Object&gt;

	&lt;Object&gt;
		...
		Handle Condition="&cmp($c,internal/x-fastcgi)" FastCGI \
			File="App1.fcgi" IOObject="FcgiIO" FCGIHost="localhost" \
			FCGIPort=3333
		...
	&lt;/Object&gt;

	</PRE>
/*___+++HTMLDOC_END+++___*/
#endif

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#define KEY_CONF_IOOBJECT						"IOObject"
#define KEY_CONF_FILE							"File"
#define KEY_CONF_FCGIHOST						"FCGIHost"
#define KEY_CONF_FCGIPORT						"FCGIPort"
#define KEY_CONF_VARIABLE						"Variable"

#if 1
/*
 * Listening socket file number
 */
#define FCGI_LISTENSOCK_FILENO 0

typedef struct {
    unsigned char version;
    unsigned char type;
    unsigned char requestIdB1;
    unsigned char requestIdB0;
    unsigned char contentLengthB1;
    unsigned char contentLengthB0;
    unsigned char paddingLength;
    unsigned char reserved;
} FCGI_Header;

/*
 * Number of bytes in a FCGI_Header.  Future versions of the protocol
 * will not reduce this number.
 */
#define FCGI_HEADER_LEN  8

/*
 * Value for version component of FCGI_Header
 */
#define FCGI_VERSION_1           1

/*
 * Values for type component of FCGI_Header
 */
#define FCGI_BEGIN_REQUEST       1
#define FCGI_ABORT_REQUEST       2
#define FCGI_END_REQUEST         3
#define FCGI_PARAMS              4
#define FCGI_STDIN               5
#define FCGI_STDOUT              6
#define FCGI_STDERR              7
#define FCGI_DATA                8
#define FCGI_GET_VALUES          9
#define FCGI_GET_VALUES_RESULT  10
#define FCGI_UNKNOWN_TYPE       11
#define FCGI_MAXTYPE (FCGI_UNKNOWN_TYPE)

/*
 * Value for requestId component of FCGI_Header
 */
#define FCGI_NULL_REQUEST_ID     0

typedef struct {
    unsigned char roleB1;
    unsigned char roleB0;
    unsigned char flags;
    unsigned char reserved[5];
} FCGI_BeginRequestBody;

typedef struct {
    FCGI_Header header;
    FCGI_BeginRequestBody body;
} FCGI_BeginRequestRecord;

/*
 * Mask for flags component of FCGI_BeginRequestBody
 */
#define FCGI_KEEP_CONN  1

/*
 * Values for role component of FCGI_BeginRequestBody
 */
#define FCGI_RESPONDER  1
#define FCGI_AUTHORIZER 2
#define FCGI_FILTER     3

typedef struct {
    unsigned char appStatusB3;
    unsigned char appStatusB2;
    unsigned char appStatusB1;
    unsigned char appStatusB0;
    unsigned char protocolStatus;
    unsigned char reserved[3];
} FCGI_EndRequestBody;

typedef struct {
    FCGI_Header header;
    FCGI_EndRequestBody body;
} FCGI_EndRequestRecord;

/*
 * Values for protocolStatus component of FCGI_EndRequestBody
 */
#define FCGI_REQUEST_COMPLETE 0
#define FCGI_CANT_MPX_CONN    1
#define FCGI_OVERLOADED       2
#define FCGI_UNKNOWN_ROLE     3

/*
 * Variable names for FCGI_GET_VALUES / FCGI_GET_VALUES_RESULT records
 */
#define FCGI_MAX_CONNS  "FCGI_MAX_CONNS"
#define FCGI_MAX_REQS   "FCGI_MAX_REQS"
#define FCGI_MPXS_CONNS "FCGI_MPXS_CONNS"

typedef struct {
    unsigned char type;    
    unsigned char reserved[7];
} FCGI_UnknownTypeBody;

typedef struct {
    FCGI_Header header;
    FCGI_UnknownTypeBody body;
} FCGI_UnknownTypeRecord;

#endif

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class FastCGI : public HandlerBaseFcgi
{
private:
	FastCGI( const FastCGI &t )
	:	HandlerBaseFcgi( t )
		{ assert( 0 ); };

	/* ---
	Configuration data
	--- */
	PIObject *pIO;						/* prototype IO object */
	PIString sFile;						/* file to match */
	PIString sHost;						/* host of fcgi application server */
	int iPort;							/* port of fcgi application server */
	enum { PARAM_BLOCK_SIZE=4096 };		/* buffer size for parameters */
	DblList lVariables;					/* expression for parameters */

protected:
	int Parameter( const char *pVariable, const char *pValue,
		const char *pWhere )
		{
		assert( pVariable && pValue );
		PIOStrStream os;
		os << pWhere << "FastCGI: ";
		if ( !PIUtil_stricmp( KEY_CONF_FILE, pVariable ) )
			{
			sFile = pValue;
			}
		else if ( !PIUtil_stricmp( KEY_CONF_FCGIHOST, pVariable ) )
			{
			sHost = pValue;
			}
		else if ( !PIUtil_stricmp( KEY_CONF_FCGIPORT, pVariable ) )
			{
			iPort = pValue ? atoi( pValue ) : 0;
			}
		else if ( !PIUtil_stricmp( KEY_CONF_IOOBJECT, pVariable ) )
			{
			if ( pIO )
				{
				os << "'IOObject' may only be specified once." << ends;
				CONFIG_ERR( Object(), os.str() );
				return 0;
				};

			PIObject *pTmpIO = PIObject_loadFromLine( 
				PIObject_getDB( Object() ),
				PIObject_getConfigurationDB( Object() ),
				pValue );

			if ( !pTmpIO )
				{ return 0; };

			pIO = pTmpIO;
			}
		else if ( !PIUtil_stricmp( KEY_CONF_VARIABLE, pVariable ) )
			{
			Pi3String *pError = Pi3String_new( 0 );
			Pi3Expression *pExpr = Pi3Expression_new( pValue, 0,
				pError );
			if ( !pExpr )
				{
				os << Pi3String_getPtr( pError ) << ends;
				Pi3String_delete( pError );
				CONFIG_ERR( Object(), os.str() ); 
				return 0;
				};
			Pi3String_delete( pError );
			lVariables.Append( (DblList::type)pExpr );
			}
		else
			{
			os << "Unknown directive '" << pVariable <<
				"'" << ends;
			CONFIG_ERR( Object(), os.str() );
			return 0;
			};

		return 1;
		};

public:
	FastCGI( PIObject *pObject, int iArgc, const char *ppArgv[] )
	:	HandlerBaseFcgi( pObject ),
		pIO( 0 ),
		sHost( "localhost" ),
		iPort( 3333 )
		{
		ReadParameters( iArgc, ppArgv );
		if ( !IsOK() )
			{ return; };

		if ( !pIO )
			{
			CONFIG_ERR( Object(), "FastCGI: Missing definition for \
'IOObject'." );
			SetOK( 0 );
			return;
			};

#if 0
		/* ---
		Make an identification string for error messages
		--- */
		PIOStrStream os;
		os << "file<" << sFile << ">_host<" << sHost << ">_port<" 
			<< iPort << ">" << ends;
		strncpy( szIdent, os.str(), IDENT_LEN );
		szIdent[IDENT_LEN]='\0';
#endif
		
		assert( pIO );
		};

	~FastCGI()
		{
		for( DblListIterator i( lVariables ); !i.BadIndex(); i++ )
			{ Pi3Expression_delete( (Pi3Expression *)i.Current() ); };
		PIObject_delete( pIO, 0, 0 );
		};

	/* -------------- +++++++++++++++++++++++++++++++ ------------------ *\

							Utilities

	\* -------------- +++++++++++++++++++++++++++++++ ------------------ */
	/* ---
	Write environment variables (params) into the supplied block.

	Returns 1 on success, 0 on failure.
	--- */
	int AddEnvironmentVariables(
			PIHTTP &tPIHTTP,
			char *pBlock,
			int &iBlockSize,
			int &iNumberOfVariables )
		{
		for( DblListIterator i( lVariables ); !i.BadIndex(); i++ )
			{
			Pi3Expression *pExpr = (Pi3Expression *)i.Current();
			assert( pExpr );

			int iWritten = Pi3Expression_write( pExpr, &tPIHTTP, this, pBlock,
				iBlockSize );
			if ( iWritten==-1 )
				{
				return 0;
				}
			/*
			** Only add non empty variable patterns
			*/
			else if ( iWritten > 0 )
				{
				/* ---
				Check first, before looking at contents.
				--- */
				if ( iWritten>=iBlockSize )
					{
					*pBlock = '\0';
					return 0;
					};

				/* --- buffer for lengths --- */
				unsigned char szLengths[8];
				int iLengthPos=0;

				/* --- get index of '=' --- */
				int i=0;
				for( ; i<iWritten && pBlock[i]!='='; i++ );
				int iNameLen = i;
				if ( i<iWritten ) { i++; };
				int iValueLen = iWritten - i;

				/* --- i is index of first character of value in pBlock --- */

				/* --- lengths --- */
				if ( iNameLen < 0x80 )
					{ szLengths[iLengthPos++] = iNameLen; }
				else
					{
					iNameLen &= 0xFFFFFFFF;
					szLengths[iLengthPos++] = ( iNameLen >> 24 ) | 0x80;
					szLengths[iLengthPos++] = ( iNameLen >> 16 );
					szLengths[iLengthPos++] = ( iNameLen >> 8 );
					};
				if ( iValueLen < 0x80 )
					{ szLengths[iLengthPos++] = iValueLen; }
				else
					{
					iValueLen &= 0xFFFFFFFF;
					szLengths[iLengthPos++] = ( iValueLen >> 24 ) | 0x80;
					szLengths[iLengthPos++] = ( iValueLen >> 16 );
					szLengths[iLengthPos++] = ( iValueLen >> 8 );
					};

				/* ---
				Check again because we've expanded the size
				--- */
				if ( iWritten>=iBlockSize )
					{
					*pBlock = '\0';
					return 0;
					};
			
				iWritten = iLengthPos + iNameLen + iValueLen;

				/* --- sanity --- */
				assert( ((iWritten==(iNameLen+iValueLen))==i) ||
						((iWritten>(iNameLen+iValueLen))&&(i<iWritten)) );

				memmove(
					&(pBlock[iWritten-iValueLen]),
					&(pBlock[i]),
					iValueLen );
				memmove(
					&(pBlock[(iWritten-iValueLen)-iNameLen]),
					pBlock,
					iNameLen );
				memcpy( pBlock, szLengths, iLengthPos );

				pBlock = &( pBlock[iWritten] );
				iBlockSize -= iWritten;

				iNumberOfVariables++;
				};
			};

		return 1;
		};

	/* ---
	Write output from one buffer into another.
	iReadLen gives number of bytes to transfer, must be > 0

	Returns 0 on success.
	--- */
	int WriteBufferToBuffer( PIIOBuffer &tInBuffer, PIIOBuffer &tOutBuffer,
		int iReadLen )
		{
		assert( iReadLen>0 );
		if ( iReadLen<=0 )
			{ return 0; };
		enum { BUF_SIZE=4096 };
		char szBuffer[BUF_SIZE];
		int iAttemptToRead = (iReadLen >= BUF_SIZE) ? BUF_SIZE : iReadLen;

		int iLen;
		iLen = PIIOBuffer_readToBuffer( &tInBuffer, szBuffer, iAttemptToRead );
		while( iLen>0 )
			{
			if ( PIIOBuffer_write( &tOutBuffer, szBuffer, iLen,
				PIIOBUF_NONE )<=0 )
				{ return -1; };

			iAttemptToRead -= iLen;
			if ( !iAttemptToRead )
				{ return 0; };

			iLen = PIIOBuffer_readToBuffer( &tInBuffer, szBuffer,
				iAttemptToRead );
			};

		/* ---
		Done
		--- */
		return iLen;
		};

	/* ---
	Initialize a fast CGI header record
	--- */
	void GenerateFcgiHeader( FCGI_Header &tHeader, int iType, 
		int iRequestId, int iContentLength )
		{
		tHeader.version = FCGI_VERSION_1;
		tHeader.type = iType;
		tHeader.requestIdB1 = (iRequestId >> 8 ) & 0xff;
		tHeader.requestIdB0 = (iRequestId ) & 0xff;
		tHeader.contentLengthB1 = (iContentLength >> 8 ) & 0xff;
		tHeader.contentLengthB0 = (iContentLength ) & 0xff;
		tHeader.paddingLength = 0;
		};

	/* ---
	Wrap a Fcgi body in a record and send it into the specified
	IO buffer object.

	Returns 0 on success, -1 on failure.
	--- */
	int SendFcgiBody( PIIOBuffer *pIOBuffer, void *pBody, int iBodyLength,
		int iType, int iRequestId )
		{
		assert( pIOBuffer );

		/* ---
		Generate header record
		--- */
		FCGI_Header tHeader;
		GenerateFcgiHeader( tHeader, iType, iRequestId, iBodyLength );

		/* ---
		Send header
		--- */
		if ( PIIOBuffer_write( pIOBuffer, (const char *)&tHeader,
			sizeof( tHeader ), PIIOBUF_NONE )!=sizeof( tHeader ) )
			{
			return -1;
			};

		/* ---
		Send body and flush
		--- */
		if ( iBodyLength>0 && PIIOBuffer_write( pIOBuffer, (const char *)pBody,
			iBodyLength, PIIOBUF_NONE )!=iBodyLength )
			{
			return -1;
			};

		return 0;	/* --- OK --- */
		};

	/* ---
	Correct response variables just before they're sent
	--- */
	void CorrectResponseVariables( PIHTTP &tPIHTTP )
		{
		/* --- disable keep alive --- */
		PIDB_replace( tPIHTTP.pConnectionDB, PIDBTYPE_OPAQUE, 
			KEY_INT_KEEPOPEN, (void *)0, 0 );

		/* --- set response status code --- */
		tPIHTTP.iStatus = ST_OK;
		};

	/* ---
	Handle stdout from an FCGI application server

	Return 0 on success, -1 on error.
	--- */
	int HandleFcgiStdout(
			PIHTTP &tPIHTTP,
			char *pData,
			int iDataLength,
			int &iVariablePos,			/* --- note pass by reference --- */
			int &iReadingVariables, 	/* --- note pass by reference --- */
			const char *szIdent )
	{
	assert( pData );

	if ( iReadingVariables )
		{
		/* ---
		reading CGI response headers
		--- */
		assert( iVariablePos<iDataLength );

		/* --- get index of EOL (\r\n) --- */
		int iEOL = 0;
		for( ;
			(iEOL<iDataLength) &&
			(pData[iEOL]!='\r') &&
			(pData[iEOL+1]!='\n');
			iEOL++ );

		if ( iEOL==iDataLength )
			{
			/* --- end of variables --- */
			iReadingVariables = 0;
			}
		else
			{

			/* ---
			EOL reached read in the variable.
			--- */
			assert( iEOL<=iDataLength );
			if ( iEOL )
				{
				/* ---
				Seperate variable from value
				--- */
				int i=0;
				for( ; i<iEOL && pData[i]!=':'; i++);	/* scan to ':' */
				/* ---
				real rfc822 variable, value pair
				--- */
				if ( i < iEOL - 3 /* : x\0 */)
					{
					pData[iEOL] = '\0';
					pData[i]='\0';	
					const char *pVariable = pData;
					if ( i<iEOL ) { i++; };
					for( ; i<iEOL && isspace(pData[i]); i++);/* scan to !ws */
					const char *pValue = &(pData[i]);
		
					/* ---
					Place variable value pair in response DB
					--- */
					assert( pVariable );
					assert( *pVariable );
					assert( pValue );
					PIDB_replace( tPIHTTP.pResponseDB, PIDBTYPE_RFC822, pVariable,
						(void *)pValue, 0 );

					/* --- reset variable buffer --- */
					iEOL+=2;		/* --- increment to next character ---  */
					assert( iEOL<=iDataLength );
					iVariablePos += iEOL;
					iReadingVariables = iVariablePos<iDataLength;
					/* ---
					Recurse to read the rest of the buffer if there's more
					--- */
					iDataLength -= iEOL;
					pData = &(pData[iEOL]);
					}
				else
					{
					iReadingVariables = 0;
					};
				}
			else
				{
				/* --- end of variables --- */
				if ( iVariablePos )
					{
					iDataLength -= 2;
					pData = &(pData[2]);
					};
				iReadingVariables = 0;
				};
			};

		/* ---
		Recurse to read more headers
		--- */
		if ( iDataLength && iReadingVariables )
			{
			return HandleFcgiStdout( tPIHTTP, pData, iDataLength,
					iVariablePos, iReadingVariables, szIdent );
			};
		};

	if (!iReadingVariables )
		{
		/* ---
		End of the stdout variables 
		start the http response
		--- */
		PIDB *pR = tPIHTTP.pResponseDB;
		CorrectResponseVariables( tPIHTTP );
		if (	HTTPCore_sendGeneralHeaders( &tPIHTTP ) ||
				HTTPCore_sendEntityHeaders( &tPIHTTP, pR ) )
			{
			HTTPCore_logError( &tPIHTTP, "FastCGI: Failed \
to start HTTP response from fast cgi application '%s'.", szIdent );
			return -1;
			};

		/* ---
		reading CGI response data
		Just ship it straight back to the client browser.
		--- */
		int iRet = 1;
		if ( iDataLength )
			{
			iRet = PIIOBuffer_write( tPIHTTP.pBuffer, pData, iDataLength,
				PIIOBUF_NONE );
			};
			if ( iRet<=0 )
			{
			HTTPCore_logError( &tPIHTTP, "FastCGI: Error sending \
response to browser from fast cgi application '%s'.", szIdent );
			return -1;
			};
		};

	return 0;		/* --- OK --- */
	};

	/* ---
	Read a Fcgi response.

	returns 0 on success, -1 on failure.
	--- */
	int ReadFcgiResponse( PIHTTP &tPIHTTP, PIIOBuffer *pIOBuffer,
		const char *szIdent )
		{
		PIString sStdErr;		/* for stderr responses */
		const char *pData;		/* pointer to current data block */
		FCGI_Header tHeader;	/* for reading FCGI headers */
		int iReadSoFar = 0;		/* offset into record or data */
		int iRecordLength;		/* length of current fcgi record */
		char *pBuffer = 0;		/* buffer for record data */
		int iExit = 1;			/* exit code or 1 for no exit */
		int iVariablePos=0;	/* current pos in HTTP vars */
		int iReadingVariables = 1;
								/* initially reading variables */
		enum { HEADER=1, RECORD } eRecordState = HEADER;
		int iLen;				/* length of data read from fcgi */
		for( ; (pData = PIIOBuffer_read( pIOBuffer, &iLen ))!=0 &&
			iExit == 1; )
			{
			HTTPCore_logDebug( DBG_LOW, "FastCGI: %s, Read %d bytes from fcgi.",
				szIdent, iLen );
	
			if ( eRecordState == HEADER )
				{
				/* ---
				Read fast cgi header
				--- */
				assert( iReadSoFar<(int)sizeof( FCGI_Header ) );
				int iToRead = sizeof( FCGI_Header ) - iReadSoFar;
				if ( iToRead>iLen )
					{ iToRead = iLen; };
				memcpy( &(((char *)&tHeader)[iReadSoFar]), pData, iToRead );
				iReadSoFar += iToRead;
				if ( iReadSoFar<(int)sizeof( FCGI_Header ) )
					{ /* --- read more into buffer --- */; continue; };

				/* ---
				Reset buffer pointers 
				--- */
				pData = &(pData[iToRead]);
				iLen -= (int)sizeof( FCGI_Header );
				iReadSoFar = 0;
				eRecordState = RECORD;

				/* ---
				prepare for record data
				--- */
				iRecordLength = ( tHeader.contentLengthB1 << 8 ) +
						tHeader.contentLengthB0;
				pBuffer = PI_NEW( char[iRecordLength] );
				};

			if ( eRecordState == RECORD )
				{
				/* ---
				Read fast cgi record
				--- */
				assert( iReadSoFar<iRecordLength );
				int iToRead = iRecordLength - iReadSoFar;
				// Can't read more than in the buffer at once
				if (iToRead > iLen - tHeader.paddingLength) {
					// Cut padding bytes
					iToRead = iLen - tHeader.paddingLength;
				};
				memcpy( &(pBuffer[iReadSoFar]), pData, iToRead );
				iReadSoFar += iToRead;
				if ( iReadSoFar<iRecordLength )
					{ /* --- read more into buffer --- */; continue; };

				/* ---
				Record fully read
				--- */
				};

			HTTPCore_logDebug( DBG_MED, "FastCGI: %s, Read header of type %d.",
				szIdent, (int)tHeader.type );
	
			/* ---
			Process this fcgi record
			--- */
			switch( tHeader.type )
				{
				case FCGI_STDERR:
					{
					/* ---
					Append to the stderr message
					--- */
					PIString sTmp( pBuffer, iRecordLength );
					sStdErr.Concatenate( sTmp );
					};
					break;

				case FCGI_STDOUT:
					if ( HandleFcgiStdout( tPIHTTP, pBuffer, iRecordLength,
						iVariablePos, iReadingVariables, szIdent )==-1)				
						{
						iExit = -1;
						};
					break;

				case FCGI_END_REQUEST:
					iExit = 0;
					break;

				case FCGI_GET_VALUES_RESULT:
					/* not  implemented */
				case FCGI_UNKNOWN_TYPE:
					/* not  implemented */
				default:
					iExit = -1;
				};

			/* ---
			Done with this record
			--- */
			PI_DELETE( [] pBuffer );

			/* ---
			Reset state to read next fcgi header
			--- */
			iReadSoFar = 0;
			eRecordState = HEADER;
			};

		HTTPCore_logDebug( DBG_LOW, "FastCGI: %s, done reading response.",
				szIdent );

		/* ---
		At this point the protocol headers should have been sent to the
		client and we should have been reading the data to send to the
		client
		--- */
		if ( iReadingVariables )
			{
			/* --- but I guess it didn't work out that way... --- */
			HTTPCore_logError( &tPIHTTP, "FastCGI: Failure to read \
proper fcgi STDOUT response from fcgi application '%s'.", szIdent );
			iExit = -1;
			};

		/* ---
		If the CGI application wrote stderr log it 
		Log stderr message
		--- */
		if ( sStdErr.Len()>0 )
			{
			HTTPCore_logError( &tPIHTTP, "FastCGI: Fast cgi \
application '%s' wrote STDERR output. Output follows:\n\
______________________________________________________________________________\
__\n%s\n\
______________________________________________________________________________\
__", szIdent, (const char *)sStdErr );
			iExit = -1;
			};

		return iExit;
		};

	/* -------------- +++++++++++++++++++++++++++++++ ------------------ *\

								Guts

	\* -------------- +++++++++++++++++++++++++++++++ ------------------ */
	/* ---
	GetFcgiConnection(), get a connection to the Fcgi application. This
	function can copy a new connection or reuse an existing one.

	For simplicity this function creates a new connection right now.
	--- */
	PIObject *GetFcgiConnection( PIHTTP &tPIHTTP, const PIString &sFcgiHost,
		int iFcgiPort )
		{
		(void)tPIHTTP;

		/* --- establish a connection to the remote Fcgi app. server --- */
		const char *ppConnectParameters[2];
		ppConnectParameters[0] = (const char *)sFcgiHost;
		ppConnectParameters[1] = (const char *)iFcgiPort;
		PIObject *pConn = PIObject_copy( pIO, 2, ppConnectParameters );

		return pConn;
		};
	
	/* ---
	ReleaseFcgiConnection(), release a connection to the fast cgi application,
	in the future this could be used to allow connection reuse.
	--- */
	void ReleaseFcgiConnection( PIObject *pConnection )
		{
		/* --- always destroy the connection --- */
		if ( pConnection )
			{ PIObject_delete( pConnection, 0, 0 ); };
		};

	/* ---
	Top level proxy server stuff
	--- */
	int DoFastCGI( PIHTTP &tPIHTTP, PIIOBuffer *pIOBuffer,
		const char *szIdent )
		{
		assert( pIO );

/*		PIDB *pR = tPIHTTP.pResponseDB; */
		PIDB *pQ = tPIHTTP.pRequestDB;

		/* ---
		Send Fcgi begin message
		--- */
		FCGI_BeginRequestBody tBody;
		tBody.roleB1 = ( FCGI_RESPONDER >> 8 ) & 0xff;
		tBody.roleB0 = ( FCGI_RESPONDER ) & 0xff;
		tBody.flags = 0;
		memset( tBody.reserved, 0, sizeof( tBody.reserved ) );
		if ( SendFcgiBody( pIOBuffer, &tBody, sizeof( tBody ),
			FCGI_BEGIN_REQUEST, 1 )==-1 )
			{
			/* --- error sending begin request --- */
			HTTPCore_logError( &tPIHTTP, "FastCGI: Error sending \
FCGI_BEGIN_REQUEST \
to fast cgi application '%s'.", szIdent );
			return HTTPUtil_doHTTPError( &tPIHTTP, ST_INTERNALERROR );
			};

		/* ---
		Send Fcgi parameters,
		--- */
		char *pParamsBlock = PI_NEW( char[PARAM_BLOCK_SIZE] );
		int iParamsBlockSize = PARAM_BLOCK_SIZE;
		int iNumVariables = 0;
		int iOK = AddEnvironmentVariables( tPIHTTP, pParamsBlock,
			iParamsBlockSize, iNumVariables );
		int iToSend = PARAM_BLOCK_SIZE - iParamsBlockSize;
		assert( iToSend>0 );
		if ( iOK && iToSend )
			{
			iOK = ( SendFcgiBody( pIOBuffer, pParamsBlock,
				iToSend, FCGI_PARAMS, 1 )!=-1 );
			};
		PI_DELETE( [] pParamsBlock );
		if ( !iOK )
			{
			HTTPCore_logError( &tPIHTTP, "FastCGI: Error \
generating or sending \
FCGI_PARAMS to fast cgi application '%s'.", szIdent );
			return HTTPUtil_doHTTPError( &tPIHTTP, ST_INTERNALERROR );
			};

		/* ---
		Determine whether or not there is any input to be sent to the
		fcgi application server.
		--- */
		const char *pContentLength = (const char *)PIDB_lookup( pQ,
			PIDBTYPE_RFC822, KEY_HTTP_CONTENTLENGTH, 0 );
		int iStdinLen = pContentLength ? atoi( pContentLength ) : 0;
		if ( iStdinLen>0 )
			{
			/* ---
			Generate header record
			--- */
			FCGI_Header tHeader;
			GenerateFcgiHeader( tHeader, FCGI_STDIN, 1, iStdinLen );

			/* ---
			Send header
			--- */
			if ( PIIOBuffer_write( pIOBuffer, (const char *)&tHeader,
				sizeof( tHeader ), PIIOBUF_NONE )!=sizeof( tHeader ) )
				{
				HTTPCore_logError( &tPIHTTP, "FastCGI: Error \
sending FCGI_STDIN header to fast cgi application '%s'.", szIdent );
				return HTTPUtil_doHTTPError( &tPIHTTP, ST_INTERNALERROR );
				};

			/* ---
			Write iStdinLen bytes from browser buffer to the fcgi buffer.
			--- */
			if ( WriteBufferToBuffer( *pIOBuffer, *(tPIHTTP.pBuffer),
				iStdinLen ) )
				{
				/* --- error --- */
				HTTPCore_logError( &tPIHTTP, "FastCGI: Error \
sending %d bytes of STDIN from browser to fast \
cgi application '%s'.", iStdinLen, szIdent );
				return HTTPUtil_doHTTPError( &tPIHTTP, ST_INTERNALERROR );
				};
			}
		else
			{
			/* ---
			Send an empty fcgi record
			--- */
			if ( SendFcgiBody( pIOBuffer, 0, 0, FCGI_STDIN, 1 )==-1 )
				{
				HTTPCore_logError( &tPIHTTP, "FastCGI: Error sending \
empty FCGI_STDIN \
to fast cgi application '%s'.", szIdent );
				return HTTPUtil_doHTTPError( &tPIHTTP, ST_INTERNALERROR );
				};
			};

		/* ---
		Flush output to fcgi application
		--- */
		PIIOBuffer_flush( pIOBuffer );

		/* ---
		Read Fcgi stdout and stderr or Fcgi end request
		--- */
		if ( ReadFcgiResponse( tPIHTTP, pIOBuffer, szIdent )==-1 )
			{
			/* ---
			Error message already given 
			--- */
			return HTTPUtil_doHTTPError( &tPIHTTP, ST_INTERNALERROR );
			};

		/* --- done --- */
		return PIAPI_COMPLETED;
		};

	/* -------------- +++++++++++++++++++++++++++++++ ------------------ *\

						Standard Handle Function

	\* -------------- +++++++++++++++++++++++++++++++ ------------------ */
	int Handle( int iPhase, PIHTTP &tPIHTTP, PIIOBuffer &/* tBuffer */ )
		{
		if ( iPhase!=PH_HANDLE )
			{ return PIAPI_ERROR; };

		assert( pIO );

		const char *pPath = (const char *)PIDB_lookup( tPIHTTP.pResponseDB,
			PIDBTYPE_STRING, KEY_INT_PATH, 0 );

		/* ---
		Check if the filename matches
		--- */
		if ( sFile!=PIString::Empty() )
			{
			if ( !pPath )
				{ return PIAPI_CONTINUE; };
			PIFInfo *pFInfo = HTTPCore_getCachedFile( pPath );
			int iPathMatches = ( sFile==PIFInfo_getName( pFInfo ) );
			HTTPCore_releaseCachedFile( pFInfo );
			if ( !iPathMatches )
				{ return PIAPI_CONTINUE; };
			};

		/* ---
		If this is a real file then attempt to read the host and port
		information from it
		--- */
		PIString sFcgiHost( sHost );
		int iFcgiPort = iPort;
		PIFInfo *pFInfo = HTTPCore_getCachedFile( pPath );
		if ( pFInfo && PIFInfo_exists( pFInfo ) )
			{
			if (	!PIFInfo_isRegular( pFInfo ) ||
					!PIFInfo_isReadable( pFInfo ) )
				{
				HTTPCore_logError( &tPIHTTP, "FastCGI: Fast CGI information \
file '%s' is not readable.", PIFInfo_getPath( pFInfo ) );
				HTTPCore_releaseCachedFile( pFInfo );
				return HTTPUtil_doHTTPError( &tPIHTTP, ST_INTERNALERROR );
				};
			HTTPCore_releaseCachedFile( pFInfo );
			
			/* ---
			Read host and port information 
			--- */
			ifstream ifs( pPath );
			if ( !ifs.good() || ifs.eof() )
				{
				HTTPCore_logError( &tPIHTTP, "FastCGI: Error opening \
fast CGI information file '%s'.", pPath );
				return HTTPUtil_doHTTPError( &tPIHTTP, ST_INTERNALERROR );
				};
			enum { BUF_SIZE=1023 };
			char szBuf[BUF_SIZE+1];
			szBuf[BUF_SIZE] = '\0';
			*szBuf = '\0';
			/* XXX, buffer overrun risk, be careful with lengths! */
			ifs >> szBuf;
			sFcgiHost = szBuf;
			*szBuf = '\0';
			ifs >> szBuf;
			iFcgiPort = atoi( szBuf );
			if ( sFcgiHost==PIString::Empty() || iFcgiPort<=0 )
				{
				HTTPCore_logError( &tPIHTTP, "FastCGI: Bad \
host:port pair (%s:%d) from fast CGI information file '%s'.", 
					(const char *)sFcgiHost, iFcgiPort, pPath );
				return HTTPUtil_doHTTPError( &tPIHTTP, ST_INTERNALERROR );
				};
			}
		else
			{
			HTTPCore_releaseCachedFile( pFInfo );
			};

		/* ---
		Make identification string for error messages and logging
		-- */
		PIString sIdent( "host:port=" );
		sIdent.Concatenate( sFcgiHost );
		enum { SMALL_BUF=63 };
		char szSmall[SMALL_BUF+1];
		sprintf( szSmall, ":%d", iFcgiPort );
		sIdent.Concatenate( szSmall );

		/* ---
		Get Connection Object
		--- */
		PIObject *pIOObject = GetFcgiConnection( tPIHTTP,
			sFcgiHost, iFcgiPort );

		if ( !pIOObject )
			{
			HTTPCore_logError( &tPIHTTP, "FastCGI: Failed to \
establish connection \
to fast cgi application '%s'.", (const char *)sIdent );
			return HTTPUtil_doHTTPError( &tPIHTTP, ST_INTERNALERROR );
			};

		/* --- 
		Make buffer object
		--- */
		PIIOBuffer *pIOBuffer = PIIOBuffer_new( pIOObject );

		int iRet = DoFastCGI( tPIHTTP, pIOBuffer, sIdent );

		PIIOBuffer_delete( pIOBuffer );
		ReleaseFcgiConnection( pIOObject );

		return iRet;
		};
};

/*____________________________________________________________________________*\
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int FastCGI_constructor( PIObject *pObj,
	int iArgc, const char *ppArgv[] )
{
	return HandlerBaseFcgi_constructor( pObj, PI_NEW( FastCGI( pObj,
		iArgc, ppArgv ) ) );
}

#if 0
/*___+++CNF_BEGIN+++___*/
	<Class>
		Name FastCGIClass
		Type LogicExtension
		Library Fcgi
		OnClassLoad HandlerBaseFcgi_onClassLoad
		Constructor FastCGI_constructor
		CopyConstructor HandlerBaseFcgi_copyConstructor
		Destructor HandlerBaseFcgi_destructor
		Execute HandlerBaseFcgi_execute
	</Class>

	<Object>
		Name FastCGI
		Class FastCGIClass
	</Object>
/*___+++CNF_END+++___*/
#endif

