/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2001 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    and Holger Zimmermann (http://pi3web.sourceforge.net/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/HTTP/DelTemp.cpp,v $
 * $Date: 2002/01/26 20:34:00 $
 *
 Description:
	Handler to delete temporary files.

\*____________________________________________________________________________*/
//$SourceTop:$

#include <stdio.h>

#include "HandBase.h"
#include "PIStrStr.h"
#include "HTTPCore.h"

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#if 0
	/*
	** HTML documentation for this handler
	*/
/*___+++HTMLDOC_BEGIN+++___*/
Name:
	DeleteTemporaryFiles

Description:
	Deletes temporary files from a response cycle. Temporary files
	are listed in the "TemporaryFile", string variable of the response
	DB.
	<P>
	If debugging is enabled this handler will not delete any files.

Phase:
	DESTROY

Returns:
	PIAPI_CONTINUE, or PIAPI_ERROR if invoked for a phase other than
	DESTROY.

Note:
Example:
	<PRE>
	&lt;Object&gt;
		Name DeleteTemporaryFiles
		Class DeleteTemporaryFilesClass
	&lt;/Object&gt;

	&lt;Object&gt;
		...
		Destroy DeleteTemporaryFiles
		...
	&lt;/Object&gt;

	</PRE>
/*___+++HTMLDOC_END+++___*/
#endif

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class DeleteTemporaryFiles : public HandlerBaseHTTP
{
private:
protected:
	int Parameter( const char *pVariable, const char *pValue,
		const char *pWhere )
		{
		assert( pVariable && pValue );
		(void)pValue;
		PIOStrStream os;
		os << pWhere << "DeleteTemporaryFiles: ";
		if ( 0 )
			{
			}
		else
			{
			os << "Unknown directive '" << pVariable <<
				"'" << ends;
			CONFIG_ERR( Object(), os.str() );
			return 0;
			};
		return 1;
		};

public:
	DeleteTemporaryFiles( PIObject *pObject, int iArgc, const char *ppArgv[] )
	:	HandlerBaseHTTP( pObject )
		{
		ReadParameters( iArgc, ppArgv );
		};

	int Handle( int iPhase, PIHTTP &tPIHTTP, PIIOBuffer &/* tIOBuffer */ )
		{
		if ( iPhase!=PH_DESTROY )
			{ return PIAPI_ERROR; };

		if ( HTTPCore_debugEnabled() )
			{
			/* don't do anything if debugging is enabled */
			return PIAPI_CONTINUE;
			};

		PIDBIterator *pIter = PIDB_getIterator( tPIHTTP.pResponseDB,
			PIDBTYPE_STRING, KEY_INT_TEMPORARYFILE, 0 );

		if ( pIter )
			{
			for( ;
				PIDBIterator_atValidElement( pIter );
				PIDBIterator_next( pIter ) )
				{
				const char *pFile = (const char *)PIDBIterator_current( pIter,
					0 );
				if ( pFile )
					{
					::remove( pFile );	/* --- remove() from C library --- */
					};
				};
			PIDBIterator_delete( pIter );
			};

		return PIAPI_CONTINUE;
		};
};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int DeleteTemporaryFiles_constructor( PIObject *pObj,
	int iArgc, const char *ppArgv[] )
{
	return HandlerBaseHTTP_constructor( pObj, PI_NEW( DeleteTemporaryFiles( pObj,
		iArgc, ppArgv ) ) );
}

#if 0
/*___+++CNF_BEGIN+++___*/
	<Class>
		Name DeleteTemporaryFilesClass
		Type LogicExtension
		Library HTTP
		OnClassLoad HandlerBaseHTTP_onClassLoad
		Constructor DeleteTemporaryFiles_constructor
		CopyConstructor HandlerBaseHTTP_copyConstructor
		Destructor HandlerBaseHTTP_destructor
		Execute HandlerBaseHTTP_execute
	</Class>

	<Object>
		Name DeleteTemporaryFiles
		Class DeleteTemporaryFilesClass
	</Object>

/*___+++CNF_END+++___*/
#endif

