/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2001 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    and Holger Zimmermann (http://pi3web.sourceforge.net/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/HTTP/Host.cpp,v $
 * $Date: 2002/01/26 20:34:00 $
 *
 Description:
	Dummy handler object used as placeholder for DB values associated
	with a virtual host.

\*____________________________________________________________________________*/
//$SourceTop:$

#include "HandBase.h"
#include "HTTPCore.h"
#include "PICompat.h"			/* from Base library */
#include "PIString.h"			/* from Base library */

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/

#if 0
	/*
	** HTML documentation for this handler
	*/
/*___+++HTMLDOC_BEGIN+++___*/
Name:
	VirtualHost

Description:
	A virtual host object is a placeholder for variable-value attribute
	pairs associated with virtual hosts. 
	This object creates its own PIDB and stores parameters in it for
	reference. A pointer to this PIDB is access by other parts of the
	HTTP server via the GetHostDB() method of the PIHTTP object.

Options:
	Ad hoc options. All parameters are stored in this object PIDB with
	type PIDBTYPE_RFC822.

Phase:
	No phase. It is an error to invoke this object as part of phase handling.

Returns:
	PIAPI_ERROR always, because this is not a handler.

Example:
	<PRE>
	&lt;Object&gt;
		Name VirtualHost
		Class VirtualHostClass
		HostName "host1"
		Administrator "jroy@johnroy.com"
	&lt;/Object&gt;

	</PRE>
/*___+++HTMLDOC_END+++___*/
#endif

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class VirtualHost : public HandlerBaseHTTP
{
private:
protected:
	int Parameter( const char *pVariable, const char *pValue,
		const char * /* pWhere */ )
		{
		assert( pVariable );
		assert( pValue );
		PIDB *pDB = PIObject_getDB( Object() );
		assert( pDB );

		PIDB_add( pDB, PIDBTYPE_RFC822, pVariable, (void *)pValue, 0 );
		return 1;
		};

public:
	VirtualHost( PIObject *pObject, int iArgc, const char *ppArgv[] )
	:	HandlerBaseHTTP( pObject )
		{
		/* --- create a new DB --- */
		PIDB *pDB = PIDB_new( PIObject_getDB( pObject ), "VirtualHostDB" );
		
		/* --- add the name of this object into it --- */
		PIDB_add( pDB, PIDBTYPE_RFC822, KEY_INT_OBJECTNAME,
			(void *)(const char *)PIObject_getName( pObject ), 0 );
		PIObject_setDB( pObject, pDB );

		ReadParameters( iArgc, ppArgv );
		};

	int Handle( int, PIHTTP &, PIIOBuffer & )
		{
		/* --- this is an error --- */
		return PIAPI_ERROR;
		};
};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int VirtualHost_constructor( PIObject *pObj,
	int iArgc, const char *ppArgv[] )
{
	return HandlerBaseHTTP_constructor( pObj, PI_NEW( VirtualHost( pObj,
		iArgc, ppArgv ) ) );
}

#if 0
/*___+++CNF_BEGIN+++___*/
	<Class>
		Name VirtualHostClass
		Type LogicExtension
		Library HTTP
		OnClassLoad HandlerBaseHTTP_onClassLoad
		Constructor VirtualHost_constructor
		CopyConstructor HandlerBaseHTTP_copyConstructor
		Destructor HandlerBaseHTTP_destructor
		Execute HandlerBaseHTTP_execute
	</Class>

	<Object>
		Name VirtualHost
		Class VirtualHostClass
	</Object>

/*___+++CNF_END+++___*/
#endif

