/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2001 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in
	the documentation and/or other materials provided with the
	distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
	mentioning features must contain the following
	acknowledgment:
	"This product includes software developed by John Roy
	and Holger Zimmermann (http://pi3web.sourceforge.net/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Intrface/General.cpp,v $
 * $Date: 2002/08/02 06:58:56 $
 *
 Description:

\*____________________________________________________________________________*/
//$SourceTop:$

#include <iostream.h>
#include <stdio.h>
#include <windowsx.h>

#include "IFaceBse.h"
#include "PIStrStr.h"
#include "StrToken.h"

#include "QConfig.h"
#include "General.hrc"

/*
** #define D { cerr << __FILE__ << ": " << __LINE__ << endl; }
*/
#define D

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#define C_RESOURCENAME                    "GeneralPage"

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#if 0
	/*
	** HTML documentation for this handler
	*/
/*___+++HTMLDOC_BEGIN+++___*/
Name:
    GeneralPage

Description:
    Creates the eponymous property page and is used by the PropSheet object.

Notes:
	The changes you make in this won't partial immediately affect the behaviour of
	the server but only after restart. This applies to the connection parameters.
	Changes may lead to a configuration, which is inapplicable for remote administration
	or requires corresponding changes in the configuration of the remote interface
	application, particular of the PropSheet object.

Example:
	<PRE>
	&lt;Object&gt;
		Name GeneralPage
		Class GeneralPageClass
		PropertyPage GeneralPage
	&lt;/Object&gt;

	</PRE>

/*___+++HTMLDOC_END+++___*/
#endif

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
static BOOL CALLBACK fnDialogProc(HWND hDlg, UINT uMsg, WPARAM wParam,
    LPARAM lParam);

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class GeneralPage : public IFaceBase
{
private:
    QConfig *pConfig;
    PROPSHEETPAGE tPsp;
    Interface *pInterface;

    /*
    ** window private values
    */
    HWND hWndHost;
    HWND hWndIP;
    HWND hWndLocal;
    HWND hWndEdit;
    HWND hWndPort;
    HWND hWndURL;
    HWND hWndAdmin;
	HWND hWndSsl;
	HWND hWndRemote;
	HWND hWndVerbose;
    PIString pHost;
    PIString pIP;
    PIString pLocal;
    int __iCurrentRadioButton;

	int Parameter( const char *pVariable, const char *pValue,
		const char *pWhere )
		{
		assert( pVariable && pValue );

		PIOStrStream os;

		os << pWhere << "GeneralPage: ";

        if ( 0 )
            {
            }
		else
			{
			os << "Unknown directive '" << pVariable <<
				"'" << ends;
			CONFIG_ERR( Object(), os.str() );
			return 0;
			};

		return 1;
		};

public:
	GeneralPage( PIObject *pObject, int iArgc, const char *ppArgv[] )
	:	IFaceBase( pObject ),
        pConfig( FindParentQConfig( PIObject_getDB( pObject ) ) ),
        pInterface( 0 ),
        hWndHost( 0 ),
        hWndIP( 0 ),
        hWndLocal( 0 ),
        hWndEdit( 0 ),
        hWndPort( 0 ),
        hWndURL( 0 ),
        hWndAdmin( 0 ),
		pHost( "localhost" ),
		pIP( "127.0.0.1" ),
		pLocal ( "localhost" ),
        __iCurrentRadioButton( 0 )
		{
		ReadParameters( iArgc, ppArgv );

        if ( !IsOK() )
            { return; };

        /*
        ** Setup the PROPSHEETPAGE
        */
        memset( &tPsp, 0, sizeof( PROPSHEETPAGE ) );
        tPsp.dwSize = sizeof( PROPSHEETPAGE );
        tPsp.dwFlags = PSP_DEFAULT;
        tPsp.hInstance = (HINSTANCE)PIDLL_getSystemHandle( PIClass_getLibrary(
            PIObject_getClass( Object() ) ) );

        /* Status page */
        tPsp.pszTemplate = C_RESOURCENAME;
        tPsp.pfnDlgProc = (DLGPROC)fnDialogProc;
        tPsp.lParam = (LPARAM)this;

        /*
        ** Set the data DB of this object
        */
        PIDB *pDB = PIObject_getDB( Object() );
        PIDB_add( pDB, PIDBTYPE_OPAQUE, INTRFACE_WINDOWTYPE,
            (void *)INTRFACE_PROPPAGE, 0 );
        assert(!PIDB_lookup( pDB, PIDBTYPE_OPAQUE, INTRFACE_PROPSHEETPAGE, 0));
        PIDB_add( pDB, PIDBTYPE_OPAQUE, INTRFACE_PROPSHEETPAGE,
            (void *)&tPsp, 0 );
        
        /*
        ** Done
        */
		};

	~GeneralPage()
		{
		};

    virtual int Handle( Interface &tWindow )
        {
        switch( tWindow.iMessage )
            {
            case INTRFACE_MSG_INIT:
                pInterface = &tWindow;
                break;

            case INTRFACE_MSG_WIN32CREATE:
                return PIAPI_ABORT;

            case INTRFACE_MSG_LOAD:
                return PIAPI_ABORT;

            case INTRFACE_MSG_DEFAULTS:
				M_AddValue( pConfig, "Internal", "URLProtocol", "http://" );
				M_AddValue( pConfig, "Internal", "DefaultPort", "80" );
				M_AddValue( pConfig, "Internal", "ServerIOObjectLine", "\tIOObject ServerIOObject" );
				M_AddValue( pConfig, "Internal", "DefaultMIMEType", "application/octet-stream" );
				M_AddValue( pConfig, "Internal", "HandlerLine", DEFAULT_HANDLERS );
				M_AddValue( pConfig, "Identity", "Use", "local" );
				M_AddValue( pConfig, "General", "Port", "80" );
				M_AddValue( pConfig, "General", "Host", "localhost" );
				M_AddValue( pConfig, "General", "Admin", "webmaster@yourhost.dom" );
				M_AddValue( pConfig, "General", "ErrorMessages", "Short" );
                break;

            default:
                assert( 0 );
                return PIAPI_ABORT;
            };

        return PIAPI_COMPLETED;
        };

	PIString *Internal_stringFromRadioButton( int iId )
	    {
		switch( iId )
			{
			case IDC_REMOTEHOST:	return &pHost;
			case IDC_REMOTEIP:	return &pIP;
			case IDC_LOCAL:	return &pLocal;
			default:
				assert( 0 );
			};
	
		return 0;
	    };
	
	void Internal_radioButtonChange( int iNewId )
	    {
		enum { BUF_SIZE=1023 };
		char szBuf[BUF_SIZE+1];
	
		/* --- Get current value --- */
		*szBuf = '\0';
		SendMessage( hWndEdit, WM_GETTEXT, BUF_SIZE, (LPARAM)szBuf );
		*(Internal_stringFromRadioButton( __iCurrentRadioButton )) = szBuf;
	
		__iCurrentRadioButton = iNewId;
	
		SendMessage( hWndEdit, WM_SETTEXT, 0, (LPARAM)(const char *)
			*(Internal_stringFromRadioButton( __iCurrentRadioButton )) );
	    };
	
	void Internal_updateURL( HWND hDlg )
	    {
		enum { BUF_SIZE=1023 };
		char szBuf[BUF_SIZE+1];
	
		/* --- Get current value --- */
		assert( M_LookupValue( pConfig, "Internal", "URLProtocol" ) );
		strcpy( szBuf, M_LookupValue( pConfig, "Internal", "URLProtocol" ) );
		int iLen = strlen( szBuf );
		SendMessage( hWndEdit, WM_GETTEXT, BUF_SIZE-7, (LPARAM)&(szBuf[iLen]) );
		iLen = strlen( szBuf );
		szBuf[iLen++] = ':';
		SendMessage( hWndPort, WM_GETTEXT, BUF_SIZE-iLen, (LPARAM)&(szBuf[iLen]) );
		assert( M_LookupValue( pConfig, "Internal", "DefaultPort" ) );
		if ( atoi( &(szBuf[iLen]) )==atoi(M_LookupValue( pConfig, "Internal", 
				"DefaultPort" )) )
			{ szBuf[--iLen] = '\0'; };
	
		/* --- now update the URL --- */
		SendMessage( hWndURL, WM_SETTEXT, 0, (LPARAM)szBuf );
		
		/* --- mark property sheet as modified --- */
		Changed( hDlg );
	    };
	
	BOOL DialogProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
	    {
	    enum { BUF_SIZE=1023 };
	    char szBuf[BUF_SIZE+1];

	    switch (uMsg)
	        {
            case WM_INITDIALOG:
				{
				const char *pA = M_LookupValue( pConfig, "General", "Admin" );
				hWndHost = GetDlgItem( hDlg, IDC_REMOTEHOST );
				hWndIP = GetDlgItem( hDlg, IDC_REMOTEIP );
				hWndLocal = GetDlgItem( hDlg, IDC_LOCAL );
				hWndEdit = GetDlgItem( hDlg, IDC_HOST );
				hWndPort = GetDlgItem( hDlg, IDC_PORT );
				hWndURL = GetDlgItem( hDlg, IDC_URL );
				hWndAdmin = GetDlgItem( hDlg, IDC_ADMIN );
				hWndSsl = GetDlgItem( hDlg, IDC_SSL );
				hWndRemote = GetDlgItem( hDlg, IDC_REMOTE );
				hWndVerbose = GetDlgItem( hDlg, IDC_VERBOSE );

				assert( hWndHost && hWndIP && hWndLocal );
				assert( hWndEdit && hWndPort && hWndURL && hWndAdmin );

				assert( pA );

				::SendMessage( hWndAdmin, WM_SETTEXT, 0, (LPARAM)pA );

				::SendMessage( hWndHost, BM_SETCHECK, BST_UNCHECKED, 0 );
				::SendMessage( hWndIP, BM_SETCHECK, BST_UNCHECKED, 0 );
				::SendMessage( hWndLocal, BM_SETCHECK, BST_UNCHECKED, 0 );
				::SendMessage( hWndSsl, BM_SETCHECK, BST_UNCHECKED, 0 );
				::SendMessage( hWndRemote, BM_SETCHECK, BST_UNCHECKED, 0 );
				::SendMessage( hWndVerbose, BM_SETCHECK, BST_UNCHECKED, 0 );

				/* --- just for sanity --- */
				::SendMessage( hWndEdit, EM_SETLIMITTEXT, 256, 0 );
				::SendMessage( hWndPort, EM_SETLIMITTEXT, 256, 0 );

				::SendMessage( ::GetDlgItem( hDlg, IDC_INSTALLDIR ), WM_SETTEXT, 0,
					(LPARAM)M_LookupValue( pConfig, "General", "InstallDir" ));
				::SendMessage( ::GetDlgItem( hDlg, IDC_OS ), WM_SETTEXT, 0,
					(LPARAM)M_LookupValue( pConfig, "General", "OS" ));

				const char *pUse = M_LookupValue( pConfig, "Identity", "Use" );
				const char *pTheHost = M_LookupValue( pConfig, "General", "Host" );
				SendMessage( hWndEdit, WM_SETTEXT, 0, (LPARAM)(const char *)pTheHost );
				if ( !stricmp( pUse, "local" ) )
					{
					SendMessage( hWndLocal, BM_SETCHECK, BST_CHECKED, 0 );
					__iCurrentRadioButton = IDC_LOCAL;
					SendMessage( hWndEdit, EM_SETREADONLY, (BOOL)TRUE, 0 );
					}
				else if ( !stricmp( pUse, "remote" ) )
					{
					SendMessage( hWndHost, BM_SETCHECK, BST_CHECKED, 0 );
					__iCurrentRadioButton = IDC_REMOTEHOST;
					SendMessage( hWndEdit, EM_SETREADONLY, (BOOL)FALSE, 0 );
					}
				else if ( !stricmp( pUse, "ip" ) )
					{
					SendMessage( hWndIP, BM_SETCHECK, BST_CHECKED, 0 );
					__iCurrentRadioButton = IDC_REMOTEIP;
					SendMessage( hWndEdit, EM_SETREADONLY, (BOOL)FALSE, 0 );
					};

				const char *pTest = M_LookupValue( pConfig, "Internal", "HandlerLine" );
				if ( strstr( pTest, "TopSiteRoot" ) )
					{
					SendMessage( hWndRemote, BM_SETCHECK, BST_CHECKED, 0 );
					}

				if ( strstr( M_LookupValue( pConfig, "General", "ErrorMessages" ), "Verbose" ))
					{
					SendMessage( hWndVerbose, BM_SETCHECK, BST_CHECKED, 0 );
					}

				// Port and SSL flag are set and URL is updated
				// in PSN_SETACTIVE handler
			
				Unchanged( hDlg );
				break;
				};

		case WM_COMMAND:
			if ( GET_WM_COMMAND_CMD(wParam, lParam)==BN_CLICKED )
				{
				switch( GET_WM_COMMAND_ID( wParam, lParam) )
					{
					case IDC_REMOTEHOST:
						Internal_radioButtonChange( IDC_REMOTEHOST );	
						SendMessage( hWndEdit, EM_SETREADONLY, (BOOL)FALSE, 0 );
						break;

					case IDC_REMOTEIP:	
						Internal_radioButtonChange( IDC_REMOTEIP );	
						SendMessage( hWndEdit, EM_SETREADONLY, (BOOL)FALSE, 0 );
						break;

					case IDC_LOCAL:	
						Internal_radioButtonChange( IDC_LOCAL );	
						SendMessage( hWndEdit, EM_SETREADONLY, (BOOL)TRUE, 0 );
						break;

					case IDC_SSL:
						*szBuf = '\0';
						SendMessage( hWndPort, WM_GETTEXT, BUF_SIZE, (LPARAM)szBuf );
						if ( ::SendMessage( hWndSsl, BM_GETCHECK, 0, 0 )==BST_CHECKED )
							{
							M_ReplaceValue( pConfig, "Internal", "URLProtocol", "https://" );
							M_ReplaceValue( pConfig, "Internal", "DefaultPort", "443" );
							if (!stricmp(szBuf, "80")) {
								SendMessage( hWndPort, WM_SETTEXT, 0, (LPARAM)"443");
								};
							}
						else
							{
							M_ReplaceValue( pConfig, "Internal", "URLProtocol", "http://" );
							M_ReplaceValue( pConfig, "Internal", "DefaultPort", "80" );
							if (!stricmp(szBuf, "443")) {
								SendMessage( hWndPort, WM_SETTEXT, 0, (LPARAM)"80");
								};
							};
						Internal_updateURL( hDlg );
						break;

					case IDC_REMOTE:
					case IDC_VERBOSE:
						Changed( hDlg );
						break;

					default:;
					};
				}
			else if ( GET_WM_COMMAND_CMD(wParam, lParam)==EN_CHANGE )
				{ Internal_updateURL( hDlg ); };
			break;

		case WM_DESTROY:
			break;

			case WM_NOTIFY:
				switch( ((NMHDR *)lParam)->code )
					{
					case PSN_SETACTIVE:
						{
						int iSaveChanged = IsChanged();
						SendMessage( hWndPort, WM_SETTEXT, 0, (LPARAM)
							M_LookupValue( pConfig, "General", "Port" ) );

						const char *pTest = M_LookupValue( pConfig, "Internal", \
							"ServerIOObjectLine" );
						if ( strstr( pTest, "SSLIO" ))
							{
							SendMessage( hWndSsl, BM_SETCHECK, BST_CHECKED, 0 );
							}
						Internal_updateURL( hDlg );
						if ( !iSaveChanged ) { Unchanged( hDlg ); };
						SetWindowLong( hDlg, DWL_MSGRESULT, FALSE );
						return (TRUE);
						}

					case PSN_QUERYCANCEL:
						return SHEET->Cancel(pInterface, hDlg);
	
					case PSN_KILLACTIVE:
						if ( IsChanged() )
						{
						*szBuf = '\0';
						::SendMessage( ::GetDlgItem( hDlg, IDC_INSTALLDIR ),
							WM_GETTEXT, BUF_SIZE, (LPARAM)szBuf );
						M_ReplaceValue( pConfig, "General", "InstallDir", szBuf ); 

						*szBuf = '\0';
						::SendMessage( ::GetDlgItem( hDlg, IDC_OS ),
							WM_GETTEXT, BUF_SIZE, (LPARAM)szBuf );
						M_ReplaceValue( pConfig, "General", "OS", szBuf ); 

						*szBuf = '\0';
						SendMessage( hWndPort, WM_GETTEXT, BUF_SIZE, (LPARAM)szBuf );
						int iPort = atoi( szBuf );
						if ( iPort<0 || iPort>65535 )
							{
							/*
							** Give message and abort change
							*/
							MessageBox( hDlg, "Port must be in the range [0..65536]",
								"Port Field Error", MB_OK );
							SetFocus( hWndPort );
							SendMessage( hWndPort, EM_SETSEL, 0, -1 );
							SetWindowLong( hDlg, DWL_MSGRESULT, PSNRET_INVALID_NOCHANGEPAGE );
							return (TRUE);
							};
						M_ReplaceValue( pConfig, "General", "Port", szBuf ); 
						SendMessage( hWndAdmin, WM_GETTEXT, BUF_SIZE, (LPARAM)szBuf );
						M_ReplaceValue( pConfig, "General", "Admin", szBuf ); 
						SendMessage( hWndEdit, WM_GETTEXT, BUF_SIZE, (LPARAM)szBuf );
						M_ReplaceValue( pConfig, "General", "Host", szBuf ); 
						SendMessage( hWndURL, WM_GETTEXT, BUF_SIZE, (LPARAM)szBuf );
						M_ReplaceValue( pConfig, "Identity", "URL", szBuf );
						switch( __iCurrentRadioButton )
							{
							case IDC_LOCAL:
								M_ReplaceValue( pConfig, "Identity", "Use", "Local" );	
								break;

							case IDC_REMOTEHOST:
								M_ReplaceValue( pConfig, "Identity", "Use", "Remote" );	
								break;

							case IDC_REMOTEIP:
								M_ReplaceValue( pConfig, "Identity", "Use", "IP" );	
								break;

							default:
								assert( 0 );
							};

						if ( ::SendMessage( hWndSsl, BM_GETCHECK, 0, 0 )==BST_CHECKED )
							{
							M_ReplaceValue( pConfig, "Internal", "ServerIOObjectLine", \
								"\tIOObject SSLIOObject" );
							}
						else
							{
							M_ReplaceValue( pConfig, "Internal", "ServerIOObjectLine", \
								"\tIOObject ServerIOObject" );
							}

						if ( ::SendMessage( hWndRemote, BM_GETCHECK, 0, 0 )==BST_CHECKED )
							{
							M_ReplaceValue( pConfig, "Internal", "HandlerLine", \
								"\tHandlers TopSiteRoot" );
  							}
						else
							{
							M_ReplaceValue( pConfig, "Internal", "HandlerLine", \
								DEFAULT_HANDLERS );
							}

						if ( ::SendMessage( hWndVerbose, BM_GETCHECK, 0, 0 )==BST_CHECKED )
							{
							/*
							** Give warning message
							*/
							if ( __iCurrentRadioButton!=IDC_LOCAL )
								{
								MessageBox( hDlg, "This kind of error messages are extremely verbose. \
Don't even think about using them in an internet server.", "Verbose Errors Warning", MB_OK | MB_ICONWARNING );
								}
							M_ReplaceValue( pConfig, "General", "ErrorMessages", "Verbose" );
							}
						else
							{
							M_ReplaceValue( pConfig, "General", "ErrorMessages", "Short" );
							}
						}
						SetWindowLong( hDlg, DWL_MSGRESULT, FALSE );
						return (TRUE);
	
					case PSN_APPLY:
						return SHEET->SaveConfig(pInterface, hDlg);

					default:
						return (FALSE);
					};
				break;

            default:
                return (FALSE);
            };

        return FALSE;
        };
};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static BOOL CALLBACK fnDialogProc(HWND hDlg, UINT uMsg, WPARAM wParam,
    LPARAM lParam)
{
    if ( uMsg==WM_INITDIALOG )
        {
        ::SetWindowLong( hDlg, DWL_USER, ((LPPROPSHEETPAGE)lParam)->lParam );
        };

    GeneralPage *pPage = (GeneralPage *)::GetWindowLong( hDlg, DWL_USER );
    return pPage ? pPage->DialogProc( hDlg, uMsg, wParam, lParam ) : FALSE;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int GeneralPage_constructor( PIObject *pObj,
	int iArgc, const char *ppArgv[] )
{
	return IFaceBase_constructor( pObj, PI_NEW( GeneralPage( pObj,
		iArgc, ppArgv ) ) );
}

#if 0
/*___+++CNF_BEGIN+++___*/
	<Class>
		Name GeneralPageClass
		Type LogicExtension
		Library IntrFace
		OnClassLoad IFaceBase_onClassLoad
		Constructor GeneralPage_constructor
		CopyConstructor IFaceBase_copyConstructor
		Destructor IFaceBase_destructor
		Execute IFaceBase_execute
	</Class>

	<Object>
		Name GeneralPage
		Class GeneralPageClass
	</Object>

/*___+++CNF_END+++___*/
#endif

