/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2001 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in
	the documentation and/or other materials provided with the
	distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
	mentioning features must contain the following
	acknowledgment:
	"This product includes software developed by John Roy
	and Holger Zimmermann (http://pi3web.sourceforge.net/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Intrface/LogFile.cpp,v $
 * $Date: 2002/08/02 06:58:56 $
 *
 Description:

\*____________________________________________________________________________*/
//$SourceTop:$

#include <iostream.h>
#include <stdio.h>
#include <windowsx.h>

#include "IFaceBse.h"
#include "PIStrStr.h"
#include "StrToken.h"

#include "QConfig.h"
#include "LogFile.hrc"

/*
** #define D { cerr << __FILE__ << ": " << __LINE__ << endl; }
*/
#define D

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#define C_RESOURCENAME                    "LogFilePage"

static const char *__pLoggingTypes[] =
{
"Access (CLF)",
"Referer Logfile",
0
};

static const char *__pLogfiles[] =
{
"Access (CLF)\tLogs\\Access.txt",

/* --- leave this last --- */
0
};

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#if 0
	/*
	** HTML documentation for this handler
	*/
/*___+++HTMLDOC_BEGIN+++___*/
Name:
    LogFilePage

Description:
    Creates the eponymous property page and is used by the PropSheet object.

Example:
	<PRE>
	&lt;Object&gt;
		Name LogFilePage
		Class LogFilePageClass
		PropertyPage LogFilePage
	&lt;/Object&gt;

	</PRE>

/*___+++HTMLDOC_END+++___*/
#endif

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
static BOOL CALLBACK fnDialogProc(HWND hDlg, UINT uMsg, WPARAM wParam,
    LPARAM lParam);

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class LogFilePage : public IFaceBase
{
private:
    QConfig *pConfig;
    PROPSHEETPAGE tPsp;
    Interface *pInterface;

    /*
    ** window private values
    */
    HWND hWndList;
    HWND hWndType;
    HWND hWndPath;
    HWND hWndAdd;
    HWND hWndReplace;
    HWND hWndDelete;
    DblList lTypes;
    DblList lPaths;

	int Parameter( const char *pVariable, const char *pValue,
		const char *pWhere )
		{
		assert( pVariable && pValue );

		PIOStrStream os;

		os << pWhere << "LogFilePage: ";

        if ( 0 )
            {
            }
		else
			{
			os << "Unknown directive '" << pVariable <<
				"'" << ends;
			CONFIG_ERR( Object(), os.str() );
			return 0;
			};

		return 1;
		};

public:
	LogFilePage( PIObject *pObject, int iArgc, const char *ppArgv[] )
	:	IFaceBase( pObject ),
        pConfig( FindParentQConfig( PIObject_getDB( pObject ) ) ),
        hWndList( 0 ),
        hWndType( 0 ),
        hWndPath( 0 ),
        hWndAdd( 0 ),
        hWndReplace( 0 ),
        hWndDelete( 0 ),
        pInterface( 0 )
		{
        /*
        ** Read input parameters
        */
		ReadParameters( iArgc, ppArgv );

        if ( !IsOK() )
            { return; };

        /*
        ** Setup the PROPSHEETPAGE
        */
        memset( &tPsp, 0, sizeof( PROPSHEETPAGE ) );
        tPsp.dwSize = sizeof( PROPSHEETPAGE );
        tPsp.dwFlags = PSP_DEFAULT;
        tPsp.hInstance = (HINSTANCE)PIDLL_getSystemHandle( PIClass_getLibrary(
            PIObject_getClass( Object() ) ) );

        /* Status page */
        tPsp.pszTemplate = C_RESOURCENAME;
        tPsp.pfnDlgProc = (DLGPROC)fnDialogProc;
        tPsp.lParam = (LPARAM)this;

        /*
        ** Set the data DB of this object
        */
        PIDB *pDB = PIObject_getDB( Object() );
        PIDB_add( pDB, PIDBTYPE_OPAQUE, INTRFACE_WINDOWTYPE,
            (void *)INTRFACE_PROPPAGE, 0 );
        PIDB_add( pDB, PIDBTYPE_OPAQUE, INTRFACE_PROPSHEETPAGE,
            (void *)&tPsp, 0 );
        
        /*
        ** Done
        */
		};

	~LogFilePage()
		{
		};

    virtual int Handle( Interface &tWindow )
        {
        switch( tWindow.iMessage )
            {
            case INTRFACE_MSG_INIT:
                pInterface = &tWindow;
                break;

            case INTRFACE_MSG_WIN32CREATE:
                return Win32Create( tWindow );

            case INTRFACE_MSG_LOAD:
                return PIAPI_ABORT;

            case INTRFACE_MSG_DEFAULTS:
                {
                for( int i=0; __pLoggingTypes[i]; i++ )
                    {
                    M_AddValue( pConfig, "Internal", "LoggingTypes",
                        __pLoggingTypes[i] );
                    };
                };
                {
                for( int i=0; __pLogfiles[i]; i++ )
                    {
                    M_AddValue( pConfig, "Logfiles", "Logging",
                        __pLogfiles[i] );
                    };
                };
                break;

            default:
                assert( 0 );
                return PIAPI_ABORT;
            };

        return PIAPI_COMPLETED;
        };

    void Internal_updateButtons( WPARAM wParam, LPARAM lParam )
        {
	    int iCBSel = ::SendMessage( hWndType, CB_GETCURSEL, 0, 0 );
	    const char *pType = 0;

	    int iExistingIndex = ::SendMessage( hWndList, LB_GETCURSEL, 0, 0 );

	    /*
	    ** Set buttons
	    */
	    if ( iCBSel<0 || ::SendMessage( hWndType, CB_GETLBTEXTLEN, iCBSel, 0 )
            ==0 )
		    {
		    if ( ::IsWindowEnabled( hWndAdd ) )
                ::EnableWindow( hWndAdd, FALSE );
		    if ( ::IsWindowEnabled( hWndReplace ) )
                ::EnableWindow( hWndReplace, FALSE );
		    }
	    else if ( SendMessage( hWndPath, EM_LINELENGTH, 0, 0 )!=0 )
		    {
		    if ( iExistingIndex!=LB_ERR )/* ---disable add, enable replace--- */
			    {
			    if ( ::IsWindowEnabled( hWndAdd ) )
                    ::EnableWindow( hWndAdd, FALSE );
			    if ( !::IsWindowEnabled( hWndReplace ) )
                    ::EnableWindow( hWndReplace, TRUE );
			    }
		    else					/* --- disable replace, enable add --- */
			    {
			    if ( !::IsWindowEnabled( hWndAdd ) )
                    ::EnableWindow( hWndAdd, TRUE );
			    if ( ::IsWindowEnabled( hWndReplace ) )
                    ::EnableWindow( hWndReplace, FALSE );
			    };
		    };

	    /*
	    ** If there is a selected item in the listbox then
	    ** enable delete, otherwise disable delete	
	    */
	    if ( ::SendMessage( hWndList, LB_GETCURSEL, 0, 0 )==LB_ERR )
		    { ::EnableWindow( hWndDelete, FALSE ); }
	    else
		    { ::EnableWindow( hWndDelete, TRUE ); };
        };

    void Internal_addLine( const char *pLine )
        {
        assert( pLine );
	    assert( lTypes.Size()==lPaths.Size() );

	    int i;
	    /* --- scan to tab --- */
	    for( i=0; pLine[i] && pLine[i]!='\t'; i++ );
	
	    lTypes.Append( (DblList::type)PI_NEW( PIString( pLine, i ) ) );

	    /* --- scan over whitespace --- */
	    for( ; pLine[i] && isspace(pLine[i]); i++ );
	    lPaths.Append( (DblList::type)PI_NEW( PIString( &(pLine[i]) ) ) );	

	    assert( lTypes.Size()==lPaths.Size() );
        };

    BOOL Internal_save( HWND hDlg, char *pszBuf, int iBufSize )
        {
		int i;
		int iCount = ::SendMessage( hWndList, LB_GETCOUNT, 0, 0 );

		/*
		** Save all mime types
		*/
		/* --- remove old ones first --- */
		M_DeleteValue( pConfig, "Logfiles", "Logging" );
		for( i=0; i<iCount; i++ )
			{
			*pszBuf = '\0';
			pszBuf[iBufSize] = '\0';
			/*
			** safety only
			*/
			if ( ::SendMessage( hWndList, LB_GETTEXTLEN, 0, 0 ) >= iBufSize )
				{
				::MessageBox( hDlg, "Logging specification too long", 
					"Internal Error", MB_OK );
				break;
				};
			::SendMessage( hWndList, LB_GETTEXT, i, (LPARAM)pszBuf );
				M_AddValue( pConfig, "Logfiles", "Logging", pszBuf );
			};
        ::SetWindowLong( hDlg, DWL_MSGRESULT, FALSE );
        return (TRUE);
        };

    int Win32Create( Interface &tWindow )
		{
        /* --- property sheet pages should not be created --- */
        return PIAPI_ABORT;
		};

    BOOL DialogProc(HWND hDlg, UINT uMsg, WPARAM wParam,
	LPARAM lParam)
        {
		enum { BUF_SIZE=1023 };
		char szBuf[BUF_SIZE+1];
		int aTabstop[2];
	
		switch (uMsg)
			{
			case WM_INITDIALOG:
			    HourglassOn();
			    {
			    const char *pTmp;
			    int iTmp;
			    hWndList = GetDlgItem( hDlg, IDC_LIST );
			    hWndType = GetDlgItem( hDlg, IDC_TYPE );
			    hWndPath= GetDlgItem( hDlg, IDC_PATH );
			    hWndAdd = GetDlgItem( hDlg, IDC_ADD );
			    hWndReplace = GetDlgItem( hDlg, IDC_REPLACE );
			    hWndDelete= GetDlgItem( hDlg, IDC_DELETE );

			    assert( hWndList && hWndType && hWndPath );
			    assert( hWndAdd && hWndReplace && hWndDelete );

			    ::EnableWindow( hWndAdd, FALSE );
			    ::EnableWindow( hWndReplace, FALSE );
			    ::EnableWindow( hWndDelete, FALSE );

			    /*
			    ** Load all logging types.
			    */
			    iTmp = 0;
			    for(;; iTmp++ )
				    {
				    pTmp = M_LookupValueEx( pConfig, "Internal", 
					    "LoggingTypes", iTmp );
				    if ( !pTmp )
					    { break; };

				    ::SendMessage( hWndType, CB_ADDSTRING, 0, (LPARAM)pTmp );
				    };

			    /*
			    ** Limit edit field lengths, really just a sanity thing
			    */
			    ::SendMessage( hWndPath, EM_SETLIMITTEXT, 256, 0 );

			    /*
			    ** Setup tabstops
			    */
			    aTabstop[0] = 0;
			    aTabstop[1] = 75;
			    ::SendMessage( hWndList, LB_SETTABSTOPS, 2, (LPARAM)&aTabstop );

			    /*
			    ** Load all logfiles
			    */
			    iTmp = 0;
			    for(;; iTmp++ )
				    {
				    pTmp = M_LookupValueEx( pConfig, "Logfiles", 
					    "Logging", iTmp );
				    if ( !pTmp )
					    { break; };

				    if ( ::SendMessage( hWndList, LB_ADDSTRING, 0,
                        (LPARAM)pTmp )>=0 )
					    {
					    Internal_addLine( pTmp );
					    }
				    else
					    { break; /* error */; };
				    };
			    }

			    HourglassOff();
			    break;
	
			case WM_DESTROY:
                {
			    int i;
			    int iCount = ::SendMessage( hWndList, LB_GETCOUNT, 0, 0 );

			    /* --- remove old ones first --- */
			    DblListIterator x( lTypes );
			    DblListIterator y( lPaths );
			    for( i=0; i<iCount; i++, x++, y++ )
				    {
				    /*
				    ** Delete this entry in lists
				    */
				    assert( !x.BadIndex() && !y.BadIndex() );
				    PI_DELETE( (PIString *)x.Current() );
				    PI_DELETE( (PIString *)y.Current() );
				    };
			    lTypes.Clear();
			    lPaths.Clear();
                };
			    break;

			case WM_COMMAND:
				switch (GET_WM_COMMAND_ID(wParam, lParam))
                    {
				    case IDC_ADD:
					    {
					    int i;
					    *szBuf = '\0';
					    ::SendMessage( hWndType, WM_GETTEXT, BUF_SIZE,
                            (LPARAM)szBuf );
					    for( i=0; i<BUF_SIZE && szBuf[i] && szBuf[i]!='\t';
                            i++ );
					    if ( szBuf[i] ) { szBuf[i] = '\0'; };
					    lTypes.Append( (DblList::type)PI_NEW( PIString( szBuf )
                            ) );
					    szBuf[i++] = '\t';
					    ::SendMessage( hWndPath, WM_GETTEXT, BUF_SIZE-i,
                            (LPARAM)&(szBuf[i]) );
					    lPaths.Append( (DblList::type)PI_NEW( PIString(
                            &(szBuf[i]) ) ) );
					    ::SendMessage( hWndList, LB_ADDSTRING,0,(LPARAM)szBuf );

					    /*
					    ** Clear edit fields, focus to list
					    */
					    ::SendMessage( hWndType, CB_SETCURSEL, (WPARAM)-1, 0 );
					    ::SendMessage( hWndPath, WM_SETTEXT, 0, (LPARAM)"" );
                        ::SetFocus( hWndList );
					    Changed( hDlg );
					    };
					    break;
	
				    case IDC_REPLACE:
					    {
					    /*
					    ** Get selected item index 
					    */
					    int iCurSel = ::SendMessage(hWndList, LB_GETCURSEL,0,0);

					    /*
					    ** Get text to add and add it to internal lists only
					    */
					    int i;
					    *szBuf = '\0';
					    ::SendMessage( hWndType, WM_GETTEXT, BUF_SIZE,
                            (LPARAM)szBuf );
					    for( i=0;i<BUF_SIZE && szBuf[i] && szBuf[i]!='\t';i++);
					    if ( szBuf[i] ) { szBuf[i] = '\0'; };
					    lTypes.Append( (DblList::type)PI_NEW(PIString(szBuf)));
					    szBuf[i++] = '\t';
					    ::SendMessage( hWndPath, WM_GETTEXT, BUF_SIZE-i,
                            (LPARAM)&(szBuf[i]) );
					    lPaths.Append( (DblList::type)PI_NEW( PIString(
                            &(szBuf[i]) ) ) );
	
					    /*
					    ** Delete the item from the list and internal lists
					    */
					    ::SendMessage( hWndList, LB_DELETESTRING, iCurSel, 0 );
					    PI_DELETE( (PIString *)lTypes.Detach( iCurSel) );
					    PI_DELETE( (PIString *)lPaths.Detach( iCurSel) );
					
					    /*
					    ** Add new item to listbox
					    */
					    ::SendMessage( hWndList, LB_ADDSTRING,0,(LPARAM)szBuf);

					    /*
					    ** Clear edit fields, focus to list
					    */
					    ::SendMessage( hWndType, CB_SETCURSEL, (WPARAM)-1, 0 );
					    ::SendMessage( hWndPath, WM_SETTEXT, 0, (LPARAM)"" );
					    ::SetFocus( hWndList );
					    Changed( hDlg );
					    };
					    break;

				    case IDC_PATH:
					    if ( GET_WM_COMMAND_CMD(wParam, lParam)==EN_CHANGE )
						    { Internal_updateButtons( wParam, lParam ); };
					    break;

				    case IDC_DELETE:
					    {
					    int iCurSel = ::SendMessage( hWndList,LB_GETCURSEL,0,0);
					    ::SendMessage(hWndList,LB_DELETESTRING,iCurSel,0 );
					    PI_DELETE( (PIString *)lTypes.Detach( iCurSel) );
					    PI_DELETE( (PIString *)lPaths.Detach( iCurSel) );
			
					    /*
					    ** Clear edit fields, focus to list
					    */
					    ::SendMessage( hWndType, CB_SETCURSEL, (WPARAM)-1, 0 );
					    ::SendMessage( hWndPath, WM_SETTEXT, 0, (LPARAM)"" );
					    ::SetFocus( hWndList );
					    Changed( hDlg );
					    };
					    break;

				    case IDC_TYPE:
					    if ( GET_WM_COMMAND_CMD(wParam, lParam)==CBN_SELCHANGE )
						    { 
						    ::SendMessage( hWndList,LB_SETCURSEL,(WPARAM)-1,0);
						    Internal_updateButtons( wParam, lParam ); 
						    };
					    break;

                    case IDC_LIST:
					    if ( GET_WM_COMMAND_CMD(wParam, lParam)==LBN_SELCHANGE )
						    {
						    int iSel = SendMessage( hWndList,
                                LB_GETCURSEL, 0, 0 );
						    if ( iSel!=LB_ERR )
                                {
							    assert( iSel>=0 && iSel<lTypes.Size() );
							    ::EnableWindow( hWndDelete, TRUE );
							    int iIndex = ::SendMessage( hWndType,
                                    CB_FINDSTRINGEXACT, 
								    (WPARAM)-1,  (LPARAM)(const char *)
                                        *((PIString *)lTypes[iSel]) );
							    ::SendMessage( hWndType, CB_SETCURSEL, (WPARAM)
								    (iIndex==CB_ERR ? -1 : iIndex), 0 );
							    ::SendMessage( hWndPath, WM_SETTEXT, 0,
								    (LPARAM)(const char *)*
                                        ((PIString *)lPaths[iSel]) );
							    };
						    };
                        break;

					default:;		
					}
				break;

			case WM_NOTIFY:
				switch( ((NMHDR *)lParam)->code )
					{
					case PSN_QUERYCANCEL:
						return SHEET->Cancel( pInterface, hDlg );
	
					case PSN_KILLACTIVE:
                        return Internal_save( hDlg, szBuf, BUF_SIZE );
	
					default:
						return (FALSE);
                    };
				break;
	
			default:
                return (FALSE);
			};
	
		return (TRUE);
        };

};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static BOOL CALLBACK fnDialogProc(HWND hDlg, UINT uMsg, WPARAM wParam,
    LPARAM lParam)
{
    if ( uMsg==WM_INITDIALOG )
        {
        ::SetWindowLong( hDlg, DWL_USER, ((LPPROPSHEETPAGE)lParam)->lParam );
        };

    LogFilePage *pPage = (LogFilePage *)::GetWindowLong( hDlg, DWL_USER );
    return pPage ? pPage->DialogProc( hDlg, uMsg, wParam, lParam ) : FALSE;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int LogFilePage_constructor( PIObject *pObj,
	int iArgc, const char *ppArgv[] )
{
	return IFaceBase_constructor( pObj, PI_NEW( LogFilePage( pObj,
		iArgc, ppArgv ) ) );
}

#if 0
/*___+++CNF_BEGIN+++___*/
	<Class>
		Name LogFilePageClass
		Type LogicExtension
		Library IntrFace
		OnClassLoad IFaceBase_onClassLoad
		Constructor LogFilePage_constructor
		CopyConstructor IFaceBase_copyConstructor
		Destructor IFaceBase_destructor
		Execute IFaceBase_execute
	</Class>

	<Object>
		Name LogFilePage
		Class LogFilePageClass
	</Object>

/*___+++CNF_END+++___*/
#endif

