/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2001 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in
	the documentation and/or other materials provided with the
	distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
	mentioning features must contain the following
	acknowledgment:
	"This product includes software developed by John Roy
	and Holger Zimmermann (http://pi3web.sourceforge.net/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Intrface/Mappings.cpp,v $
 * $Date: 2002/08/02 06:58:56 $
 *
 Description:

\*____________________________________________________________________________*/
//$SourceTop:$

#include <iostream.h>
#include <stdio.h>
#include <windowsx.h>

#include "IFaceBse.h"
#include "PIStrStr.h"
#include "StrToken.h"

#include "QConfig.h"
#include "Mappings.hrc"

/*
** #define D { cerr << __FILE__ << ": " << __LINE__ << endl; }
*/
#define D

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#define C_RESOURCENAME                    "MappingsPage"

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
static const char *__pMappingTypes[]=
{ 
"Documents", 
"CGI",
"Windows CGI",
"Fast CGI",
"ISAPI",
/* "Redirect", */
0 
};
static const char *__pDefaultMappings[] =
{
"CGI\t/cgi-bin/\tCgi-Bin\\\t(none)",
"Windows CGI\t/cgi-win/\tCgi-Win\\\t(none)",
"Fast CGI\t/cgi-fcgi/\tCgi-Fcgi\\\t(none)",
"Documents\t/pidocs/\tPi3Docs\\\t(none)",
"Documents\t/icons/\tIcons\\\t(none)",
"Documents\t/images/\tImages\\\t(none)",
"Documents\t/errors/\tFragment\\\t(none)",
"ISAPI\t/isapi/\tIsapi\\\t(none)",
"Documents\t/admin/\tLogs\\\tAdministration",
"Documents\t/\tWebRoot\\\t(none)",

/* --- leave this last --- */
0
};

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#if 0
	/*
	** HTML documentation for this handler
	*/
/*___+++HTMLDOC_BEGIN+++___*/
Name:
    MappingsPage

Description:
    Creates the eponymous property page and is used by the PropSheet object.

Example:
	<PRE>
	&lt;Object&gt;
		Name MappingsPage
		Class MappingsPageClass
		PropertyPage MappingsPage
	&lt;/Object&gt;

	</PRE>

/*___+++HTMLDOC_END+++___*/
#endif

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
static BOOL CALLBACK fnDialogProc(HWND hDlg, UINT uMsg, WPARAM wParam,
    LPARAM lParam);

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class MappingsPage : public IFaceBase
{
private:
    QConfig *pConfig;
    PROPSHEETPAGE tPsp;
    Interface *pInterface;

    /*
    ** window private values
    */
    HWND hWndList;
    HWND hWndType;
    HWND hWndFrom;
    HWND hWndTo;
    HWND hWndRealm;
    HWND hWndAdd;
    HWND hWndReplace;
    HWND hWndDelete;
    DblList lTypes;
    DblList lFroms;
    DblList lTos;
    DblList lRealms;

	int Parameter( const char *pVariable, const char *pValue,
		const char *pWhere )
		{
		assert( pVariable && pValue );

		PIOStrStream os;

		os << pWhere << "MappingsPage: ";

        if ( 0 )
            {
            }
		else
			{
			os << "Unknown directive '" << pVariable <<
				"'" << ends;
			CONFIG_ERR( Object(), os.str() );
			return 0;
			};

		return 1;
		};

public:
	MappingsPage( PIObject *pObject, int iArgc, const char *ppArgv[] )
	:	IFaceBase( pObject ),
        pConfig( FindParentQConfig( PIObject_getDB( pObject ) ) ),
        hWndList( 0 ),
        hWndType( 0 ),
        hWndFrom( 0 ),
        hWndTo( 0 ),
        hWndRealm( 0 ),
        hWndAdd( 0 ),
        hWndReplace( 0 ),
        hWndDelete( 0 ),
        pInterface( 0 )
		{
        /*
        ** Read input parameters
        */
		ReadParameters( iArgc, ppArgv );

        if ( !IsOK() )
            { return; };

        /*
        ** Setup the PROPSHEETPAGE
        */
        memset( &tPsp, 0, sizeof( PROPSHEETPAGE ) );
        tPsp.dwSize = sizeof( PROPSHEETPAGE );
        tPsp.dwFlags = PSP_DEFAULT;
        tPsp.hInstance = (HINSTANCE)PIDLL_getSystemHandle( PIClass_getLibrary(
            PIObject_getClass( Object() ) ) );

        /* Status page */
        tPsp.pszTemplate = C_RESOURCENAME;
        tPsp.pfnDlgProc = (DLGPROC)fnDialogProc;
        tPsp.lParam = (LPARAM)this;

        /*
        ** Set the data DB of this object
        */
        PIDB *pDB = PIObject_getDB( Object() );
        PIDB_add( pDB, PIDBTYPE_OPAQUE, INTRFACE_WINDOWTYPE,
            (void *)INTRFACE_PROPPAGE, 0 );
        PIDB_add( pDB, PIDBTYPE_OPAQUE, INTRFACE_PROPSHEETPAGE,
            (void *)&tPsp, 0 );
        
        /*
        ** Done
        */
		};

	~MappingsPage()
		{
		};

    virtual int Handle( Interface &tWindow )
        {
        switch( tWindow.iMessage )
            {
            case INTRFACE_MSG_INIT:
                pInterface = &tWindow;
                break;

            case INTRFACE_MSG_WIN32CREATE:
                return Win32Create( tWindow );

            case INTRFACE_MSG_LOAD:
                return PIAPI_ABORT;

            case INTRFACE_MSG_DEFAULTS:
                {
                for( int i=0; __pMappingTypes[i]; i++ )
                    {
                    M_AddValue( pConfig, "Internal", "MappingTypes",
                        __pMappingTypes[i] );
                    };
                };
                {
                for( int i=0; __pDefaultMappings[i]; i++ )
                    {
                    M_AddValue( pConfig, "Mappings", "Mapping",
                        __pDefaultMappings[i] );
                    };
                };
                break;

            default:
                assert( 0 );
                return PIAPI_ABORT;
            };

        return PIAPI_COMPLETED;
        };

    int Internal_verifyFields( HWND hDlg, char *pszBuf, int iBufSize )
        {
		assert( pszBuf && iBufSize );
		*pszBuf = '\0';
		int i;
		int iFromLen;
		int iToLen;
	
		SendMessage( hWndFrom, WM_GETTEXT, iBufSize, (LPARAM)pszBuf );
	
		/*
		** 'From' must begin and end with '/'
		*/
		assert( *pszBuf );
		iFromLen = strlen(pszBuf);
		if ( *pszBuf!='/' || pszBuf[iFromLen-1]!='/' )
			{
			MessageBox( hDlg, "'From' field must begin and end with a \
forward-slash ('/').",
				"Error in 'From' Field", MB_OK | MB_ICONEXCLAMATION );
	
			::SetFocus( hWndFrom );
			::SendMessage( hWndFrom, EM_SETSEL, (WPARAM)(INT)0,
                (LPARAM)(INT)-1 );
			goto verify_fail;
			};
	
		/*
		** Scan for spaces
		*/
		for( i=0; i<iFromLen; i++)
			{
			if ( isspace( pszBuf[i] ) )
				{
				MessageBox( hDlg, "'From' field may not \
contain spaces or tabs.",
					"Error in 'From' Field", MB_OK | MB_ICONEXCLAMATION );
	
				goto verify_fail;
				};
			};
	
		SendMessage( hWndTo, WM_GETTEXT, iBufSize, (LPARAM)pszBuf );
	
		/*
		** 'To' must end with '/'
		*/
		assert( *pszBuf );
		iToLen = strlen(pszBuf);
		if ( pszBuf[iToLen-1]!='/' && pszBuf[iToLen-1]!='\\' )
			{
			MessageBox( hDlg, "'To' field must end with a forward-slash ('/') \
or a backslash ('\\').",
				"Error in 'To' Field", MB_OK | MB_ICONEXCLAMATION );
	
			::SetFocus( hWndTo );
			::SendMessage( hWndTo, EM_SETSEL, (WPARAM)(INT)0, (LPARAM)(INT)-1 );
	
			goto verify_fail;
			};
	
		/*
		** Some realm must be selected, even if it is '(none)'
		*/
		if ( ::SendMessage( hWndRealm, CB_GETCURSEL, 0, 0 )==CB_ERR )
			{
			::SendMessage( hWndRealm, CB_SETCURSEL, (WPARAM)::SendMessage( 
				hWndRealm,
				CB_SETCURSEL,
				::SendMessage( hWndRealm, CB_FINDSTRINGEXACT, 0,
                (LPARAM)"(none)" ), 0 ), 0 );
			assert( ::SendMessage( hWndRealm, CB_GETCURSEL, 0, 0 )!=CB_ERR  );
			};
	
		/*
		** Good
		*/
		return 1;
	
verify_fail:
	
		return 0;
	    };
	
    void Internal_saveValues( HWND hDlg )
        {
		enum { BUF_SIZE=1023 };
		char pszBuf[BUF_SIZE+1];
	
		for(;;)
			{
			int iCount = ::SendMessage( hWndList, LB_GETCOUNT, 0, 0 );
	
			if ( !iCount )
				{ return; };
	
			int iBiggest = -1;
			int iBiggestIndex = -1;
	
			for( int j=0; j<iCount; j++ )	
				{
				/*
				** Get string
				*/
				if (::SendMessage( hWndList, LB_GETTEXTLEN, j, 0 ) >= BUF_SIZE)
					{
					::MessageBox( hDlg, "Mapping line too long",
                        "Internal Error", MB_OK );
					return;
					};
				::SendMessage( hWndList, LB_GETTEXT, j, (LPARAM)pszBuf );
				
				/*
				** Get number of components in path
				*/
				int iNumComponents = 0;
				int i=0;
				for( ; pszBuf[i] && pszBuf[i]!='\t'; i++ );
				for( ; pszBuf[i] && isspace(pszBuf[i]); i++ );
				for( ; pszBuf[i] && !isspace(pszBuf[i]); i++ )
					{
					if ( pszBuf[i]=='/' )
						{
						iNumComponents++;
						i++; 
						if ( !pszBuf[i] || isspace(pszBuf[i]) )
							{ break; };
						};
					};
	
				if ( iNumComponents>iBiggest )
					{
					iBiggestIndex = j;
					iBiggest = iNumComponents;
					};
				};
	
			/*
			** Add biggest line to list and remove it from
			** listbox
			*/
			SendMessage( hWndList, LB_GETTEXT, iBiggestIndex, (LPARAM)pszBuf );
			SendMessage( hWndList, LB_DELETESTRING, iBiggestIndex, 0 );
			M_AddValue( pConfig, "Mappings", "Mapping", pszBuf );
			};
        };
		
	void Internal_addLine( const char *pLine )
	    {
		assert( pLine );
		assert( lTypes.Size()==lFroms.Size() );
		assert( lFroms.Size()==lTos.Size() );
	
		int i = 0;
		/* --- scan to tab --- */
		for( ; pLine[i] && pLine[i]!='\t'; i++ );
		lTypes.Append( (DblList::type)PI_NEW( PIString( pLine, i ) ) );
	
		/* --- scan over whitespace to start of from --- */
		for( ; pLine[i] && isspace(pLine[i]); i++ );
		pLine = &( pLine[i] );	
		i=0;
		/* --- scan to tab --- */
		for( ; pLine[i] && pLine[i]!='\t'; i++ );
		lFroms.Append( (DblList::type)PI_NEW( PIString( pLine, i ) ) );	
	
		/* --- scan over whitespace to start of to --- */
		for( ; pLine[i] && isspace(pLine[i]); i++ );
		pLine = &( pLine[i] );	
		i=0;
		/* --- scan to tab --- */
		for( ; pLine[i] && pLine[i]!='\t'; i++ );
		lTos.Append( (DblList::type)PI_NEW( PIString( pLine, i ) ) );	
	
		/* --- scan over whitespace to start of realm --- */
		for( ; pLine[i] && isspace(pLine[i]); i++ );
		pLine = &( pLine[i] );	
		i=0;
		/* --- scan to tab --- */
		for( ; pLine[i] && pLine[i]!='\t'; i++ );
		if ( i )
			{ lRealms.Append( (DblList::type)PI_NEW( PIString( pLine, i ) ) ); }
		else
			{ lRealms.Append( (DblList::type)PI_NEW( PIString( "(none)" ) ) ); }
	
		assert( lTypes.Size()==lFroms.Size() );
		assert( lFroms.Size()==lTos.Size() );
	    };
	
    void Internal_updateButtons( WPARAM wParam, LPARAM lParam, 
		const char *pFrom )
	    {
		int iCBSel = SendMessage( hWndType, CB_GETCURSEL, 0, 0 );
		const char *pType = 0;
		if ( iCBSel>-1 )
			{
			pType = M_LookupValueEx( pConfig, "Internal", 
				"MappingTypes", iCBSel );
			assert( pType );
			};
	
		/*
		** If the 'from' changed check is it an existing one.
		*/
		int iExistingIndex = -1;
		assert( pFrom );
		int iIndex = 0;
		if ( *pFrom )
			{
			for( DblListIterator i( lFroms ); !i.BadIndex(); i++, iIndex++ )
				{
				PIString *pString = (PIString *)i.Current();
				assert( pString );
				if ( (*pString)==pFrom )
					{
					iExistingIndex = iIndex;
					break;
					};
				};
			};
	
		/*
		** Remove listbox selection or scroll to correct line
		** as appropriate 
		*/
		SendMessage( hWndList, LB_SETCURSEL, iExistingIndex, 0 );
	
		/*
		** Set buttons
		*/
		if ( iCBSel<0 || SendMessage( hWndType, CB_GETLBTEXTLEN, iCBSel, 0 )
                ==0 ||
				SendMessage( hWndFrom, EM_LINELENGTH, 0, 0 )==0 ||
				SendMessage( hWndTo, EM_LINELENGTH, 0, 0 )==0 )
			{
			if ( IsWindowEnabled( hWndAdd ) )
                ::EnableWindow( hWndAdd, FALSE );
			if ( IsWindowEnabled( hWndReplace ) )
                ::EnableWindow( hWndReplace, FALSE );
			}
		else
			{
			if ( iExistingIndex>-1 )/* --- disable add, enable replace --- */
				{
				if ( IsWindowEnabled( hWndAdd ) )
                    ::EnableWindow( hWndAdd, FALSE );
				if ( !IsWindowEnabled( hWndReplace ) )
                    ::EnableWindow( hWndReplace, TRUE );
				}
			else					/* --- disable replace, enable add --- */
				{
				if ( !IsWindowEnabled( hWndAdd ) )
                    ::EnableWindow( hWndAdd, TRUE );
				if ( IsWindowEnabled( hWndReplace ) )
                    ::EnableWindow( hWndReplace, FALSE );
				};
			};
	
		/*
		** If there is a selected item in the listbox then
		** enable delete, otherwise disable delete	
		*/
		if ( SendMessage( hWndList, LB_GETCURSEL, 0, 0 )==LB_ERR )
			{ EnableWindow( hWndDelete, FALSE ); }
		else
			{ EnableWindow( hWndDelete, TRUE ); };
	    };
	
    int Win32Create( Interface &tWindow )
        {
        /* --- property sheet pages should not be created --- */
        return PIAPI_ABORT;
		};

    BOOL CALLBACK DialogProc(HWND hDlg, UINT uMsg, WPARAM wParam,
		LPARAM lParam)
	    {
		enum { BUF_SIZE=1023 };
		char szBuf[BUF_SIZE+1];
		int aTabstop[4];
	
		switch (uMsg)
			{
			case WM_INITDIALOG:
				HourglassOn();
				{
				const char *pTmp;
				int iTmp;
				hWndList = GetDlgItem( hDlg, IDC_LIST );
				hWndType = GetDlgItem( hDlg, IDC_TYPE );
				hWndFrom = GetDlgItem( hDlg, IDC_FROM );
				hWndTo = GetDlgItem( hDlg, IDC_TO );
				hWndRealm = GetDlgItem( hDlg, IDC_REALM );
				hWndAdd = GetDlgItem( hDlg, IDC_ADD );
				hWndReplace = GetDlgItem( hDlg, IDC_REPLACE );
				hWndDelete= GetDlgItem( hDlg, IDC_DELETE );
	
				assert( hWndList && hWndType && hWndFrom && hWndTo );
				assert( hWndAdd && hWndReplace && hWndDelete && hWndRealm );
	
				EnableWindow( hWndAdd, FALSE );
				EnableWindow( hWndReplace, FALSE );
				EnableWindow( hWndDelete, FALSE );
	
				/*
				** Add options to 'Types' combo-box
				*/
				iTmp = 0;
				for(;; iTmp++ )
					{
					pTmp = M_LookupValueEx( pConfig, "Internal", 
						"MappingTypes", iTmp );
					if ( !pTmp )
						{ break; };
	
					::SendMessage( hWndType, CB_ADDSTRING, 0, 
						(LPARAM)pTmp ); 
					};
	
				/*
				** Limit edit field lengths, really just a sanity thing
				*/
				SendMessage( hWndFrom, EM_SETLIMITTEXT, 256, 0 );
				SendMessage( hWndTo, EM_SETLIMITTEXT, 256, 0 );
	
				/*
				** Setup tabstops
				*/
				aTabstop[0] = 0;
				aTabstop[1] = 53;
				aTabstop[2] = 107;
				aTabstop[3] = 161;
				SendMessage( hWndList, LB_SETTABSTOPS, 4, (LPARAM)&aTabstop );
	
				/*
				** Load all mappings
				*/
				iTmp = 0;
				for(;; iTmp++ )
					{
					pTmp = M_LookupValueEx( pConfig, "Mappings", 
						"Mapping", iTmp );
					if ( !pTmp )
						{ break; };
	
					if ( ::SendMessage( hWndList, LB_ADDSTRING, 0,
                        (LPARAM)pTmp )>=0 )
						{
						Internal_addLine( pTmp );
						}
					else
						{ break; /* error */; };
					};
				}
	
				HourglassOff();
				break;
	
			case WM_DESTROY:
				{
				/*
				** Delete values
				*/
				DblListIterator x( lTypes );
				DblListIterator y( lFroms );
				DblListIterator z( lTos );
				DblListIterator w( lRealms );
				for( ; !x.BadIndex(); x++, y++, z++, w++ )
					{
					assert( !x.BadIndex() && !y.BadIndex() && !z.BadIndex() 
						&& !w.BadIndex() );
					PI_DELETE( (PIString *)x.Current() );
					PI_DELETE( (PIString *)y.Current() );
					PI_DELETE( (PIString *)z.Current() );
					PI_DELETE( (PIString *)w.Current() );
					};
				lTypes.Clear();
				lFroms.Clear();
				lTos.Clear();
				lRealms.Clear();
				};
				break;
	
			case WM_COMMAND:
				switch (GET_WM_COMMAND_ID(wParam, lParam))
					{
					case IDC_ADD:
						{
						/*
						** Verify fields are valid
						*/
						if ( !Internal_verifyFields( hDlg, szBuf, BUF_SIZE ) )
							{ break; };
	
						int i, j;
						*szBuf = '\0';
	
						SendMessage( hWndType, WM_GETTEXT, BUF_SIZE,
                            (LPARAM)szBuf );
						for( i=0; i<BUF_SIZE && szBuf[i]!='\t' && szBuf[i];
                            i++ );
						if ( szBuf[i] ) { szBuf[i] = '\0'; };
						lTypes.Append( (DblList::type)PI_NEW(
                            PIString( szBuf ) ) );
						szBuf[i++] = '\t';
	
						SendMessage( hWndFrom, WM_GETTEXT, BUF_SIZE-i,
                            (LPARAM)&(szBuf[i]) );
						for( j=i; i<BUF_SIZE && szBuf[i] && szBuf[i]!='\t';
                            i++ );
						if ( szBuf[i] ) { szBuf[i] = '\0'; };
						lFroms.Append( (DblList::type)PI_NEW( PIString(
                            &(szBuf[j]) ) ) );
						szBuf[i++] = '\t';
	
						SendMessage( hWndTo, WM_GETTEXT, BUF_SIZE-i,
                            (LPARAM)&(szBuf[i]) );
						for( j=i; i<BUF_SIZE && szBuf[i] && szBuf[i]!='\t';
                            i++ );
						if ( szBuf[i] ) { szBuf[i] = '\0'; };
						lTos.Append( (DblList::type)PI_NEW( PIString(
                            &(szBuf[j]) ) ) );
						szBuf[i++] = '\t';
	
						szBuf[i] = '\0';
						SendMessage( hWndRealm, WM_GETTEXT, BUF_SIZE-i,
                            (LPARAM)&(szBuf[i]) );
						if ( szBuf[i]=='\0' )
							{ strncpy( &(szBuf[i]), "(none)", BUF_SIZE-i ); };
						lRealms.Append( (DblList::type)PI_NEW( PIString(
                            &(szBuf[i]) ) ) );
		
						SendMessage( hWndList, LB_ADDSTRING, 0, (LPARAM)szBuf );
	
						/*
						** Clear edit fields, focus to list
						*/
						SendMessage( hWndType, CB_SETCURSEL, (WPARAM)-1, 0 );
						SendMessage( hWndFrom, WM_SETTEXT, 0, (LPARAM)"" );
						SendMessage( hWndTo, WM_SETTEXT, 0, (LPARAM)"" );
						SendMessage( hWndRealm, CB_SETCURSEL, (WPARAM)-1, 0 );
						SetFocus( hWndList );
						Changed( hDlg );
						};
						break;
		
					case IDC_REPLACE:
						{
						/*
						** Verify fields are valid
						*/
						if ( !Internal_verifyFields( hDlg, szBuf, BUF_SIZE ) )
							{ break; };
	
						/*
						** Get selected item index 
						*/
						int iCurSel = SendMessage( hWndList, LB_GETCURSEL, 0,
                            0 );
						assert( iCurSel>=0 );
	
						/*
						** Get text to add and add it to internal lists only
						*/
						int i, j;
						*szBuf = '\0';
	
						SendMessage( hWndType, WM_GETTEXT, BUF_SIZE,
                            (LPARAM)szBuf );
						for( i=0; i<BUF_SIZE && szBuf[i]!='\t' && szBuf[i];
                            i++ );
						if ( szBuf[i] ) { szBuf[i] = '\0'; };
						lTypes.Append( (DblList::type)PI_NEW(
                            PIString( szBuf ) ) );
						szBuf[i++] = '\t';
	
						SendMessage( hWndFrom, WM_GETTEXT, BUF_SIZE-i,
                            (LPARAM)&(szBuf[i]) );
						for( j=i; i<BUF_SIZE && szBuf[i] && szBuf[i]!='\t';
                            i++ );
						if ( szBuf[i] ) { szBuf[i] = '\0'; };
						lFroms.Append( (DblList::type)PI_NEW(
	                        PIString( &(szBuf[j]) ) ) );
						szBuf[i++] = '\t';
	
						SendMessage( hWndTo, WM_GETTEXT, BUF_SIZE-i,
                            (LPARAM)&(szBuf[i]) );
						for( j=i; i<BUF_SIZE && szBuf[i] && szBuf[i]!='\t';
                            i++ );
						if ( szBuf[i] ) { szBuf[i] = '\0'; };
						lTos.Append( (DblList::type)PI_NEW( PIString(
                            &(szBuf[j]) ) ) );
						szBuf[i++] = '\t';
	
						szBuf[i] = '\0';
						SendMessage( hWndRealm, WM_GETTEXT, BUF_SIZE-i,
                            (LPARAM)&(szBuf[i]) );
						if ( szBuf[i]=='\0' )
							{ strncpy( &(szBuf[i]), "(none)", BUF_SIZE-i ); };
						lRealms.Append( (DblList::type)PI_NEW( PIString(
                            &(szBuf[i]) ) ) );
		
						/*
						** Delete the item from the list and internal lists
						*/
						SendMessage( hWndList, LB_DELETESTRING,	iCurSel, 0 );
						PI_DELETE( (PIString *)lTypes.Detach( iCurSel) );
						PI_DELETE( (PIString *)lFroms.Detach( iCurSel) );
						PI_DELETE( (PIString *)lTos.Detach( iCurSel) );
						PI_DELETE( (PIString *)lRealms.Detach( iCurSel) );
						
						/*
						** Add new item to listbox
						*/
						SendMessage( hWndList, LB_ADDSTRING, 0, (LPARAM)szBuf );
	
						/*
						** Clear edit fields, focus to list
						*/
						SendMessage( hWndType, CB_SETCURSEL, (WPARAM)-1, 0 );
						SendMessage( hWndFrom, WM_SETTEXT, 0, (LPARAM)"" );
						SendMessage( hWndTo, WM_SETTEXT, 0, (LPARAM)"" );
						SendMessage( hWndRealm, CB_SETCURSEL, (WPARAM)-1, 0 );
						SetFocus( hWndList );
						Changed( hDlg );
						};
						break;
	
					case IDC_DELETE:
						{
						int iCurSel = SendMessage( hWndList, LB_GETCURSEL,
                            0, 0 );
						SendMessage( hWndList, LB_DELETESTRING,	iCurSel, 0 );
						PI_DELETE( (PIString *)lTypes.Detach( iCurSel ) );
						PI_DELETE( (PIString *)lFroms.Detach( iCurSel ) );
						PI_DELETE( (PIString *)lTos.Detach( iCurSel ) );
						PI_DELETE( (PIString *)lRealms.Detach( iCurSel ) );
				
						/*
						** Clear edit fields, focus to list
						*/
						SendMessage( hWndType, CB_SETCURSEL, (WPARAM)-1, 0 );
						SendMessage( hWndFrom, WM_SETTEXT, 0, (LPARAM)"" );
						SendMessage( hWndTo, WM_SETTEXT, 0, (LPARAM)"" );
						SendMessage( hWndRealm, CB_SETCURSEL, (WPARAM)-1, 0 );
						SetFocus( hWndList );
						Changed( hDlg );
						};
						break;
	
					case IDC_TYPE:
						if ( GET_WM_COMMAND_CMD(wParam, lParam)!=CBN_SELCHANGE )
							{ break; };
						*szBuf = '\0';
						SendMessage( hWndFrom, WM_GETTEXT, BUF_SIZE,
                            (LPARAM)szBuf ); 
						Internal_updateButtons( wParam, lParam, szBuf );
						break;
	
					case IDC_FROM:
						if ( GET_WM_COMMAND_CMD(wParam, lParam)!=EN_CHANGE )
							{ break; };
						*szBuf = '\0';
						SendMessage( hWndFrom, WM_GETTEXT, BUF_SIZE,
                            (LPARAM)szBuf ); 
						Internal_updateButtons( wParam, lParam, szBuf );
						break;
	
					case IDC_TO:
						if ( GET_WM_COMMAND_CMD(wParam, lParam)!=EN_CHANGE )
							{ break; };
						*szBuf = '\0';
						SendMessage( hWndFrom, WM_GETTEXT, BUF_SIZE,
                            (LPARAM)szBuf );
						Internal_updateButtons( wParam, lParam, szBuf );
						break;
	
					case IDC_REALM:
						if ( GET_WM_COMMAND_CMD(wParam, lParam)!=CBN_SELCHANGE )
							{ break; };
						*szBuf = '\0';
						SendMessage( hWndFrom, WM_GETTEXT, BUF_SIZE,
                            (LPARAM)szBuf ); 
						Internal_updateButtons( wParam, lParam, szBuf );
						break;
	
					case IDC_LIST:
						if ( GET_WM_COMMAND_CMD(wParam, lParam)==LBN_SELCHANGE )
							{
							int iSel = SendMessage( hWndList, LB_GETCURSEL,
                                0, 0 );
							if ( iSel!=LB_ERR )
								{
								assert( iSel>=0 && iSel<lFroms.Size() );
								EnableWindow( hWndDelete, TRUE );
								int iIndex = SendMessage( hWndType,
                                    CB_FINDSTRINGEXACT, 
									(WPARAM)-1, 
									(LPARAM)(const char *)*((PIString *)
                                        lTypes[iSel]) );
								assert( iIndex>=0 );
								SendMessage( hWndType, CB_SETCURSEL, (WPARAM)
									(iIndex==CB_ERR ? -1 : iIndex), 0 );
								SendMessage( hWndFrom, WM_SETTEXT, 0,
									(LPARAM)(const char *)*((PIString *)
                                        lFroms[iSel]) );
								SendMessage( hWndTo, WM_SETTEXT, 0,
									(LPARAM)(const char *)*((PIString *)
                                        lTos[iSel]) );
								iIndex = SendMessage( hWndRealm,
                                    CB_FINDSTRINGEXACT, 
									    (WPARAM)-1, 
									(LPARAM)(const char *)*((PIString *)
                                        lRealms[iSel]) );
								assert( iIndex>=0 );
								SendMessage( hWndRealm, CB_SETCURSEL, (WPARAM)
									(iIndex==CB_ERR ? -1 : iIndex), -1 );
								};
							};
						break;
	
					default:;		
					}
				break;
	
			case WM_NOTIFY:
				{
				switch( ((NMHDR *)lParam)->code )
					{
                    case PSN_SETACTIVE:
                        {
                        int iCurSel = ::SendMessage( hWndRealm, CB_GETCURSEL,
                            0, 0 );

                        ::SendMessage( hWndRealm, CB_RESETCONTENT, 0, 0 );
                        int iTmp = 0;
                        for(;; iTmp++ )
                            {
                            const char *pTmp = M_LookupValueEx( pConfig,
                                "Realms", "Basic", iTmp );
                            if ( !pTmp ) { break; };

                            ::SendMessage( hWndRealm, CB_ADDSTRING, 0,
                                (LPARAM)pTmp );
                            };

                        for(iTmp = 0;; iTmp++ )
                            {
                            const char *pTmp = M_LookupValueEx( pConfig,
                                "Realms", "Digest", iTmp );
                            if ( !pTmp ) { break; };

                            ::SendMessage( hWndRealm, CB_ADDSTRING, 0,
                                (LPARAM)pTmp );
                            };

                        for(iTmp = 0;; iTmp++ )
                            {
                            const char *pTmp = M_LookupValueEx( pConfig,
                                "Realms", "Ntlm", iTmp );
                            if ( !pTmp ) { break; };

                            ::SendMessage( hWndRealm, CB_ADDSTRING, 0,
                                (LPARAM)pTmp );
                            };
                        ::SendMessage( hWndRealm, CB_ADDSTRING, 0,
                            (LPARAM)"(none)" );

                        ::SendMessage( hWndRealm, CB_SETCURSEL, iCurSel, 0 );
                        };
                        break;
	
					case PSN_QUERYCANCEL:
						return SHEET->Cancel(pInterface, hDlg);
	
					case PSN_KILLACTIVE:
						{
						int i = 0;
						int iCurSel = ::SendMessage( hWndList, LB_GETCURSEL,
                            0, 0 );
						int iCount = ::SendMessage( hWndList, LB_GETCOUNT,
                            0, 0 );
	
						/* --- remove old ones first --- */
						M_DeleteValue( pConfig, "Mappings", "Mapping" );
	
						/*
						** Save all mappings
						*/
						Internal_saveValues( hDlg );
	
						/*
						** Delete values
						*/
						DblListIterator x( lTypes );
						DblListIterator y( lFroms );
						DblListIterator z( lTos );
						DblListIterator w( lRealms );
						for( ; !x.BadIndex(); x++, y++, z++, w++ )
							{
							assert( !x.BadIndex() && !y.BadIndex() &&
                                !z.BadIndex() && !w.BadIndex() );
							PI_DELETE( (PIString *)x.Current() );
							PI_DELETE( (PIString *)y.Current() );
							PI_DELETE( (PIString *)z.Current() );
							PI_DELETE( (PIString *)w.Current() );
							};
						lTypes.Clear();
						lFroms.Clear();
						lTos.Clear();
						lRealms.Clear();
	
						/*
						** Load all mappings
						*/
						int iTmp = 0;
						const char *pTmp;
						for(;; iTmp++ )
							{
							pTmp = M_LookupValueEx( pConfig, "Mappings", 
								"Mapping", iTmp );
							if ( !pTmp )
								{ break; };
	
							if ( SendMessage( hWndList, LB_ADDSTRING, 0,
								(LPARAM)pTmp )>=0 )
								{
								Internal_addLine( pTmp );
								}
							else
								{ break; /* error */; };
							};
						::SendMessage( hWndList, LB_SETCURSEL, iCurSel, 0 );
						};
						SetWindowLong( hDlg, DWL_MSGRESULT, FALSE );
						break;
	
					default:
						return (FALSE);
					};
				};
				break;
	
			default:
				return (FALSE);
			};
	
		return (TRUE);
        };
};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static BOOL CALLBACK fnDialogProc(HWND hDlg, UINT uMsg, WPARAM wParam,
    LPARAM lParam)
{
    if ( uMsg==WM_INITDIALOG )
        {
        ::SetWindowLong( hDlg, DWL_USER, ((LPPROPSHEETPAGE)lParam)->lParam );
        };

    MappingsPage *pPage = (MappingsPage *)::GetWindowLong( hDlg, DWL_USER );
    return pPage ? pPage->DialogProc( hDlg, uMsg, wParam, lParam ) : FALSE;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int MappingsPage_constructor( PIObject *pObj,
	int iArgc, const char *ppArgv[] )
{
	return IFaceBase_constructor( pObj, PI_NEW( MappingsPage( pObj,
		iArgc, ppArgv ) ) );
}

#if 0
/*___+++CNF_BEGIN+++___*/
	<Class>
		Name MappingsPageClass
		Type LogicExtension
		Library IntrFace
		OnClassLoad IFaceBase_onClassLoad
		Constructor MappingsPage_constructor
		CopyConstructor IFaceBase_copyConstructor
		Destructor IFaceBase_destructor
		Execute IFaceBase_execute
	</Class>

	<Object>
		Name MappingsPage
		Class MappingsPageClass
	</Object>

/*___+++CNF_END+++___*/
#endif

