/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2001 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    and Holger Zimmermann (http://pi3web.sourceforge.net/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Pi3API/HTTPUtil.cpp,v $
 * $Date: 2002/04/14 06:57:23 $
 *
 Description:
	HTTP utilities which require no content other than thier parameters.

\*____________________________________________________________________________*/
//$SourceTop:$

#include <assert.h>
#include <string.h>
#include <time.h>
#include <ctype.h>

#include <iostream.h>
#include <fstream.h>
#include <stdio.h>

#include "PIString.h"
#include "HTTPDefs.h"
#include "HTTPCore.h"
#include "HTTPUtil.h"
#include "PIStrStr.h"

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
/*
#define D { cerr << __FILE__ << ", " << __LINE__ << ": " << endl; }
*/
#define D

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Returns 0 on success.
\*____________________________________________________________________________*/
int HTTPUtil_rFC822Time( struct tm *pT, Pi3String *pResult )
{
	assert( pT );
	assert( pResult );
	if ( !pT || !pResult )
		{ return -1; };

	/* --- RFC822 is strftime "%a, %d %b %Y %T GMT" --- */
	enum { SMALL_BUFFER=255 };
	char szBuf[SMALL_BUFFER+1];
	szBuf[SMALL_BUFFER] = '\0';
	int iRet = strftime(szBuf, SMALL_BUFFER, "%a, %d %b %Y %H:%M:%S GMT", pT);
	*pResult = szBuf;
	return iRet>0 ? 0 : -1;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Contract a path containing './' and '../' sequences
	Change any '\' characters to '/'. Returns the number of
	directories above this one that the path should be contracted.
\*____________________________________________________________________________*/
int HTTPUtil_contractPath( char *pPath )
{
	int iNumDots=0;
	int i;
	for( i=0; pPath[i]; i++ )
		{
		if ( pPath[i]=='.' && iNumDots>=0 )
			{ iNumDots++; }
		else
			{
			if ( pPath[i]=='\\' )
				{ pPath[i]='/'; break; }
			else if ( pPath[i]=='/' )
				{ break; };
			iNumDots=-1;		/* --- file or dir. starting with '.' --- */
			};
		};

	/* ---
	iNumDots is the number of directories to ignore, starting with
	this one.
	--- */
	if ( iNumDots==-1 ) { iNumDots=0; };
	if ( pPath[i] )
		{ iNumDots+=HTTPUtil_contractPath( &(pPath[i+1]) ); };
	if ( iNumDots>0 )
		{
		/* --- skip this directory --- */
/*
** modified this 3/31/97 JR, to handle 'xxx/..' like 'xxx/../'
**		memmove( pPath, &(pPath[i+1]), strlen( &(pPath[i]) ) );
*/
		memmove( pPath, &(pPath[i]), strlen( &(pPath[i]) )+1 );
		iNumDots--;
		};
	return iNumDots;
}

#define RCNAME_UNKNOWN			"UNKNOWN"
#define RCNAME_COMPLETED		"COMPLETED"
#define RCNAME_CONTINUE			"CONTINUE"
#define RCNAME_ERROR			"ERROR"
#define RCNAME_ABORT			"ABORT"
#define RCNAME_REDIRECT			"REDIRECT"

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
const char *HTTPUtil_rcNumberToName( int iRC )
{
	switch( iRC )
		{
		case PIAPI_COMPLETED:		return RCNAME_COMPLETED;
		case PIAPI_CONTINUE:		return RCNAME_CONTINUE;
		case PIAPI_ERROR:			return RCNAME_ERROR;
		case PIAPI_ABORT:			return RCNAME_ABORT;
		case INT_REDIRECT:			return RCNAME_REDIRECT;
		default:
			return RCNAME_UNKNOWN;
		};
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int HTTPUtil_rcNameToNumber( const char *pName )
{
	if ( !pName ) { return PIAPI_ABORT; };
	if ( !PIUtil_stricmp( RCNAME_ABORT, pName ) )
		{ return PIAPI_ABORT; }
	else if ( !PIUtil_stricmp( RCNAME_CONTINUE, pName ) )
		{ return PIAPI_CONTINUE; }
	else if ( !PIUtil_stricmp( RCNAME_COMPLETED, pName ) )
		{ return PIAPI_COMPLETED; }
	else if ( !PIUtil_stricmp( RCNAME_ERROR, pName ) )
		{ return PIAPI_ERROR; }
	else if ( !PIUtil_stricmp( RCNAME_REDIRECT, pName ) )
		{ return INT_REDIRECT; }
	else
		{ return PIAPI_ABORT; };
}

#define PHNAME_INVALID			"INVALID"
#define PHNAME_INIT				"INIT"
#define PHNAME_HEADERS			"HEADERS"
#define PHNAME_HOSTMAP			"HOSTMAP"
#define PHNAME_MAPPING			"MAPPING"
#define PHNAME_CHECKPATH		"CHECKPATH"
#define PHNAME_CHECKAUTH		"CHECKAUTH"
#define PHNAME_CHECKACCESS		"CHECKACCESS"
#define PHNAME_CHECKTYPE		"CHECKTYPE"
#define	PHNAME_HANDLE			"HANDLE"
#define PHNAME_LOG				"LOG"
#define PHNAME_DESTROY			"DESTROY"

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
const char *HTTPUtil_phaseNumberToName( int iPhaseNumber )
{
	switch( iPhaseNumber )
		{
		case PH_INVALID:		return PHNAME_INVALID;
		case PH_INIT:			return PHNAME_INIT;
		case PH_HEADERS:		return PHNAME_HEADERS;
		case PH_HOSTMAP:		return PHNAME_HOSTMAP;
		case PH_MAPPING:		return PHNAME_MAPPING;
		case PH_CHECKPATH:		return PHNAME_CHECKPATH;
		case PH_CHECKAUTH:		return PHNAME_CHECKAUTH;
		case PH_CHECKACCESS:	return PHNAME_CHECKACCESS;
		case PH_CHECKTYPE:		return PHNAME_CHECKTYPE;
		case PH_HANDLE:			return PHNAME_HANDLE;
		case PH_LOG:			return PHNAME_LOG;
		case PH_DESTROY:		return PHNAME_DESTROY;
		default:
			return PHNAME_INVALID;
		};
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int HTTPUtil_phaseNameToNumber( const char *pName )
{
	if ( !pName ) { return PH_INVALID; };
	if ( !PIUtil_stricmp( PHNAME_INVALID, pName ) )
		{ return PH_INVALID; }
	else if ( !PIUtil_stricmp( PHNAME_INIT, pName ) )
		{ return PH_INIT; }
	else if ( !PIUtil_stricmp( PHNAME_HEADERS, pName ) )
		{ return PH_HEADERS; }
	else if ( !PIUtil_stricmp( PHNAME_HOSTMAP, pName ) )
		{ return PH_HOSTMAP; }
	else if ( !PIUtil_stricmp( PHNAME_MAPPING, pName ) )
		{ return PH_MAPPING; }
	else if ( !PIUtil_stricmp( PHNAME_CHECKPATH, pName ) )
		{ return PH_CHECKPATH; }
	else if ( !PIUtil_stricmp( PHNAME_CHECKAUTH, pName ) )
		{ return PH_CHECKAUTH; }
	else if ( !PIUtil_stricmp( PHNAME_CHECKACCESS, pName ) )
		{ return PH_CHECKACCESS; }
	else if ( !PIUtil_stricmp( PHNAME_CHECKTYPE, pName ) )
		{ return PH_CHECKTYPE; }
	else if ( !PIUtil_stricmp( PHNAME_HANDLE, pName ) )
		{ return PH_HANDLE; }
	else if ( !PIUtil_stricmp( PHNAME_LOG, pName ) )
		{ return PH_LOG; }
	else if ( !PIUtil_stricmp( PHNAME_DESTROY, pName ) )
		{ return PH_DESTROY; }
	else
		{ return PH_INVALID; };	
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
const char *HTTPUtil_getHostName( PIHTTP *pPIHTTP )
{
	return (const char *)PIDB_lookup( pPIHTTP->GetHostDB(),
		PIDBTYPE_RFC822, KEY_INT_HOSTNAME, 0 );
}

static const char *http_default_port = "80";
static const char *https_default_port = "443";

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
const char *HTTPUtil_getHostPort( PIHTTP *pPIHTTP )
{
	const char *pValue = (const char *)PIDB_lookup( pPIHTTP->GetHostDB(),
		PIDBTYPE_RFC822, KEY_INT_SERVERPORT, 0 );
	
	if (pValue) return pValue;

	pValue = (const char *)PIDB_lookup( pPIHTTP->GetConnectionDB(),
		PIDBTYPE_STRING, KEY_HTTPS_KEYSIZE, 0 );
	return pValue ? https_default_port : http_default_port;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
const char *HTTPUtil_protocolNumberToName( int iProtocol )
{
	switch( iProtocol )
		{
		case PR_HTTP09:
			return PR_NAME_HTTP09;

		case PR_HTTP10:
			return PR_NAME_HTTP10;

		case PR_HTTP11:
			return PR_NAME_HTTP11;

		default:;
			return PR_NAME_UNKNOWN;
		};
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	State machine based resolution of weekday name to number
	First letter must be upper case, rest lower, no error checking.
	0-6 on success
	-1 on error
\*____________________________________________________________________________*/
int Internal_weekdayNumber( const char *pWeekday )
{
	assert( pWeekday );
	int iState = 0;
	for( int i=0; pWeekday[i]; i++ )
		{
		char c = pWeekday[i];
		switch( iState )
			{
			case 0:
				switch( c )
					{
					case 'S': iState=1; break;
					case 'M': return 1;	/* Monday */
					case 'T': iState=2; break;
					case 'W': return 3;	/* Wednesday */
					case 'F': return 5;	/* Friday */
					default:
						return -1;
					};
				break;

			case 1:
				return c=='u' ? 0 : 6; /* Sunday, Saturday */

			case 2:
				return c=='u' ? 2 : 4; /* Tuesday, Thursday */

			default:
				return -1;
			};
		};
	return -1;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	State machine based resolution of month name to number
	First letter must be upper case, rest lower, no error checking.
	0-11 on success
	-1 on error
\*____________________________________________________________________________*/
int Internal_monthNumber( const char *pMonth )
{
	assert( pMonth );
	int iState = 0;
	for( int i=0; pMonth[i]; i++ )
		{
		char c = pMonth[i];
		switch( iState )
			{
			case 0:
				switch( c )
					{
					case 'J': iState=1;	break;
					case 'F': return 1;		/* February */
					case 'M': iState=2;	break;
					case 'A': iState=3; break;
					case 'S': return 8;		/* September */
					case 'O': return 9;		/* October */
					case 'N': return 10;	/* November */
					case 'D': return 11;	/* December */
					default:
						return -1;
					};
				break;

			case 1:
				/* January, June, July */
				if ( c=='a' ) { return 0; };
				iState = 4;
				break;

			case 2:
				/* March, May */
				iState = 5;
				break;

			case 3:
				/* April, August */
				if ( c=='p' ) { return 3; }			/* April */
				else { return 7; };					/* August */

			case 4:
				/* June, July */
				if ( c=='n' ) { return 5; } 		/* June */
				else { return 6; };					/* July */

			case 5:
				if ( c=='r' )	{ return 2; }		/* March */
				else { return 4; };					/* May */

			default:
				return -1;
			};
		};
	return -1;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int Internal_readRFC1123Time(
		int iWeekday,
		const char *pLine,
		struct tm &tT )
{
	if ( iWeekday<0 )
		{ return -1; };

#if 0
	int tm_sec;	 /* seconds after the minute - [0,59] */
	int tm_min;	 /* minutes after the hour - [0,59] */
	int tm_hour;	/* hours since midnight - [0,23] */
	int tm_mday;	/* day of the month - [1,31] */
	int tm_mon;	 /* months since January - [0,11] */
	int tm_year;	/* years since 1900 */
	int tm_wday;	/* days since Sunday - [0,6] */
	int tm_yday;	/* days since January 1 - [0,365] */
	int tm_isdst;   /* daylight savings time flag */
#endif

	/*
	 * RFC1123:
	 *	
	 * Sun, 06 Nov 1994 08:49:37 GMT	; RFC 822, updated by RFC 1123
	 *		|| | | |  | || || || | |
	 *		|| | | |  11111111112222
	 *		012345678901234567890123
	 *
	 */
	tT.tm_sec = atoi( &(pLine[18]) );
	tT.tm_min = atoi( &(pLine[15]) );
	tT.tm_hour = atoi( &(pLine[12]) );
	tT.tm_mday = atoi( &(pLine[0]) );
	tT.tm_mon = Internal_monthNumber( &(pLine[3]) );
	tT.tm_year = atoi( &(pLine[7]) )-1900;
	tT.tm_wday = atoi( &(pLine[iWeekday]) );
/*
**	tT.tm_yday = atoi( &(pLine[]) );
**	tT.tm_isdst = atoi( &(pLine[]) );
*/
	if ( strncmp( &(pLine[21]), "GMT", 3 ) )
		{ return -1; }; 

	return 0;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int Internal_readRFC850Time(
		int iWeekday,
		const char *pLine,
		struct tm &tT )
{
	if ( iWeekday<0 )
		{ return -1; };

#if 0
	int tm_sec;	 /* seconds after the minute - [0,59] */
	int tm_min;	 /* minutes after the hour - [0,59] */
	int tm_hour;	/* hours since midnight - [0,23] */
	int tm_mday;	/* day of the month - [1,31] */
	int tm_mon;	 /* months since January - [0,11] */
	int tm_year;	/* years since 1900 */
	int tm_wday;	/* days since Sunday - [0,6] */
	int tm_yday;	/* days since January 1 - [0,365] */
	int tm_isdst;   /* daylight savings time flag */
#endif

	/*
	 * RFC850:
	 *	
	 * Sunday, 06-Nov-94 08:49:37 GMT   ; RFC 850, obsoleted by RFC 1036
	 *		   || | | || || || || | |
	 *		   || | | || 111111111122
	 *		   0123456789012345678901
	 *
	 */
	tT.tm_sec = atoi( &(pLine[16]) );
	tT.tm_min = atoi( &(pLine[13]) );
	tT.tm_hour = atoi( &(pLine[10]) );
	tT.tm_mday = atoi( &(pLine[0]) );
	tT.tm_mon = Internal_monthNumber( &(pLine[3]) );
	tT.tm_year = atoi( &(pLine[7]) );
	tT.tm_wday = atoi( &(pLine[iWeekday]) );
/*
**	tT.tm_yday = atoi( &(pLine[]) );
**	tT.tm_isdst = atoi( &(pLine[]) );
*/
	if ( strncmp( &(pLine[19]), "GMT", 3 ) )
		{ return -1; }; 

	return 0;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int Internal_readANSICTime(
		int iWeekday,
		const char *pLine,
		struct tm &tT )
{
	if ( iWeekday<0 )
		{ return -1; };

#if 0
	int tm_sec;	 /* seconds after the minute - [0,59] */
	int tm_min;	 /* minutes after the hour - [0,59] */
	int tm_hour;	/* hours since midnight - [0,23] */
	int tm_mday;	/* day of the month - [1,31] */
	int tm_mon;	 /* months since January - [0,11] */
	int tm_year;	/* years since 1900 */
	int tm_wday;	/* days since Sunday - [0,6] */
	int tm_yday;	/* days since January 1 - [0,365] */
	int tm_isdst;   /* daylight savings time flag */
#endif

	/*
	 * ANSI C's asctime() format:
	 *	
	 * Sun Nov  6 08:49:37 1994		 ; ANSI C's asctime() format
	 *	   | | |  || || || |  |
	 *	   | | |  || 1111111111
	 *	 01234567890123456789
	 *
	 */
	tT.tm_sec = atoi( &(pLine[13]) );
	tT.tm_min = atoi( &(pLine[10]) );
	tT.tm_hour = atoi( &(pLine[7]) );
	tT.tm_mday = atoi( &(pLine[4]) );
	tT.tm_mon = Internal_monthNumber( &(pLine[0]) );
	tT.tm_year = atoi( &(pLine[16]) )-1900;
	tT.tm_wday = atoi( &(pLine[iWeekday]) );
/*
**	tT.tm_yday = atoi( &(pLine[]) );
**	tT.tm_isdst = atoi( &(pLine[]) );
*/

	return 0;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Return zero on success.
	This is not very tolerant right now.
	Tolerant of garbage or other date after the date in the same string.
\*____________________________________________________________________________*/
int HTTPUtil_readTime( const char *pLine, struct tm *pT )
{
	assert( pLine );
	assert( pT );

	if ( !pLine || !pT )
		{ return -1; };

	struct tm &tT = *pT;

	/* --- read line to first ' ' or ',' --- */
	int i=0;
	for( ; pLine[i] && pLine[i]!=' ' && pLine[i]!=','; i++);
	switch( pLine[i] )
		{
		case ' ':
			/* --- asctime() --- */
			if ( i!=3 ) { return -1; };
			return Internal_readANSICTime( 
				Internal_weekdayNumber( pLine ), &(pLine[4]), tT );

		case ',':
			/* --- asctime() --- */
			if ( i==3 )
				{ return Internal_readRFC1123Time( 
					Internal_weekdayNumber( pLine ), &(pLine[5]), tT ); }
			else
				{ return Internal_readRFC850Time( 
					Internal_weekdayNumber( pLine ), &(pLine[i+2]), tT ); };

		default:;
			return -1;
		};
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Returns, 0 match failed, 1 match succeeded.
\*____________________________________________________________________________*/
int Internal_regexMatch(
	const char *pRegex,
	int iRegexLen,
	const char *pString,
	int iStringLen )
{
	int i=0;
	int j=0;
	for( ; i<iRegexLen; i++, j++ )
		{
		if ( pRegex[i]=='*' )
			{
			i++;
			for(; i<iRegexLen && j<iStringLen; j++)
				{
				if ( Internal_regexMatch(
					&(pRegex[i]),
					iRegexLen-i,
					&(pString[j]),
					iStringLen-j ) )
					{ return 1; };
				};
			if ( i==iRegexLen )
				{ return 1; }
			else
				{ return 0; };
			};

		if ( i==iStringLen || pString[i]!=pRegex[i] )
			{ return 0; };
		};
	return iRegexLen==iStringLen ? 1 : 0;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Returns non-zero on success.
\*____________________________________________________________________________*/
int HTTPUtil_regexMatchPi3String( const Pi3String *pRegexPattern,
	const Pi3String *pString )
{
	assert( pRegexPattern );
	assert( pString );
	int iMatch = Internal_regexMatch( *pRegexPattern, pRegexPattern->Len(),
		*pString, pString->Len() );
	return iMatch;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Returns non-zero on success.
\*____________________________________________________________________________*/
int HTTPUtil_regexMatch( const char *pRegex, int iRegexLen,
	const char *pString, int iStringLen )
{
	assert( pRegex && pString );
	if ( !pRegex || !iRegexLen || !pString || !iStringLen )
		{
		if ( !iRegexLen && !iStringLen )
			{ return 1; };
		return 0;
		};
	return Internal_regexMatch( pRegex, iRegexLen, pString, iStringLen );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void HTTPUtil_urlEncode( const Pi3String *pToEncode, Pi3String *pResult )
{
	assert( pToEncode );
	assert( pResult );
	if ( !pToEncode || !pResult )
		{ return; };
	const PIString &sToEncode = *pToEncode;
	PIString &sResult = *pResult;

	PIOStrStream os;
	int iLen = sToEncode.Len();
	const char *pS = sToEncode;
	PIString sAscii;
	for( int i=0; i<iLen; i++)
		{
		char ch=pS[i];
		if ( ch==' ' )
			{ os << '+'; }
		else if ( !(isalnum(ch)) && (ch!='_') )
			{
			if ( pi_ltoa( ch, 16, sAscii ) )
				{ os << '%' << sAscii; }
			}
		else
			{ os << ch; };
		};

	os << ends;
	sResult = os.str();
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void HTTPUtil_urlDecode( const Pi3String *pToDecode, Pi3String *pResult )
{
	assert( pToDecode );
	assert( pResult );
	if ( !pToDecode || !pResult )
		{ return; };
	const PIString &sToDecode = *pToDecode;
	PIString &sResult = *pResult;

	int iLen = sToDecode.Len();
	char *pTmp = new char[iLen+1];
	strcpy( pTmp, sToDecode );	
	int i=0, j=0;
	for(; i<iLen; i++, j++)
		{
		if ( pTmp[i]=='+' )
			{ pTmp[j] = ' '; }
		else if ( pTmp[i]=='%' )
			{
			if ( ((i+1)<iLen) && pTmp[i+1]=='%' )
				{ pTmp[j]=pTmp[i]; i++; }
			else if ( (i+2)<iLen )
				{
				char szBuf[3];
				szBuf[2]='\0';
				strncpy( szBuf, &(pTmp[i+1]), 2 );
				pTmp[j] = (char)strtol( szBuf, (char **)0, 16 ); 
				i+=2;
				};
			}
		else
			{ pTmp[j]=pTmp[i]; };
		};
	pTmp[j] = '\0';
	sResult = pTmp;
	delete [] pTmp;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int HTTPUtil_doHTTPError( PIHTTP *pPIHTTP, int iErrorCode )
{
	assert( pPIHTTP );
	assert( iErrorCode );
	
	/* --- set the error status code --- */
	pPIHTTP->iStatus = iErrorCode;

	PIDB *pR = pPIHTTP->GetResponseDB();
	PIDB *pQ = pPIHTTP->GetRequestDB();

	/* ---
	clear any content-type, content-length, objectmap,
	authentication realm, authentication type.
	--- */
	if ( PIDB_lookup( pR, PIDBTYPE_RFC822, KEY_HTTP_CONTENTTYPE, 0 ) )
		PIDB_replace( pR, PIDBTYPE_RFC822, KEY_HTTP_CONTENTTYPE, 0, 0 );
	if ( PIDB_lookup( pR, PIDBTYPE_RFC822, KEY_HTTP_CONTENTLENGTH, 0 ) )
		PIDB_replace( pR, PIDBTYPE_RFC822, KEY_HTTP_CONTENTLENGTH, 0, 0 );
	if ( PIDB_lookup( pR, PIDBTYPE_STRING, KEY_INT_OBJECTMAP, 0 ) )
		PIDB_replace( pR, PIDBTYPE_STRING, KEY_INT_OBJECTMAP, 0, 0 );
	if ( PIDB_lookup( pR, PIDBTYPE_STRING, KEY_INT_AUTHENTICATIONREALM, 0 ) )
		PIDB_replace( pR, PIDBTYPE_STRING, KEY_INT_AUTHENTICATIONREALM, 0, 0 );
	if ( PIDB_lookup( pR, PIDBTYPE_STRING, KEY_INT_AUTHTYPE, 0 ) )
		PIDB_replace( pR, PIDBTYPE_STRING, KEY_INT_AUTHTYPE, 0, 0 );

	/* --- set the method to 'GET' --- */
	PIDB_replace( pQ, PIDBTYPE_OPAQUE, KEY_HTTP_METHOD, (void *)MD_GET, 0 );
	PIDB_replace( pQ, PIDBTYPE_STRING, KEY_HTTP_METHOD, (void *)MD_NAME_GET, 0);

	/* --- disable keep alive --- */
	PIHTTP_disableKeepOpen( *pPIHTTP );

	/* --- return INT_REDIRECT, to cause a redirect --- */
	return INT_REDIRECT;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int HTTPUtil_sendFile( PIIOBuffer *pBuffer, PIFInfo *pFInfo,
	int iBufferingFlags, int iSendFlags )
{
	(void)iSendFlags;		/* --- not currently used --- */
	assert( pBuffer );
	assert( pFInfo );

	if ( !pBuffer || !pFInfo )
		{
        return PIAPI_ERROR;
        };

	/* --- get path --- */
	const char *pPath = PIFInfo_getPath( pFInfo );

	/* --- get the file map --- */
	const char *pMap = 0;
	int iLen = 0;

	/* --- get filemap object --- */
	if ( PIFInfo_getSize( pFInfo )>0 )
		{
		PIFMap *pFileMap = PIFMap_new( pFInfo );
		if ( pFileMap )
			{
			pMap = (const char *)PIFMap_lock( pFileMap, &iLen );
			};
		if ( !pFileMap || !pMap )
			{	
			HTTPCore_logError( 0, "HTTPUtil_sendFile: Could not memory \
map file with path '%s'.", pPath );
			PIFMap_delete( pFileMap );
			return PIAPI_ERROR;
			}
		else
			{
			if ( iLen )
				{
				pBuffer->Write( pMap, iLen, iBufferingFlags );
				};
			PIFMap_delete( pFileMap );
			};
		};

	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int HTTPUtil_recvFile( PIHTTP *pPIHTTP, PIIOBuffer *pBuffer, PIFInfo *pFInfo )
		{
		/* --- Open file from descriptor --- */
		PIPLATFORM_FD tFD = 0;
		tFD = PIFile_open( PIFInfo_getPath( pFInfo ), "w" );
		if ( tFD == PIPLATFORM_FD_INVALID ) return PIAPI_ERROR;

		/* --- get length to read --- */
		long iRead = 0;
		long iToRead = strtol((const char *)PIDB_lookup( pPIHTTP->pRequestDB,
			PIDBTYPE_RFC822, KEY_HTTP_CONTENTLENGTH, 0 ), NULL, 10);
		int iBufRead;

		/* --- loop until no more bytes left --- */
		do
			{
			/* --- no more data or client timeout --- */
			if ( PIIOBuffer_pollBeforeRead( pBuffer ) < 1 ) break;

			/* --- read the data --- */
			char *szData = (char *)PIIOBuffer_read( pBuffer, &iBufRead );
			iRead += iBufRead;

			/* --- this is an error --- */
			if ( ! iRead ) 
				{
				PIFile_close( tFD );
				return PIAPI_ERROR;
				};

			/* --- write the data to the file --- */
			if ( PIFile_write( tFD, iBufRead, szData ))
				{
				PIFile_close( tFD );
				return PIAPI_ERROR;
				};
			}
		while ( iRead < iToRead );

		/* --- close the file --- */
		if ( PIFile_close( tFD )) return PIAPI_ERROR;
		return PIAPI_COMPLETED;
	};


/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
 To split a HTTP Multipart Message (RFC1867) into it parts.
\*____________________________________________________________________________*/
PUBLIC_PIAPI int HTTPUtil_recvMultipartMsg( PIHTTP *pPIHTTP, PIIOBuffer *pBuffer,
										   PFN_MUTLIPARTCB cb, void *pUser )
	{
	assert( cb );

	// get length to read
	long iRead = 0;
	char *szContentLength = (char *)PIDB_lookup( pPIHTTP->pRequestDB,
		PIDBTYPE_RFC822, KEY_HTTP_CONTENTLENGTH, 0 );
	if ( !szContentLength ) return PIAPI_ERROR;
	long iToRead = atol( szContentLength );
	int iBufRead;
	unsigned int i = 0;

	enum { IDSIZE = 255 };
	char Id[IDSIZE+1] = "";
	char szName[IDSIZE+1] = "";
	char *szTemp1;
	char *szTemp2;
	enum { BUF_SIZE=8192 };
	char *szBuffer = (char *)PIUtil_malloc( BUF_SIZE );
	if ( !szBuffer ) return PIAPI_ERROR;

	char *szBufRead = szBuffer;
	char *szBufWrite = szBuffer;
	char *szBufEnd = szBuffer + BUF_SIZE;
	int iPhase = MP_PART_NONE;

	// get multipart Id from fist line of http message
	iRead += PIIOBuffer_getLine( pBuffer, Id, IDSIZE );
	iRead += strlen( HTTP_CRLF );

	// loop until no more bytes left
	do
		{
		// no more data or client timeout ?
		if ( PIIOBuffer_pollBeforeRead( pBuffer ) < 1 ) break;
		
		// read the data
		iBufRead = PIIOBuffer_readToBuffer( pBuffer, szBufRead, szBufEnd - szBufRead );

		if ( !strcmp( szName, "" ))
			{
			// Content Disposition
			szTemp1 = strstr( szBufRead, "Content-Disposition: " );
		
			// This is not a multipart message!
			if ( !szTemp1 ) goto error;

			// Name
			szTemp1 = strstr( szTemp1, "name=" );
			if ( !szTemp1 ) goto error;

			szTemp1 += strlen("name=") + 1;
			szTemp2 = strchr( szTemp1, '"' );
			if ( !szTemp2 ) goto error;

			strncpy( szName, szTemp1, szTemp2 - szTemp1 );
			};
			
		// this is a fatal error
		if ( !iBufRead ) goto error;

		// increase read count and pointer
		iRead += iBufRead;
		szBufRead += iBufRead;

		// ring buffer is full
		if ( szBufRead == szBufEnd )
			{ 
			for ( i = 0; i < BUF_SIZE - strlen( Id ); i++ )
				{
				// a new part was found
				if ( !memcmp( szBuffer + i, Id, strlen( Id ))) 
					{
					// this is the end Id, exit loop
					if ( i == BUF_SIZE - strlen( Id ) - 2 - strlen( HTTP_CRLF )) break;
					iPhase = iPhase ? MP_PART_LAST : MP_PART_FIRST;
					if ( cb( iPhase, szName, szBufWrite, i - (szBufWrite - szBuffer) - strlen( HTTP_CRLF ), pPIHTTP, pUser )) goto error;
					szBufWrite += szBufWrite - szBuffer + i + strlen( Id ) + strlen( HTTP_CRLF );

					// Content Disposition
					szTemp1 = strstr( szBufWrite, "Content-Disposition: " );
		
					// This is not a multipart message!
					if ( !szTemp1 ) goto error;

					// Name
					szTemp1 = strstr( szTemp1, "name=" );
					if ( !szTemp1 ) goto error;

					szTemp1 += strlen("name=") + 1;
					szTemp2 = strchr( szTemp1, '"' );
					if ( !szTemp2 ) goto error;

					strncpy( szName, szTemp1, szTemp2 - szTemp1 );
					szName[szTemp2 - szTemp1] = 0;

					iPhase = MP_PART_NONE;
					};
				};

			// the whole rest of the buffer
			iPhase = iPhase ? MP_PART_NEXT : MP_PART_FIRST;
			if ( cb( iPhase, szName, szBufWrite, szBufEnd - szBufWrite - strlen( Id ), pPIHTTP, pUser )) goto error;

			// move the untested margin to the start of the buffer
			memcpy( szBuffer, szBufRead - strlen( Id ), strlen( Id ));
			szBufRead = szBuffer + strlen( Id );
			szBufWrite = szBuffer;
			};
		}
	while ( iRead < iToRead );
		
	// an error occured during reading request data
	if ( iRead < iToRead ) goto error;

	// rest of ring buffer
	for ( i = 0; i < ( szBufRead - szBuffer ) - strlen( Id ); i++ )
		{
		// a new part was found
		if ( !memcmp( szBuffer + i, Id, strlen( Id ))) 
			{
			// this is the end Id, exit loop
			if ( i == ( szBufRead - szBuffer ) - strlen( Id ) - 2 - strlen( HTTP_CRLF )) break;
			iPhase = iPhase ? MP_PART_NEXT | MP_PART_LAST : MP_PART_FIRST | MP_PART_LAST;
			if ( cb( iPhase, szName, szBufWrite, i - ( szBufWrite - szBuffer ) - strlen( HTTP_CRLF ), pPIHTTP, pUser )) goto error;
			szBufWrite += szBufWrite - szBuffer + i + strlen( Id ) + strlen( HTTP_CRLF );

			// Content Disposition
			szTemp1 = strstr( szBufWrite, "Content-Disposition: " );
		
			// This is not a multipart message!
			if ( !szTemp1 ) goto error;

			// Name
			szTemp1 = strstr( szTemp1, "name=" );
			if ( !szTemp1 ) goto error;

			szTemp1 += strlen("name=") + 1;
			szTemp2 = strchr( szTemp1, '"' );
			if ( !szTemp2 ) goto error;

			strncpy( szName, szTemp1, szTemp2 - szTemp1 );
			szName[szTemp2 - szTemp1] = 0;

			iPhase = MP_PART_NONE;
			};
		};

	// the whole rest of the buffer
	iPhase = iPhase ? MP_PART_NEXT | MP_PART_LAST : MP_PART_FIRST | MP_PART_LAST;
	if ( cb( iPhase, szName, szBufWrite, szBufRead - szBufWrite - strlen( HTTP_CRLF )
		- strlen(Id) - 2 - strlen( HTTP_CRLF ),	pPIHTTP, pUser )) goto error;

	// Completion
	PIUtil_free( szBuffer );
	return PIAPI_COMPLETED;
error:
	PIUtil_free( szBuffer );
	return PIAPI_ERROR;
	};


/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static const char *pMethodMap[] = 
{
    0,
    MD_NAME_GET,
    MD_NAME_POST,
    MD_NAME_HEAD,
    MD_NAME_PUT,
    MD_NAME_DELETE,
    MD_NAME_OPTIONS,
    MD_NAME_TRACE,
    MD_NAME_CONNECT,
    0
};
static const char *pProtocolMap[] = 
{
    PR_NAME_UNKNOWN,
    PR_NAME_HTTP09,
    PR_NAME_HTTP10,
    PR_NAME_HTTP11,
    0
};
static const char *pConnectionMap[] = 
{
    "Close",
    "Keep-Alive",
    0
};
int HTTPUtil_doHTTPRequest( PIIOBuffer *pBuffer, PIDB *pQ )
{
    /*
    ** Gather items for the HTTP request and validate them
    */
    const char *pMethod = 0;
    const char *pURI = 0;
    const char *pQS = 0;
    const char *pProtocol = 0;
    const char *pConnection = 0;

    int iMethod = (int)PIDB_lookup( pQ, PIDBTYPE_OPAQUE, KEY_HTTP_METHOD, 0 );
    if ( iMethod<1 || iMethod>MD_LAST )
        { return PIAPI_EINVAL; };
    pMethod = pMethodMap[iMethod];

    int iProtocol = (int)PIDB_lookup( pQ, PIDBTYPE_OPAQUE, KEY_HTTP_PROTOCOL, 0 );
    if ( iProtocol<1 || iProtocol>PR_HTTP11 )
        { return PIAPI_EINVAL; };
    pProtocol = pProtocolMap[iProtocol];

    int iKeepAlive = (int)PIDB_lookup( pQ, PIDBTYPE_OPAQUE, KEY_HTTP_CONNECTION, 0 );
    if ( iKeepAlive<0 || iKeepAlive>1 )
        { return PIAPI_EINVAL; };
    pConnection = pConnectionMap[iKeepAlive];

    pURI = (const char *)PIDB_lookup( pQ, PIDBTYPE_STRING, KEY_HTTP_URI, 0 );
    if ( !pURI )
        { return PIAPI_EINVAL; };

    pQS = (const char *)PIDB_lookup( pQ, PIDBTYPE_STRING, KEY_HTTP_QUERYSTRING, 0 );
    
    /* ---
    Send the HTTP request line and standard headers 
    --- */
    PIIOBuffer_write( pBuffer, pMethod, -1, PIIOBUF_NONE );
    PIIOBuffer_write( pBuffer, " ", 1, PIIOBUF_NONE );
    PIIOBuffer_write( pBuffer, pURI, -1, PIIOBUF_NONE );
    if ( pQS && *pQS )
        {
        PIIOBuffer_write( pBuffer, "?", 1, PIIOBUF_NONE );
        PIIOBuffer_write( pBuffer, pQS, -1, PIIOBUF_NONE );
        };
    PIIOBuffer_write( pBuffer, " ", 1, PIIOBUF_NONE );
    PIIOBuffer_writeLn( pBuffer, pProtocol, -1, PIIOBUF_NONE );

    if (( iKeepAlive && iProtocol == PR_HTTP10 )
    || ( !iKeepAlive && iProtocol == PR_HTTP11 ))
        {
        PIIOBuffer_write( pBuffer, "Connection", -1, PIIOBUF_NONE );
        PIIOBuffer_write( pBuffer, ": ", 2, PIIOBUF_NONE );
        PIIOBuffer_writeLn( pBuffer, pConnection, -1, PIIOBUF_NONE );
        };

    /* ---
    Now send included headers
    --- */
    /* --- iterate over all headers, sending them to server --- */
    PIDBIterator *pRIter = PIDB_getIterator( pQ, PIDBTYPE_RFC822, 0, 0 );
    if ( pRIter )
        {
        for( ;  PIDBIterator_atValidElement( pRIter );
                PIDBIterator_next( pRIter ) )
            {
            const char *pKey;
            const char *pValue = (const char *)
                PIDBIterator_current( pRIter, &pKey );
            if ( pKey && *pKey )
                {
                PIIOBuffer_write( pBuffer, pKey, -1, PIIOBUF_NONE );
                PIIOBuffer_write( pBuffer, ": ", 2, PIIOBUF_NONE );
                PIIOBuffer_writeLn( pBuffer, pValue, -1, PIIOBUF_NONE );
                };
            };
        PIDBIterator_delete( pRIter );
        };
    PIIOBuffer_writeLn( pBuffer, "", 0, PIIOBUF_NONE );

    /* ---
    HTTP request has been started
    --- */
    return PIAPI_COMPLETED;
}

