/*____________________________________________________________________________*\
 *

 Copyright (c) 1997 John Roy. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in
	the documentation and/or other materials provided with the
	distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
	mentioning features must contain the following
	acknowledgment:
	"This product includes software developed by John Roy
	(http://www.johnroy.com/pi3/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Pi3Perl/PspHandler.cpp,v $
 * $Date: 2002/02/11 22:16:36 $
 *
 Description:
	Execute an ISAPI plugin.

\*____________________________________________________________________________*/
//$SourceTop:$

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#define PITHREAD_H_
#include "HandBase.h"
#include "HTTPCore.h"
#include "HTTPUtil.h"
#include "PIStrStr.h"
#include "DeQuote.h"
#include "PiAPI.h"
#include "FlexLexer.h"
#include "psp_scanner.h"
#include <EXTERN.h>
#include <perl.h>

/*
** #define D { cerr << __FILE__ << ": " << __LINE__ << endl; }
*/
#define D

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#define KEY_CONF_DESTRUCTLEVEL		"DestructLevel"

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#if 0
	/*
	** HTML documentation for this handler
	*/
/*___+++HTMLDOC_BEGIN+++___*/
Name:
	PspHandler

Description:
	Load and execute an PSP (Perl Server Pages) extension.

Options:

<TABLE BORDER=1>
<TH>Option
<TH>Default
<TH>Values
<TH>Short Description
<TH>Example(s)

<TR>
<TD>DestructLevel
<TD>-
<TD>0, 1
<TD>define destruction level for perl interpreters
<TD>DestructLevel 1

</TABLE>
<STRONG>-</STRONG> in the <IT>default</IT> indicates no default<BR>
<STRONG>+</STRONG> in the <IT>default</IT> indicates the field is mandatory<BR>

<H4>Description of Options</H4>
<H5>DestructLevel</H5>
A value of 1 will force the explicite destruction of used perl interpreter
instances during server shutdown. This may currently cause server crashes.
A value of 0 (default, recommended) does nothing on shutdown.

Phase:
	HANDLE

Returns:
	PIAPI_COMPLETED or PIAPI_ERROR according to the status returned by the extension.

Note:
	For the PSP language specification refer to the respective Pi3-HOWTO page.

Example:
	<PRE>
	&lt;Object&gt;
		Name PSP
		Class FlexibleHandlerClass
		Condition "&cmp(&dblookup(response,string,ObjectMap),PSP)"
		CheckPath ReturnCode ReturnCode=COMPLETED
		CheckType ReturnCode ReturnCode=COMPLETED
		CheckAccess AccessByFile RequirePermissions="X"
		Handle PerlSrvPages
	&lt;/Object&gt;
	</PRE>

A PSP script to perform Http redirects:
	<PRE>
	&lt;%
	sub pspSetResponse {
		$pspResponse{&quot;Location&quot;} = &quot;http://localhost&quot;.$pspRequest{&quot;PathInfo&quot;};
		$pspResponseStatus = 302;
		return 0;
	}
	%&gt;
	</PRE>

/*___+++HTMLDOC_END+++___*/
#endif


EXTERN_C void boot_DynaLoader (pTHX_ CV* cv);

EXTERN_C void xs_init(pTHX);

class thePSPParser : public PspFlexLexer
{
private:
	PerlInterpreter *my_perl;
	PIHTTP *pHttp;
	PIIOBuffer *pBuf;
	int iFlags;
	int iContentLen;

	virtual int doShebang(const char *txt) {
		/* --- don't send the shebang line to the client! --- */
		return 0;
	};
	virtual int doPlain(const char *txt) {
		if (get_pass()) {
			PIIOBuffer_write( pBuf, txt, strlen(txt), iFlags );
		} else {
			iContentLen += strlen(txt);
		}
		return 0;
	};
	virtual int doScriptlet(const char *txt) {
		PERL_SET_CONTEXT(my_perl);
		eval_pv(txt, TRUE);
		return 0;
	};
	virtual int doExpression(const char *txt) {
		STRLEN n_a;
		PERL_SET_CONTEXT(my_perl);		
		SV *pSV = eval_pv(txt, FALSE);
		const char *pRet = SvPV(pSV, n_a);
		if (get_pass()) {
			PIIOBuffer_write( pBuf, pRet, strlen(pRet), iFlags );
		} else {
			iContentLen += strlen(pRet);
		}
		return 0;
	};
	virtual int doError(const char *msg) {
		HTTPCore_logError(pHttp, msg);
		return 0;
	};
	virtual const char *makeAbsolute(const char *path) {
		Pi3String *s = Pi3String_new("");
		PIDB *pDB = pHttp->pHostDB;
		HTTPCore_relativeToAbsolutePath( pDB, path, s);
		return Pi3String_getPtr( s );
	};

public:
	thePSPParser( const char *ifname, PerlInterpreter *pPerl,
		PIHTTP &tPIHTTP, PIIOBuffer &tBuffer, int theFlags ) :
		PspFlexLexer( ifname ),
		pHttp( &tPIHTTP ),
		pBuf( &tBuffer ),
		my_perl( pPerl ),
		iFlags( theFlags ),
		iContentLen( 0 )
	{
	};

	int getContentLen() { return iContentLen; };
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class PspHandler : public HandlerBasePi3Perl
{
private:
	/* ---
	Configuration data
	--- */
	PIDB *pContextDB;
	const char *pFKPath;
	const char *pFKContentType;
	const char *pFKContentLength;
	const char *pFKMethod;
	const char *pFKQueryString;
	int iDestructLevel;
	PISync *pMutex;

	PerlInterpreter *getPerl( const char *pPath, int *iNew ) {
		*iNew = 0;
		PerlInterpreter *my_perl = (PerlInterpreter *)PIDB_lookup( pContextDB,
			PIDBTYPE_OPAQUE, pPath, 0 );
		if (!my_perl) {
			my_perl = perl_alloc();
			if ( my_perl ) {
				PERL_SET_CONTEXT(my_perl);
				perl_construct(my_perl);
				char *embedding[] = { "", "-e", "0" };
				perl_parse(my_perl, xs_init, 3, embedding, NULL);
				perl_run(my_perl);

				PIDB_add( pContextDB, PIDBTYPE_OPAQUE, pPath, (void *)my_perl, 0 );
				*iNew = 1;
			};
		};
		return my_perl;
	};

	int SetResponse( PerlInterpreter *my_perl, PIHTTP *pPIHTTP, PIIOBuffer *pBuffer ) {

#define SPLIT_HASH(HASH) "foreach $pair (split(/&/, $pspPlain)) {	\
	($name, $value) = split(/=/, $pair);							\
	$name =~ tr/+/ /;												\
	$name =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack(\"C\", hex($1))/eg;	\
	$name =~ s/~!/ ~!/g;											\
	$value =~ tr/+/ /;												\
	$value =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack(\"C\", hex($1))/eg;	\
	$value =~ s/~!/ ~!/g;											\
	$"HASH"{$name} = $value;										\
};"
		PIDB *pQ = pPIHTTP->pRequestDB;
		PIDB *pR = pPIHTTP->pResponseDB;
		const char *pPath = (const char *)PIDB_lookup( pR, PIDBTYPE_STRING,
			pFKPath, PIDBFLAG_FASTKEY );
		const char *pContentType = (const char *)PIDB_lookup( pQ, PIDBTYPE_RFC822,
			pFKContentType, PIDBFLAG_FASTKEY );
		const char *pMethod = (const char *)PIDB_lookup( pQ, PIDBTYPE_STRING,
			pFKMethod, PIDBFLAG_FASTKEY );

		/* --- Set default Content-Type to 'text/html' --- */
		PIDB_replace( pR, PIDBTYPE_RFC822, pFKContentType, (void *)"text/html", PIDBFLAG_FASTKEY );

		PIDBIterator *pIter;
		const char *pKey;
		const char *pValue;

		PERL_SET_CONTEXT(my_perl);

		/* --- Populate %pspRequest --- */
		HV* reqHash = get_hv("pspRequest", FALSE);
		PIString sTmp;
		if (reqHash) {
			pIter = PIDB_getIterator( pQ, PIDBTYPE_RFC822, 0, 0 );
			if ( pIter ) {
				for( ; PIDBIterator_atValidElement( pIter ); PIDBIterator_next( pIter )) {

					pValue = (char *)PIDBIterator_current( pIter, &pKey );
					hv_store(reqHash, pKey, strlen(pKey), newSVpv(pValue, 0), 0);
					HTTPCore_logDebug( DBG_LOW,
						"PSP: Variable Request|RFC822|%s|%s added to request hash.",
						pKey, pValue );

				};
				PIDBIterator_delete( pIter );
			};

			pIter = PIDB_getIterator( pQ, PIDBTYPE_STRING, 0, 0 );
			if ( pIter ) {
				for( ; PIDBIterator_atValidElement( pIter ); PIDBIterator_next( pIter )) {

					pValue = (char *)PIDBIterator_current( pIter, &pKey );
					hv_store(reqHash, pKey, strlen(pKey), newSVpv(pValue, 0), 0);
					HTTPCore_logDebug( DBG_LOW,
						"PSP: Variable Request|STRING|%s|%s added to request hash.",
						pKey, pValue );
				};
				PIDBIterator_delete( pIter );
			};

			pIter = PIDB_getIterator( pR, PIDBTYPE_STRING, 0, 0 );
			if ( pIter ) {
				for( ; PIDBIterator_atValidElement( pIter ); PIDBIterator_next( pIter )) {

					pValue = (char *)PIDBIterator_current( pIter, &pKey );
					hv_store(reqHash, pKey, strlen(pKey), newSVpv(pValue, 0), 0);
					HTTPCore_logDebug( DBG_LOW,
						"PSP: Variable Response|STRING|%s|%s added to request hash.",
						pKey, pValue );
				};
				PIDBIterator_delete( pIter );
			};
		};

		/* --- Populate %pspResponse --- */
		reqHash = get_hv("pspResponse", FALSE);
		if (reqHash) {
			pIter = PIDB_getIterator( pR, PIDBTYPE_RFC822, 0, 0 );
			if ( pIter ) {
				for( ; PIDBIterator_atValidElement( pIter ); PIDBIterator_next( pIter )) {

					pValue = (char *)PIDBIterator_current( pIter, &pKey );
					hv_store(reqHash, pKey, strlen(pKey), newSVpv(pValue, 0), 0);
					HTTPCore_logDebug( DBG_LOW,
						"PSP: Variable Response|RFC822|%s|%s added to response hash.",
						pKey, pValue );
				}
				PIDBIterator_delete( pIter );
			}
		};

		/* --- Populate %pspForm --- */
		if ( pMethod && !PIUtil_stricmp( pMethod, "POST" )) {
			HV* frmHash = get_hv("pspForm", FALSE);
			if ( frmHash ) {

				/* --- reset the temporary --- */
//				SV *tmp = sv_2mortal(get_sv("tmp", TRUE));
				SV *tmp = get_sv("pspPlain", TRUE);
//				sv_clear(tmp);				

				char *pData = (char *)PIDB_lookup( pQ, PIDBTYPE_RFC822,
					pFKContentLength, PIDBFLAG_FASTKEY );
				if (!pData) {
					HTTPCore_logError( pPIHTTP, "PSP: \
HTTP client didn't provide Content-Length in POST request." );
					return HTTPUtil_doHTTPError( pPIHTTP, ST_INTERNALERROR );			
				};
				int iContentLength = atoi(pData);
				int iRead = 0;
				
				enum { BUF_SIZE=2100 }; // 2048 from PIIOBuf.h plus some bytes
				pData = (char *)PIUtil_malloc(BUF_SIZE);
				int iLen = 0;
				for( ; iRead<iContentLength && (iLen=PIIOBuffer_readToBuffer(pBuffer,pData,BUF_SIZE - 1))>0; ) {
					HTTPCore_logDebug( DBG_LOW, "PSP: Read %d bytes from client.", iLen );
					pData[iLen] = 0;
					sv_catpv(tmp, pData);
					iRead += iLen;
				}
				PIUtil_free( pData );

				if ( pContentType && !PIUtil_stricmp( pContentType, "application/x-www-form-urlencoded" )) {

					/* --- urldecode and add to Hash pspForm --- */
					eval_pv( SPLIT_HASH("pspForm"), FALSE );
					/* --- cleanup --- */
//					sv_clear(tmp);
//					sv_free(get_sv("pair", FALSE ));
//					sv_free(get_sv("name", FALSE ));
//					sv_free(get_sv("value", FALSE ));
					HTTPCore_logDebug( DBG_LOW,
						"PSP: %d bytes POST data of type '%s' added to form hash.",
						iRead, pContentType );
				} else {

					/* --- simple scalar pspPlain --- */
					HTTPCore_logDebug( DBG_LOW,
						"PSP: %d bytes POST data of type '%s' added to scalar variable.",
						iRead, pContentType );
				}
			}

		} else {

			const char *pQueryString = (const char *)PIDB_lookup( pQ, PIDBTYPE_STRING,
				pFKQueryString, PIDBFLAG_FASTKEY );
			if (pQueryString) {
				SV *tmp = get_sv("pspPlain", TRUE);
				sv_catpv(tmp, pQueryString);
				/* --- urldecode and add to Hash pspForm --- */
				eval_pv( SPLIT_HASH("pspQueryString"), FALSE );
				/* --- cleanup --- */
//				sv_clear(tmp);
//				sv_free(get_sv("pair", FALSE ));
//				sv_free(get_sv("name", FALSE ));
//				sv_free(get_sv("value", FALSE ));
				HTTPCore_logDebug( DBG_LOW,
					"PSP: QueryString from '%s' request added to form hash.", pMethod );
			};
		}

#undef SPLIT_HASH

		if (!get_cv("pspSetResponse", FALSE)) return PIAPI_COMPLETED;

		/* --- Invoke pspSetResponse() --- */
		char *args[] = { NULL };

		dSP;                            /* initialize stack pointer      */
		ENTER;                          /* everything created after here */
		SAVETMPS;                       /* ...is a temporary variable.   */
		PUSHMARK(SP);                   /* remember the stack pointer    */
		SV *ref = sv_setref_iv(newSViv(0), "const PIObjectPtr", (long)Object());
		XPUSHs(sv_2mortal(ref));
		SV *ref1 = sv_setref_iv(newSViv(0), "PIHTTPPtr", (long)pPIHTTP);
		XPUSHs(sv_2mortal(ref1));
		SV *ref2 = sv_setref_iv(newSViv(0), "PIIOBufferPtr", (long)pBuffer);
		XPUSHs(sv_2mortal(ref2));
		PUTBACK;                        /* make local stack pointer global */
		if ( call_pv("pspSetResponse", G_SCALAR) != 1  )      /* call the function             */
			{
			/* --- an error occurred --- */
			HTTPCore_logError( pPIHTTP, "PSP: \
'pspSetResponse()' returned wrong number of arguments in file %s.", pPath );
			return HTTPUtil_doHTTPError( pPIHTTP, ST_INTERNALERROR );			
			};
		SPAGAIN;                        /* refresh stack pointer         */
		int iRet = POPi;                /* pop the return value from stack */
		PUTBACK;
		FREETMPS;                       /* free that return value        */
		LEAVE;

		switch (iRet) {
			case PIAPI_ERROR:
				HTTPCore_logError( pPIHTTP, "PSP: \
'pspSetResponse()' returned with PIAPI_ERROR.");
				iRet = 1;
				break;

			case PIAPI_COMPLETED:
				{
					/* --- Read %pspResponse --- */
					if (reqHash) {
						hv_iterinit(reqHash);
						long iKeyLen;
						unsigned int iValLen;
						SV* sValue = hv_iternextsv(reqHash, (char **)&pKey, &iKeyLen);
						while (sValue) {
							char *pValue = SvPV(sValue, iValLen);
							PIDB_replace( pR, PIDBTYPE_RFC822, pKey, pValue, 0 );
							HTTPCore_logDebug( DBG_LOW,
								"PSP: Variable Response|RFC822|%s|%s written back to DB.",
								pKey, pValue );
							sValue = hv_iternextsv(reqHash, (char **)&pKey, &iKeyLen);
						};
					};
            
					/* --- Read %pspResponseStatus --- */
					SV* resStatus = get_sv("pspResponseStatus", FALSE);
					if (resStatus) {
						pPIHTTP->iStatus = SvIV(resStatus);
						HTTPCore_logDebug( DBG_LOW,
							"PSP: StatusCode %d written back to HTTP object.",
							pPIHTTP->iStatus );
					};
					break;
				}

			default:
				HTTPCore_logError( pPIHTTP,	"PSP: \
'pspSetResponse()' returned unknown returncode '%d'.", iRet );
				iRet = 1;
		};
		return iRet;
	};

	int OnLoad( PerlInterpreter *my_perl, PIHTTP *pPIHTTP, PIIOBuffer *pBuffer ) {

		PIDB *pQ = pPIHTTP->pRequestDB;
		PIDB *pR = pPIHTTP->pResponseDB;
		const char *pPath = (const char *)PIDB_lookup( pR, PIDBTYPE_STRING,
			pFKPath, PIDBFLAG_FASTKEY );

		/* --- Invoke pspLoad() --- */
		char *args[] = { NULL };

		PERL_SET_CONTEXT(my_perl);
		if (!get_cv("pspLoad", FALSE)) return PIAPI_COMPLETED;

		dSP;                            /* initialize stack pointer      */
		ENTER;                          /* everything created after here */
		SAVETMPS;                       /* ...is a temporary variable.   */
		PUSHMARK(SP);                   /* remember the stack pointer    */
		SV *ref = sv_setref_iv(newSViv(0), "const PIObjectPtr", (long)Object());
		XPUSHs(sv_2mortal(ref));
		PUTBACK;                        /* make local stack pointer global */
		if ( call_pv("pspLoad", G_SCALAR) != 1  )      /* call the function             */
			{
			/* --- an error occurred --- */
			HTTPCore_logError( pPIHTTP, "PSP: \
'pspLoad()' returned wrong number of arguments." );
			return PIAPI_ERROR;
			};
		SPAGAIN;                        /* refresh stack pointer         */
		int iRet = POPi;                /* pop the return value from stack */
		PUTBACK;
		FREETMPS;                       /* free that return value        */
		LEAVE;
		switch (iRet) {
			case PIAPI_ERROR:
				HTTPCore_logError( pPIHTTP, "PSP: \
'pspLoad()' returned with PIAPI_ERROR.");
				iRet = 1;
				break;
			case PIAPI_COMPLETED:
				break;
			default:
				HTTPCore_logError( pPIHTTP, "PSP: \
'pspLoad()' returned unknown returncode '%d'.", iRet);
				iRet = 1;
		};
		return iRet;
	};

	int OnUnload( PerlInterpreter *my_perl, const char *pPath ) {

		PIOStrStream os;
		os << "PSP: ";

		/* --- Invoke pspUnload() --- */
		char *args[] = { NULL };

		PERL_SET_CONTEXT(my_perl);
		if (!get_cv("pspUnload", FALSE)) return PIAPI_COMPLETED;

		dSP;                            /* initialize stack pointer      */
		ENTER;                          /* everything created after here */
		SAVETMPS;                       /* ...is a temporary variable.   */
		PUSHMARK(SP);                   /* remember the stack pointer    */
		SV *ref = sv_setref_iv(newSViv(0), "const PIObjectPtr", (long)Object());
		XPUSHs(sv_2mortal(ref));
		PUTBACK;                        /* make local stack pointer global */
		if ( call_pv("pspUnload", G_SCALAR) != 1  )      /* call the function             */
			{
			/* --- an error occurred --- */
			os << "'pspUnload()' returned wrong number of arguments.";
			CONFIG_ERR( Object(), os.str() );
			return PIAPI_ERROR;
			};
		SPAGAIN;                        /* refresh stack pointer         */
		int iRet = POPi;                /* pop the return value from stack */
		PUTBACK;
		FREETMPS;                       /* free that return value        */
		LEAVE;
		switch (iRet) {
			case PIAPI_ERROR:
				os << "'pspUnload()' returned with PIAPI_ERROR.";
				CONFIG_ERR( Object(), os.str() );
				iRet = 1;
				break;
			case PIAPI_COMPLETED:
				break;
			default:
				os << "'pspUnload()' returned returned unknown returncode '"
				<< iRet << "'.";
				CONFIG_ERR( Object(), os.str() );
				iRet = 1;
		};
		return iRet;
	};

protected:
int Parameter( const char *pVariable, const char *pValue,
	const char *pWhere )
	{
		assert( pVariable && pValue );
		PIOStrStream os;
		os << pWhere << "PSP: ";
		if ( !PIUtil_stricmp( KEY_CONF_DESTRUCTLEVEL, pVariable ) )
			{
			iDestructLevel = atoi( pValue );
			}
		else
			{
			os << "Unknown directive '" << pVariable <<	"'" << ends;
			CONFIG_ERR( Object(), os.str() );
			return 0;
			};
		
		return 1;
	};

public:
	PspHandler( PIObject *pObject, int iArgc, const char *ppArgv[] )
	:	HandlerBasePi3Perl( pObject ),
		pContextDB( PIDB_new( 0, "PSP_Context" ) ),
		pFKPath( PIDB_getFastKey( KEY_INT_PATH, PIDBTYPE_STRING ) ),
		pFKContentType( PIDB_getFastKey( KEY_HTTP_CONTENTTYPE, PIDBTYPE_RFC822 ) ),
		pFKContentLength( PIDB_getFastKey( KEY_HTTP_CONTENTLENGTH, PIDBTYPE_RFC822 ) ),
		pFKMethod( PIDB_getFastKey( KEY_HTTP_METHOD, PIDBTYPE_STRING ) ),
		pFKQueryString( PIDB_getFastKey( KEY_HTTP_QUERYSTRING, PIDBTYPE_STRING ) ),
		iDestructLevel( 0 ),
		pMutex ( PIPlatform_allocLocalMutex()) 
	{
	ReadParameters( iArgc, ppArgv );
	};

~PspHandler()
	{
	/* ---
	iterate through the Context DB and free context
	--- */
	const char *pKey;
	PerlInterpreter *pPerl;
	PIDBIterator *pIter;
	pIter = PIDB_getIterator( pContextDB, PIDBTYPE_OPAQUE, 0, 0 );
	if ( pIter ) {
		for( ; PIDBIterator_atValidElement( pIter ); PIDBIterator_next( pIter )) {

			pPerl = (PerlInterpreter *)PIDBIterator_current( pIter, &pKey );
	
			PERL_SET_CONTEXT(pPerl);
			if (iDestructLevel == 1) {
			    perl_destruct(pPerl);
				perl_free(pPerl);
			};
		};
		PIDBIterator_delete( pIter );
	};

	PIDB_delete( pContextDB );
	};

int Handle( int iPhase, PIHTTP &tPIHTTP, PIIOBuffer &tBuffer )
	{
	if (iPhase != PH_HANDLE) return PIAPI_ERROR;

	PIDB *pR = tPIHTTP.pResponseDB;

	const char *pPath = (const char *)PIDB_lookup( pR, PIDBTYPE_STRING,
		pFKPath, PIDBFLAG_FASTKEY );

	int iNew = 0;
	PerlInterpreter *pPerl = getPerl(pPath, &iNew);
	if (!pPerl) {
		HTTPCore_logError( &tPIHTTP, "PSP: \
Fatal error, couldn't allocate Perl interpreter." );
		return HTTPUtil_doHTTPError( &tPIHTTP, ST_INTERNALERROR );
	};

	/* --- transfer-encoding --- */
	int iFlags = (!PIHTTP_isChild( &tPIHTTP) && (int)PIDB_lookup( tPIHTTP.pRequestDB,
		PIDBTYPE_OPAQUE, KEY_HTTP_PROTOCOL, 0 ) == 3 ) ? PIIOBUF_CHUNKED : PIIOBUF_NONE;
	if ( iFlags & PIIOBUF_CHUNKED )
		{
		PIDB_replace( pR, PIDBTYPE_RFC822, KEY_HTTP_TRANSFERENCODING,
			(void *)KEY_HTTP_CHUNKED, 0 );
		};

	thePSPParser Parser = thePSPParser( pPath, pPerl, tPIHTTP, tBuffer, iFlags );
	int iRet = PIAPI_COMPLETED;

	PISync_lock( pMutex );
	if (Parser.exec()) iRet = HTTPUtil_doHTTPError( &tPIHTTP, ST_INTERNALERROR );
	if (Parser.getParsedLen()) {
		char szBuf[15];
		sprintf(szBuf,"%ul",Parser.getContentLen());
		PIDB_replace( pR, PIDBTYPE_RFC822, KEY_HTTP_CONTENTLENGTH, (void *)szBuf, 0 );
	};
	if ( iNew && OnLoad( pPerl, &tPIHTTP, &tBuffer )) iRet = HTTPUtil_doHTTPError( &tPIHTTP, ST_INTERNALERROR );
	if ( SetResponse( pPerl, &tPIHTTP, &tBuffer )) iRet = HTTPUtil_doHTTPError( &tPIHTTP, ST_INTERNALERROR );
	HTTPCore_sendGeneralHeaders(&tPIHTTP);
	HTTPCore_sendEntityHeaders(&tPIHTTP, pR);
	if ( iFlags & PIIOBUF_CHUNKED ) PIIOBuffer_flush( &tBuffer );
	Parser.prepare_next_pass();
	Parser.exec();
	PISync_unlock( pMutex );

	PIIOBuffer_flush( &tBuffer );
	return iRet;
	};

};


/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int PSP_constructor( PIObject *pObj,
	int iArgc, const char *ppArgv[] )
{
	return HandlerBasePi3Perl_constructor( pObj, PI_NEW( PspHandler( pObj,
		iArgc, ppArgv ) ) );
}

#if 0
/*___+++CNF_BEGIN+++___*/
	<Class>
		Name PSPClass
		Type LogicExtension
		Library Pi3Perl
		OnClassLoad HandlerBasePi3Perl_onClassLoad
		Constructor PSP_constructor
		CopyConstructor HandlerBasePi3Perl_copyConstructor
		Destructor HandlerBasePi3Perl_destructor
		Execute HandlerBasePi3Perl_execute
	</Class>

	<Object>
		Name PerlSrvPages
		Class PSPClass
	</Object>

/*___+++CNF_END+++___*/
#endif

