/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2001 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    and Holger Zimmermann (http://pi3web.sourceforge.net/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/PiAPI/_PIPlat.cpp,v $
 * $Date: 2001/07/18 13:10:49 $
 *
 Description:
\*____________________________________________________________________________*/
//$SourceTop:$

#include "Platform.h"
#include "PiAPI.h"
#include "PlatDefs.h"


/*____________________________________________________________________________*\
 *
 Description:
	See API header files in ../PIAPI for full description
\*____________________________________________________________________________*/
PUBLIC_PIAPI int PIPlatform_enter( const char *pArgv0, int iVersion,
	PIPlatformFn fn )
{
	if ( iVersion != PIPLATFORM_VERSION_1_0 || !fn )
		{ PIERROR( PIAPI_EINVAL ); return PIAPI_ERROR; };
	Platform::Enter( pArgv0, fn );
	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Description:
	See API header files in ../PIAPI for full description
\*____________________________________________________________________________*/
PUBLIC_PIAPI const char *PIPlatform_getDescription()
{
	return Platform::GetDescription();
}

/*____________________________________________________________________________*\
 *
 Description:
	See API header files in ../PIAPI for full description
\*____________________________________________________________________________*/
PUBLIC_PIAPI int PIPlatform_getUniqueId( const char *szBuf, unsigned int iLen )
{
	return Platform::GetUniqueId(szBuf, iLen);
}

/*____________________________________________________________________________*\
 *
 Description:
	See API header files in ../PIAPI for full description
\*____________________________________________________________________________*/
PUBLIC_PIAPI Semaphore *PIPlatform_allocLocalMutex()
{
	return Platform::AllocLocalMutex();
}

/*____________________________________________________________________________*\
 *
 Description:
	See API header files in ../PIAPI for full description
\*____________________________________________________________________________*/
PUBLIC_PIAPI Semaphore *PIPlatform_allocLocalSemaphore( int iInitialCount,
	int iMaxCount )
{
	return Platform::AllocLocalSemaphore( iInitialCount, iMaxCount );
}

/*____________________________________________________________________________*\
 *
 Description:
	See API header files in ../PIAPI for full description
\*____________________________________________________________________________*/
PUBLIC_PIAPI int PIPlatform_pollNetFD( PIPLATFORM_NETFD fdSocket,
	int iFlags, int iTimeout )
{
	return Platform::PollNetFD( fdSocket, iFlags, iTimeout );
}

/*____________________________________________________________________________*\
 *
 Description:
	See API header files in ../PIAPI for full description
\*____________________________________________________________________________*/
PUBLIC_PIAPI int PIPlatform_pollPipeFD( PIPLATFORM_FD fd,
	int iFlags, int iTimeout )
{
	return Platform::PollPipeFD( fd, iFlags, iTimeout );
}
/*____________________________________________________________________________*\
 *
 Description:
	See API header files in ../PIAPI for full description
\*____________________________________________________________________________*/
PUBLIC_PIAPI int PIPlatform_pollFD( PIPLATFORM_FD fd,
	int iFlags, int iTimeout )
{
	return Platform::PollFD( fd, iFlags, iTimeout );
}

/*____________________________________________________________________________*\
 *
 Description:
	See API header files in ../PIAPI for full description
\*____________________________________________________________________________*/
PUBLIC_PIAPI int PIPlatform_sleep( int iMilliseconds )
{
	return Platform::Sleep( iMilliseconds );
}

/*____________________________________________________________________________*\
 *
 Description:
	See API header files in ../PIAPI for full description
\*____________________________________________________________________________*/
PUBLIC_PIAPI int PIPlatform_getLastError()
{
	return Platform::GetLastError();
}

/*____________________________________________________________________________*\
 *
 Description:
	See API header files in ../PIAPI for full description
\*____________________________________________________________________________*/
PUBLIC_PIAPI void PIPlatform_setLastError( int iErrorCode )
{
	Platform::SetLastError( iErrorCode );
}

/*____________________________________________________________________________*\
 *
 Description:
	See API header files in ../PIAPI for full description
\*____________________________________________________________________________*/
typedef void (* Fn)(void *);
PUBLIC_PIAPI void PIPlatform_catchExceptions( Fn pF, void *pV )
{
	Platform::CatchExceptions( pF, pV );
}

/*____________________________________________________________________________*\
 *
 Description:
	See API header files in ../PIAPI for full description
\*____________________________________________________________________________*/
PUBLIC_PIAPI int PIPlatform_beforeUnsafeBlock()
{
	return Platform::BeforeUnsafeBlock();
}

/*____________________________________________________________________________*\
 *
 Description:
	See API header files in ../PIAPI for full description
\*____________________________________________________________________________*/
PUBLIC_PIAPI int PIPlatform_afterUnsafeBlock()
{
	return Platform::AfterUnsafeBlock();
}

/*____________________________________________________________________________*\
 *
 Description:
	See API header files in ../PIAPI for full description
\*____________________________________________________________________________*/
PUBLIC_PIAPI int PIPlatform_getCurrentDirectory( char *pBuffer, int iMaxLen )
{
	if ( !pBuffer || !iMaxLen )
		{ PIERROR( PIAPI_EINVAL ); return -1; };

	PIString sCurrentDirectory;
	if ( Platform::RetrieveCurrentDirectory( sCurrentDirectory )!=
		PIAPI_COMPLETED )
		{
		return -1;
		}
	else
		{
		strncpy( pBuffer, sCurrentDirectory, iMaxLen );
		return sCurrentDirectory.Len();
		};
}

/*____________________________________________________________________________*\
 *
 Description:
	See API header files in ../PIAPI for full description
\*____________________________________________________________________________*/
PUBLIC_PIAPI char PIPlatform_getDirectorySeparator()
{
	return Platform::getDirectorySeparator();
};


/*____________________________________________________________________________*\
 *
 Description:
	See API header files in ../PIAPI for full description
\*____________________________________________________________________________*/
PUBLIC_PIAPI int PIPlatform_getProcess()
{
	return Platform::GetProcess();
}

/*____________________________________________________________________________*\
 *
 Description:
	See API header files in ../PIAPI for full description
\*____________________________________________________________________________*/
PUBLIC_PIAPI int PIPlatform_getProcessId()
{
	return Platform::GetProcessId();
}

/*____________________________________________________________________________*\
 *
 Description:
	See API header files in ../PIAPI for full description
\*____________________________________________________________________________*/
PUBLIC_PIAPI int PIPlatform_waitForProcess( int tProcess, int iFlags,
	int iTimeout )
{
	return Platform::WaitForProcess( tProcess, iFlags, iTimeout );
}

/*____________________________________________________________________________*\
 *
 Description:
	See API header files in ../PIAPI for full description
\*____________________________________________________________________________*/
PUBLIC_PIAPI int PIPlatform_advisaryLockFd( PIPLATFORM_FD tFD )
{
	return Platform::LockFd( tFD );
}

/*____________________________________________________________________________*\
 *
 Description:
	See API header files in ../PIAPI for full description
\*____________________________________________________________________________*/
PUBLIC_PIAPI int PIPlatform_advisaryUnlockFd( PIPLATFORM_FD tFD )
{
	return Platform::UnlockFd( tFD );
}


