/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2001 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    and Holger Zimmermann (http://pi3web.sourceforge.net/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/PiAPI/_PITimer.cpp,v $
 * $Date: 2001/04/07 13:30:22 $
 *
 Description:
\*____________________________________________________________________________*/
//$SourceTop:$

#include "Platform.h"
#include "PiAPI.h"
#include "PlatDefs.h"

/*____________________________________________________________________________*\
 *
 Description:
	See API header files in ../PIAPI for full description
\*____________________________________________________________________________*/
PUBLIC_PIAPI PITimer *PITimer_new()
{
	Timer *pTimer = Platform::AllocTimer();
	if ( !pTimer )
		{ PIERROR( PIAPI_EXHAUSTED ); };
	if ( !pTimer->IsOK() )
		{
		/* --- error is set in Platform::AllocTimer() --- */
		PI_DELETE( pTimer );
		pTimer = 0;
		};
	return pTimer;
}

/*____________________________________________________________________________*\
 *
 Description:
	See API header files in ../PIAPI for full description
\*____________________________________________________________________________*/
PUBLIC_PIAPI int PITimer_reset( PITimer *pTimer )
{
	if ( !pTimer )
		{ PIERROR( PIAPI_EINVAL ); return PIAPI_ERROR; };
	if ( !pTimer->Reset() )
		{ /* error code set internally */ return PIAPI_ERROR; };
	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Description:
	See API header files in ../PIAPI for full description
\*____________________________________________________________________________*/
PUBLIC_PIAPI int PITimer_elapsed( PITimer *pTimer,
	long *plSeconds, long *plMicroSeconds )
{
	if ( !pTimer )
		{ PIERROR( PIAPI_EINVAL ); return PIAPI_ERROR; };
	if ( !pTimer->Elapsed( plSeconds, plMicroSeconds ) )
		{ /* error code set internally */ return PIAPI_ERROR; };
	return PIAPI_COMPLETED;
}
/*____________________________________________________________________________*\
 *
 Description:
	See API header files in ../PIAPI for full description
\*____________________________________________________________________________*/
PUBLIC_PIAPI int PITimer_delete( PITimer *pTimer )
{
	if ( !pTimer )
		{ PIERROR( PIAPI_EINVAL ); return PIAPI_ERROR; };
	PI_DELETE( pTimer );
	return PIAPI_COMPLETED;
}

