/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2001 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    and Holger Zimmermann (http://pi3web.sourceforge.net/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Platform/POSIXFSE.h,v $
 * $Date: 2001/04/07 13:30:22 $
 *
 Description:
\*____________________________________________________________________________*/
//$HeaderTop:$

#ifndef POSIXFSE_H_
#define POSIXFSE_H_

#include "OSConfig.h"

#if defined(CONFIG_OS_POSIX)
#	include "POSIXOSE.h"
	typedef POSIXFileSysType InternalFileSysType;

#elif defined(CONFIG_OS_WIN16)
#	include "Win16OSE.h"
	typedef Win16FATFileSysType InternalFileSysType;

#endif

#include <sys/stat.h>
#include <dirent.h>
#include "FSEntity.h"

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class POSIXFileSystemEntity : public FileSystemEntity
{
private:
	static InternalFileSysType tFSType;
	struct stat tData;
	bool bPopulated;
	bool Populate();
	DIR *pParentDir;
	struct dirent *pDirEntMe;
	POSIXFileSystemEntity( const FileSysType &tFSType,
		const PIString &sParentPath, 
		DIR *pParent, struct dirent *pDirEntMe );

protected:
	virtual void InternalExpireFileInformation()	{ bPopulated=false; };

public:
	POSIXFileSystemEntity( const FileSysType &tFSType,
		const PIString &sFullPath );
	POSIXFileSystemEntity( const POSIXFileSystemEntity &tFSE );
	~POSIXFileSystemEntity();
	virtual bool Exists();
	virtual bool IsRegularFile();
	virtual bool IsDirectory();
	virtual bool IsLink();
	virtual bool IsReadable();
	virtual bool IsWritable();
	virtual bool IsExecutable();
	virtual unsigned long GetFileSize();
	virtual time_t GetLastModified();
	virtual FileSystemEntity *FirstFileInDirectory( int iFlags );
	virtual bool NextSiblingFile();
};

#endif	/* POSIXFSE_H_ */

