/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2001 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    and Holger Zimmermann (http://pi3web.sourceforge.net/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Platform/Win32Err.cpp,v $
 * $Date: 2001/06/03 19:00:16 $
 *
 Description:
	Translate Win32 Error codes in PI error codes.
\*____________________________________________________________________________*/
//$SourceTop:$

#include "OSConfig.h"

#if defined(CONFIG_OS_WIN32)	/* must be Win32 OS */

#if !defined(NDEBUG)
#	include <iostream.h>
#endif

#include "Platform.h"
#include "PIDefs.h"

#include <windows.h>

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int Platform::OSErrorToPIError()
{
	return Platform::OSErrorToPIError( ::GetLastError() );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int Platform::OSErrorToPIError( long lErrorCode )
{
#if !defined(NDEBUG)
	cout << "System error #" << lErrorCode << endl;
#endif
	switch( lErrorCode )
		{
		case ERROR_PATH_NOT_FOUND:
		case ERROR_FILE_NOT_FOUND:
		case ERROR_MOD_NOT_FOUND:
		case ERROR_FILE_EXISTS:
		case ERROR_DLL_NOT_FOUND:
			return PIAPI_EEXIST;

		case ERROR_FILENAME_EXCED_RANGE:
		case ERROR_INVALID_PARAMETER:
		case ERROR_INVALID_HANDLE:
		case ERROR_FILE_INVALID:
		case ERROR_INVALID_NAME:
		case ERROR_BAD_EXE_FORMAT:
		case ERROR_PROC_NOT_FOUND:
		case ERROR_DIRECTORY:
		case ERROR_NOACCESS:
			return PIAPI_EINVAL;

		case ERROR_ACCESS_DENIED:
		case ERROR_SHARING_VIOLATION:
		case ERROR_NOT_OWNER:
			return PIAPI_EACCES;

		case ERROR_TOO_MANY_OPEN_FILES:
		case ERROR_DISK_FULL:
			return PIAPI_EXHAUSTED;

		default:
#if !defined(NDEBUG)
			cout << "Error #" << lErrorCode << endl;
			assert( 0 );
#endif
			return PIAPI_ERROR;
		};
}

#endif	/* CONFIG_OS_WIN32 */

