#!/usr/bin/perl

#*____________________________________________________________________________*\
#*
#
# Copyright (c) 1997-2001 John Roy, Holger Zimmermann. All rights reserved.
#
# These sources, libraries and applications are
# FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
# as long as the following conditions are adhered to.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer. 
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#
# 3. Redistributions of any form whatsoever and all advertising materials 
#    mentioning features must contain the following
#    acknowledgment:
#    "This product includes software developed by John Roy
#    and Holger Zimmermann (http://pi3web.sourceforge.net/)."
#
# THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
#
#*____________________________________________________________________________*|
#*
#* $Source: /cvsroot/pi3web/Pi3Web_200/Source/Scripts/linkify.pl,v $
#* $Date: 2001/04/07 13:30:22 $
#*
# Description:
#	Markup HTML documents by replace indexed keywords with HTML links.
#
# Parameters
#	ARGV[0] - name of the index datafile. Full path or relative from
#				current directory.
#	ARGV[1..n] - Files (relative from current directory, or full path )
#				to look in
#
#*____________________________________________________________________________*/

#/*___________________________________________________________________________*\
# Global declarations:
#\*___________________________________________________________________________*/
%Hoh;
$entry = 0;
$canonical_name = "";

&do_main( @ARGV );

#/*___________________________________________________________________________*\
# *
# Function:
# Synopsis:
# Description:
#\*___________________________________________________________________________*/
sub do_main
	{
	local( $data_file ) = shift;
	print "#\n# linkify.pl, using index file '$data_file' \n#\n";

	# read in index
	open( INP, '<'.$data_file ) || die "Cannot open file $data_file: $!\n";
	&read_file();
	close( INP );
	if ( $^O =~ /MSWin32/ )
		{
		local( @args );
		foreach( @_ ) { push @args, glob; };
		@_=@args;
		};
	foreach( @_ )
		{
		local( $file ) = $_;
		print "linkify.pl: Adding links to file $_...\n";
		open( OUT, '>tmp.tmp' ) || die "Could not open file tmp.tmp: $!\n";
		open( INP, '<'.$file ) || die "Cannot open file $file: $!\n";
		&process_file();
		close( OUT );
		close( INP );
		unlink( $file );
		rename( 'tmp.tmp', $file ) || die "Could not rename file $file: $!\n";
		};
	};

#/*___________________________________________________________________________*\
# *
# Function:
# Synopsis:
# Description:
#	Read indexes from the file into the hash table %Hoh
#\*___________________________________________________________________________*/
sub read_file
	{
	while( <INP> )
		{
		s/\n//;
		local ( $can_name, $location ) = split( /:\s*/, $_, 2 );
		$Hoh{$can_name} = $location;
		};
	};

#/*___________________________________________________________________________*\
# *
# Function:
# Synopsis:
# Description:
#	Scan this file looking for keywords that match words in the hash. 
#	Substitute the keywords with a link to the keywords.
#\*___________________________________________________________________________*/
sub process_file
	{
	while( <INP> )
		{
		s/(\w+)([\s|\(])/&do_subst($1).$2/ge;
		print OUT $_;
		};
	};

#/*___________________________________________________________________________*\
# *
# Function:
# Synopsis:
# Description:
#	Substitute an identifier pattern.
#\*___________________________________________________________________________*/
sub do_subst
	{
	local( $pattern ) = @_;
	local( $location ) = $Hoh{$pattern};
	if ( length( $location ) > 0 )
		{
		$pattern = "<A HREF=\"".$Hoh{$pattern}."\">$pattern".'</A>';
		};
	return $pattern;
	};

