%start page
%pure_parser
%token PLAIN NL WHITESPACE
%token START_TAG END_TAG
%token START_JSP END_JSP
%token ACTION_START JSP_USEBEAN JSP_SETPROPERTY JSP_GETPROPERTY JSP_INCLUDE JSP_FORWARD JSP_PLUGIN JSP_PARAMS JSP_PARAM JSP_FALLBACK ACTION_END
%token JSP_ID JSP_SCOPE JSP_CLASS JSP_TYPE JSP_BEANNAME
%token JSP_NAME JSP_PROPERTY JSP_PROPERTYPARAM JSP_VALUE JSP_EXPRESSION
%token JSP_PAGE JSP_FLUSH
%token JSP_CODE JSP_CODEBASE JSP_ALIGN JSP_ARCHIVE JSP_HEIGHT JSP_HSPACE JSP_JREVERSION JSP_JNAME JSP_VSPACE JSP_WIDTH JSP_NSPLUGINURL JSP_IEPLUGINURL
%token SCRIPTLET_START SCRIPTLET SCRIPTLET_END
%token DIRECTIVE_START DIR_PAGE DIR_INCLUDE DIR_TAGLIB BOOL STRING NUMBER DIRECTIVE_END
%token DIR_LANGUAGE DIR_EXTENDS DIR_IMPORT DIR_SESSION DIR_BUFFER DIR_AUTOFLUSH DIR_ISTHREADSAFE DIR_INFO DIR_ERRORPAGE DIR_CONTENTTYPE DIR_ISERRORPAGE
%token DIR_PREFIX DIR_URI
%token DIR_INCFILE
%token PUSH_INCLUDE POP_INCLUDE
%token DECLARATION_START DECLARATION DECLARATION_END
%token EXPRESSION_START EXPRESSION EXPRESSION_END
%token COMMENT_START COMMENT COMMENT_END
%token ERROR_INCLUDES_TOO_DEEPLY ERROR_CANT_OPEN_INCLUDE

%{
/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2001 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    and Holger Zimmermann (http://pi3web.sourceforge.net/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Servlet/JspParser/jsp_parser.y,v $
 * $Date: 2001/09/09 12:20:00 $
 *
 Description:
\*____________________________________________________________________________*/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <assert.h>
#include "jsp_scanner.h"

#define DIRECTIVE_HDR(name, detail) if (PG->pass) fprintf( PG->out, "\n\t\t\t/**\n\t\t\t* Embedded Directive '%s' %s\n", name, detail)
#define DIRATTR_HDR() if (PG->pass) fprintf( PG->out, "\t\t\t* %s = ", (char *)pi3lval)

#define JSPACTION_HDR(name, detail) if (PG->pass) fprintf( PG->out, "\n\t\t\t/**\n\t\t\t* Embedded Action '%s' %s\n\t\t\t**/\n", name, detail)
//#define JSPATTR_HDR() if (PG->pass) fprintf( PG->out, "\t\t\t* %s = ", (char *)pi3lval)

#define YYERROR_VERBOSE
#define MAX_ERROR_COUNT 10
%}

%%
page :
		|	part
		;

part :		part part
		|	txt
		|	START_TAG tag END_TAG
		|	START_JSP jsp END_JSP
		|	error
			{
			if (pi3nerrs < MAX_ERROR_COUNT)
				{
				if (PG->pass) 
					{
					fprintf( PG->out, "/* >>>>> PARSE ERROR near token '%s', line %d <<<<< */\n", (char *)pi3lval, get_lineno(pp));
					}
				else
					{
					fprintf( stderr, "PARSE ERROR near token '%s', line %d\n", (char *)pi3lval, get_lineno(pp));
					};

				yyerrok; yyclearin;
				}
			else
				{
				fprintf( stderr, "Giving up due to too much errors (%d).\n", pi3nerrs);
				YYABORT;
				};
			}
		;

tag :		commt
		|	script
		|	decl
		|	expr
		|	dir
		;

txt :		PLAIN
			{
			if (PG->pass)
				{
				pi3write_plain( PG, (const char *)pi3lval );
				}
			}
		|	NL
			{
			if (PG->pass) fprintf( PG->out, "\t\t\tout.newLine();\n");
			}
		;

jsp :		usebean
		|	setprop
		|	getprop
		|	include
		|	forward
		|	plugin

usebean :	usebean_start usebean_text jsp_end
		|	usebean part
		;

setprop :	setprop_start setprop_text jsp_end
		;

getprop :	getprop_start getprop_text jsp_end
		;

include :	include_start include_text jsp_end
		;

forward :	forward_start forward_text jsp_end
		;

plugin :	plugin_start plugin_text jsp_end
		|	plugin plugin_body
		;

plugin_body : params
		|	fallback
		|	params fallback
		;

params :	params_start params_body jsp_end
		;

params_body : param_start param_text jsp_end
		|	params_body params_body
		;

fallback :	fallback_start PLAIN jsp_end
		;

commt :		cmt_start cmt_text cmt_end
		|	cmt_start cmt_end
		;

cmt_start : COMMENT_START
			{
			if (PG->pass) fprintf( PG->out, "\t\t\t\n\t\t\t/**\n\t\t\t");
			}
		;

cmt_end :	COMMENT_END
			{
			if (PG->pass) fprintf( PG->out, "\n\t\t\t**/\n");
			}
		;

cmt_text :	COMMENT
			{
			if (PG->pass) fprintf( PG->out, "%s", (char *)pi3lval);
			}
		|	WHITESPACE
			{
			if (PG->pass)
				{
				fprintf( PG->out, "%s", (char *)pi3lval);
				if ((((char *)pi3lval)[0]==13) || (((char *)pi3lval)[0]==10))
					fprintf( PG->out, "\t\t\t");
				};
			}
		|	cmt_text cmt_text
		;

script :	scr_start scr_text scr_end
		|	scr_start scr_end
		;

scr_start :	SCRIPTLET_START
			{
			if (PG->pass) fprintf( PG->out, "\n\t\t\t/**\n\t\t\t* Embedded Scriptlet\n\t\t\t**/\n\t\t\t");
			}
		;

scr_end :	SCRIPTLET_END
			{
			if (PG->pass) fprintf( PG->out, "\n\n");
			}
		;

scr_text :	SCRIPTLET
			{
			if (PG->pass) fprintf( PG->out, "%s", (char *)pi3lval);
			}
		|	WHITESPACE
			{
			if (PG->pass)
				{
				fprintf( PG->out, "%s", (char *)pi3lval);
				if ((((char *)pi3lval)[0]==13) || (((char *)pi3lval)[0]==10))
					fprintf( PG->out, "\t\t\t");
				};
			}
		|	scr_text scr_text
		;

decl :		dec_start dec_text dec_end
		|	dec_start dec_end
		;

dec_start :	DECLARATION_START
			{
			if (PG->pass) fprintf( PG->out, "\n\t\t\t/**\n\t\t\t* Embedded Declaration\n\t\t\t**/\n\t\t\t");
			}
		;

dec_end :	DECLARATION_END
			{
			if (PG->pass) fprintf( PG->out, "\n\n");
			}
		;

dec_text :	DECLARATION
			{
			if (PG->pass) fprintf( PG->out, "%s", (char *)pi3lval);
			}
		|	WHITESPACE
			{
			if (PG->pass)
				{
				fprintf( PG->out, "%s", (char *)pi3lval);
				if ((((char *)pi3lval)[0]==13) || (((char *)pi3lval)[0]==10))
					fprintf( PG->out, "\t\t\t");
				};
			}
		|	dec_text dec_text
		;

expr :		exp_start exp_text exp_end
		|	exp_start exp_end
		;

exp_start :	EXPRESSION_START
			{
			if (PG->pass) fprintf( PG->out, "\n\t\t\t/**\n\t\t\t* Embedded Expression\n\t\t\t**/\n");
			if (PG->pass) fprintf( PG->out, "\t\t\tout.print(");
			}
		;

exp_end	:	EXPRESSION_END
			{
			if (PG->pass) fprintf( PG->out, ");\n\n");
			}
		;

exp_text :	EXPRESSION
			{
			if (PG->pass) fprintf( PG->out, "%s", (char *)pi3lval);
			}
		|	WHITESPACE
			{
			if (PG->pass)
				{
				fprintf( PG->out, "%s", (char *)pi3lval);
				if ((((char *)pi3lval)[0]==13) || (((char *)pi3lval)[0]==10))
					fprintf( PG->out, "\t\t\t");
				};
			}
		|	exp_text exp_text
		;

dir :		page_start page_text dir_end
		|	incl_start incl_text dir_end
		|	tagl_start tagl_text dir_end
		;

page_start : DIRECTIVE_START DIR_PAGE
			{
			DIRECTIVE_HDR((char *)$2, "");
			PG->directive = DIR_PAGE;
			}
		;

incl_start : DIRECTIVE_START DIR_INCLUDE
			{
			DIRECTIVE_HDR((char *)$2, "");
			PG->directive = DIR_INCLUDE;
			}
		;

tagl_start : DIRECTIVE_START DIR_TAGLIB
			{
			DIRECTIVE_HDR((char *)$2, "(unsupported)");
			PG->directive = DIR_TAGLIB;
			}
		;

dir_end	:	DIRECTIVE_END
			{
			if (PG->pass) fprintf( PG->out, "\t\t\t**/\n\n");
			}
		;

page_text :	page_attr '=' '"' page_value '"'
			{
			if (PG->pass && (PG->dir_attribute != DIR_IMPORT))
				{
				if (!PG->page_attribs[PG->dir_attribute - DIR_LANGUAGE])
					{
					++PG->page_attribs[PG->dir_attribute - DIR_LANGUAGE];
					}
				else
					{
					fprintf( PG->out, "\n/**\n* ERROR: Multiple definitions of attribute '%s' found in page directive.\n**/\n", (char *)$1);
					YYABORT;							
					};
				};
			}
		|	WHITESPACE
		|	page_text page_text
		;

incl_text :	incl_attr '=' '"' push_inc page pop_inc '"'
		|	incl_attr '=' '"' incl_err '"'
		|	WHITESPACE
		|	incl_text incl_text
		;

incl_err :	ERROR_INCLUDES_TOO_DEEPLY
			{
			fprintf( stderr, "\n/*\n* ERROR: Includes nested to deeply for file: %s. \
Maximum depth is %d.\n**/\n", (char *)$1, MAX_IMPORT_DEPTH);
			YYABORT;
			}

		|	ERROR_CANT_OPEN_INCLUDE
			{
			fprintf( stderr, "\n/**\n* ERROR: Can't open include file: %s.\n**/\n", (char *)$1);
			YYABORT;
			}
		;

tagl_text :	tagl_attr '=' '"' tagl_value '"'
		|	WHITESPACE
		|	tagl_text tagl_text
		;

page_attr : DIR_LANGUAGE
			{
			DIRATTR_HDR();
			PG->dir_attribute = DIR_LANGUAGE;
			}
		|	DIR_EXTENDS
			{
			DIRATTR_HDR();
			PG->dir_attribute = DIR_EXTENDS;
			}
		|	DIR_IMPORT
			{
			DIRATTR_HDR();
			PG->dir_attribute = DIR_IMPORT;
			}
		|	DIR_SESSION
			{
			DIRATTR_HDR();
			PG->dir_attribute = DIR_SESSION;
			}
		|	DIR_BUFFER
			{
			DIRATTR_HDR();
			PG->dir_attribute = DIR_BUFFER;
			}
		|	DIR_AUTOFLUSH
			{
			DIRATTR_HDR();
			PG->dir_attribute = DIR_AUTOFLUSH;
			}
		|	DIR_ISTHREADSAFE
			{
			DIRATTR_HDR();
			PG->dir_attribute = DIR_ISTHREADSAFE;
			}
		|	DIR_INFO
			{
			DIRATTR_HDR();
			PG->dir_attribute = DIR_INFO;
			}
		|	DIR_ERRORPAGE
			{
			DIRATTR_HDR();
			PG->dir_attribute = DIR_ERRORPAGE;
			}
		|	DIR_CONTENTTYPE
			{
			DIRATTR_HDR();
			PG->dir_attribute = DIR_CONTENTTYPE;
			}
		|	DIR_ISERRORPAGE
			{
			DIRATTR_HDR();
			PG->dir_attribute = DIR_ISERRORPAGE;
			}
		;

tagl_attr :	DIR_URI
			{
			DIRATTR_HDR();
			PG->dir_attribute = DIR_URI;
			}
		|	DIR_PREFIX
			{
			DIRATTR_HDR();
			PG->dir_attribute = DIR_PREFIX;
			}
		;

incl_attr : DIR_INCFILE
			{
			DIRATTR_HDR();
			PG->dir_attribute = DIR_INCFILE;
			}
		;

page_value : BOOL
			{
			if (PG->pass) fprintf( PG->out, "%d\n", pi3lval); else
				{
				switch (PG->dir_attribute)
					{
					case DIR_SESSION: PG->session = pi3lval; break;
					case DIR_ISTHREADSAFE: PG->isThreadSafe = pi3lval; break;
					case DIR_AUTOFLUSH: PG->autoFlush = pi3lval; break;
					case DIR_ISERRORPAGE: PG->isErrorPage = pi3lval;
					};
				};
			}
		|	STRING
			{
			if (PG->pass) fprintf( PG->out, "%s\n", (char *)pi3lval); else
				{
				switch (PG->dir_attribute)
					{
					case DIR_LANGUAGE: 
						if (strcmp((char *)pi3lval, "java"))
							{
							if (PG->pass) fprintf( PG->out, "\n/**\n* ERROR: Invalid value %s of attribute 'language' \
(only 'java' is allowed here).\n**/\n", (char *)pi3lval);
							YYABORT;
							};
						break;
					case DIR_BUFFER:
						if (!strcmp((char *)pi3lval, "none"))
							{
							PG->buffer = -1;
							}
						else 
							{
							// we parse this by hand to avoid trouble with numbers
							// within other quoted values
							PG->buf = strspn((char *)pi3lval, "0123456789");
							if (PG->buf)
								{
								((char *)pi3lval)[PG->buf] = 0;
								PG->buffer = atoi((const char *)pi3lval);
								if (PG->pass && strcmp((const char *)(((char *)pi3lval)[PG->buf+1]), "kB"))
									{
									fprintf( PG->out, "\n/**\n* ERROR: Invalid value %s of attribute 'buffer' \
(only 'none'|'<size>kB' are allowed here).\n**/\n", (char *)pi3lval);
									YYABORT;
									};
								}
							else
								{
								if (PG->pass)
									{
									fprintf( PG->out, "\n/**\n* ERROR: Invalid value %s of attribute 'buffer' \
(only 'none'|'<size>kB' are allowed here).\n**/\n", (char *)pi3lval);
									YYABORT;
									};
								};
							};
						break;
					case DIR_IMPORT:
						if (PG->import_ptr <= MAX_IMPORT_DEPTH)
							{
							strcpy(PG->imports[PG->import_ptr++], (char *)pi3lval);
							}
						else
							{
							if (PG->pass) fprintf( PG->out, "\n/**\n* ERROR: Too much imports in page directive, \
define %d at maximum.\n**/\n", MAX_IMPORT_DEPTH);
							};
						break;
					case DIR_EXTENDS: strcpy(PG->extends, (char *)pi3lval); break;
					case DIR_INFO: strcpy(PG->info, (char *)pi3lval); break;
					case DIR_ERRORPAGE: strcpy(PG->errorPage, (char *)pi3lval); break;
					case DIR_CONTENTTYPE: strcpy(PG->contentType, (char *)pi3lval);
					};
				};
			}
		|	NUMBER STRING
			{
			if (PG->pass)
				{
				fprintf( PG->out, "%d%s\n", $1, $2);
				if ((PG->dir_attribute == DIR_BUFFER) && strcmp((char *)$2, "kB"))
					{
					fprintf( PG->out, "\n/**\n* ERROR: Invalid unit %s of attribute 'buffer' \
(only 'kB' is allowed here).\n**/\n", (char *)pi3lval);
					YYABORT;
					};
				}
			else
				{
				switch (PG->dir_attribute)
					{
					case DIR_BUFFER: PG->buffer = $1;
					};
				};
			}
		;

tagl_value : STRING
			{
			if (PG->pass) fprintf( PG->out, "%s\n", (char *)pi3lval); else
				{
				switch (PG->dir_attribute)
					{
					case DIR_PREFIX: break;
					case DIR_URI: break;
					};
				}
			}
		;

push_inc :	PUSH_INCLUDE
			{
			if (PG->pass) fprintf( PG->out, "%s, level %d, included text follows >>>>>\n\t\t\t**/\n",
				get_incl_name(PG), get_incl_level(PG));
			}
		;

pop_inc	:	POP_INCLUDE
			{
			if (PG->pass) fprintf( PG->out, "\t\t\t/**\n\t\t\t* >>>>> End of included text, level %d\n",
				get_incl_level(PG) + 1);
			}
		;

usebean_start : ACTION_START JSP_USEBEAN
			{
			JSPACTION_HDR((char *)$2, "");
			PG->action = JSP_USEBEAN;
			}
		;

setprop_start : ACTION_START JSP_SETPROPERTY
			{
			JSPACTION_HDR((char *)$2, "");
			PG->action = JSP_SETPROPERTY;
			}
		;

getprop_start : ACTION_START JSP_GETPROPERTY
			{
			JSPACTION_HDR((char *)$2, "");
			PG->action = JSP_GETPROPERTY;
			}
		;

include_start : ACTION_START JSP_INCLUDE
			{
			JSPACTION_HDR((char *)$2, "");
			if (PG->pass) fprintf( PG->out, "\t\t\tapplication.getRequestDispatcher(");
			PG->action = JSP_INCLUDE;
			}
		;

forward_start : ACTION_START JSP_FORWARD
			{
			JSPACTION_HDR((char *)$2, "");
			if (PG->pass) fprintf( PG->out, "\t\t\tapplication.getRequestDispatcher(");
			PG->action = JSP_FORWARD;
			}
		;

plugin_start : ACTION_START JSP_PLUGIN
			{
			JSPACTION_HDR((char *)$2, "(unsupported)");
			PG->action = JSP_PLUGIN;
			}
		;

params_start : ACTION_START JSP_PARAMS
			{
			JSPACTION_HDR((char *)$2, "(unsupported)");
			PG->action = JSP_PARAMS;
			}
		;

param_start : ACTION_START JSP_PARAM
			{
			JSPACTION_HDR((char *)$2, "(unsupported)");
			PG->action = JSP_PARAM;
			}
		;

fallback_start : ACTION_START JSP_FALLBACK
			{
			JSPACTION_HDR((char *)$2, "(unsupported)");
			PG->action = JSP_FALLBACK;
			}
		;

usebean_text : usebean_attr '=' '"' jsp_value '"'
		|	WHITESPACE
		|	usebean_text usebean_text
		;

setprop_text : setprop_attr '=' '"' jsp_value '"'
		|	setprop_attr '=' ''' jsp_expr '''
		|	WHITESPACE
		|	setprop_text setprop_text
		;

getprop_text : getprop_attr '=' '"' jsp_value '"'
		|	WHITESPACE
		|	getprop_text getprop_text
		;

include_text : include_attr '=' '"' jsp_value '"'
		|	include_attr '=' ''' jsp_expr '''
		|	WHITESPACE
		|	include_text include_text
		;

forward_text : forward_attr '=' '"' jsp_value '"'
		|	forward_attr '=' ''' jsp_expr '''
		|	WHITESPACE
		|	forward_text WHITESPACE
		|	WHITESPACE forward_text
		;

plugin_text : plugin_attr  '=' '"' jsp_value '"'
		|	WHITESPACE
		|	plugin_text plugin_text
		;

param_text : param_attr  '=' '"' jsp_value '"'
		|	WHITESPACE
		|	param_text param_text
		;

usebean_attr : JSP_ID
			{
			// !!! Check duplicate ID's !!!
			PG->act_attribute = JSP_ID;
			}
		|	JSP_SCOPE
			{
			PG->act_attribute = JSP_SCOPE;
			}
		|	JSP_CLASS
			{
			PG->act_attribute = JSP_CLASS;
			}
		|	JSP_BEANNAME
			{
			PG->act_attribute = JSP_BEANNAME;
			}
		|	JSP_TYPE
			{
			PG->act_attribute = JSP_TYPE;
			}
		;

setprop_attr : JSP_NAME
			{
			PG->act_attribute = JSP_NAME;
			}
		|	JSP_PROPERTY
			{
			PG->act_attribute = JSP_PROPERTY;
			}
		|	JSP_PROPERTYPARAM
			{
			PG->act_attribute = JSP_PROPERTYPARAM;
			}
		|	JSP_VALUE
			{
			PG->act_attribute = JSP_VALUE;
			}
		;

getprop_attr : JSP_NAME
			{
			PG->act_attribute = JSP_NAME;
			}
		|	JSP_PROPERTY
			{
			PG->act_attribute = JSP_PROPERTY;
			}
		;

include_attr : JSP_PAGE
			{
			PG->act_attribute = JSP_PAGE;
			}
		|	JSP_FLUSH
			{
			PG->act_attribute = JSP_FLUSH;
			}
		;

forward_attr : JSP_PAGE
			{
			PG->act_attribute = JSP_PAGE;
			}
		;

plugin_attr : JSP_TYPE
			{
			PG->act_attribute = JSP_TYPE;
			}
		|	JSP_CODE
			{
			PG->act_attribute = JSP_CODE;
			}
		|	JSP_CODEBASE
			{
			PG->act_attribute = JSP_CODEBASE;
			}
		|	JSP_ALIGN
			{
			PG->act_attribute = JSP_ALIGN;
			}
		|	JSP_ARCHIVE
			{
			PG->act_attribute = JSP_ARCHIVE;
			}
		|	JSP_HEIGHT
			{
			PG->act_attribute = JSP_HEIGHT;
			}
		|	JSP_HSPACE
			{
			PG->act_attribute = JSP_HSPACE;
			}
		|	JSP_JREVERSION
			{
			PG->act_attribute = JSP_JREVERSION;
			}
		|	JSP_JNAME
			{
			PG->act_attribute = JSP_JNAME;
			}
		|	JSP_VSPACE
			{
			PG->act_attribute = JSP_VSPACE;
			}
		|	JSP_WIDTH
			{
			PG->act_attribute = JSP_WIDTH;
			}
		|	JSP_NSPLUGINURL
			{
			PG->act_attribute = JSP_NSPLUGINURL;
			}
		|	JSP_IEPLUGINURL
			{
			PG->act_attribute = JSP_IEPLUGINURL;
			}
		;

param_attr : JSP_NAME
			{
			PG->act_attribute = JSP_NAME;
			}
		|	JSP_VALUE
			{
			PG->act_attribute = JSP_VALUE;
			}
		;

jsp_expr :	jexp_start jexp_text jexp_end
		|	EXPRESSION_START EXPRESSION_END
		;

jexp_start : EXPRESSION_START
			{
			if (PG->pass && PG->bs.prop.name[0])
				{
				fprintf( PG->out, "\t\t\tBeanProperty.set(%s,\"%s\",",
					PG->bs.prop.name, PG->bs.prop.property );
				}
			}
		;
jexp_text :	JSP_EXPRESSION
			{
			if (PG->pass) fprintf( PG->out, "%s", (char *)pi3lval);
			}
		|	WHITESPACE
			{
			if (PG->pass)
				{
				fprintf( PG->out, "%s", (char *)pi3lval);
				if ((((char *)pi3lval)[0]==13) || (((char *)pi3lval)[0]==10))
					fprintf( PG->out, "\t\t\t\t");
				};
			}
		|	jexp_text jexp_text
		;

jexp_end : EXPRESSION_END
			{
			if (PG->pass && PG->bs.prop.name[0])
				{
				fprintf( PG->out, ");\n");
				}
			}
		;

jsp_value : BOOL
		|	STRING
			{
			if (PG->pass)
				{
				switch (PG->action)
					{
					case JSP_GETPROPERTY:
						switch (PG->act_attribute)
							{
							case JSP_NAME:
								strcpy((char *)&(PG->bs.prop.name), (char *)pi3lval);
								break;
							case JSP_PROPERTY:
								strcpy((char *)&(PG->bs.prop.property), (char *)pi3lval);
								break;
							default:
								{
								fprintf( PG->out, "\nPARSE ERROR\n");
//								YYABORT;
								};
							}
						break;
					case JSP_SETPROPERTY:
						switch (PG->act_attribute)
							{
							case JSP_NAME:
								strcpy((char *)&(PG->bs.prop.name), (char *)pi3lval);
								break;
							case JSP_PROPERTY:
								strcpy((char *)&(PG->bs.prop.property), (char *)pi3lval);
								break;
							case JSP_PROPERTYPARAM:
								strcpy((char *)&(PG->bs.prop.param), (char *)pi3lval);
								break;
							case JSP_VALUE:
								strcpy((char *)&(PG->bs.prop.value), (char *)pi3lval);
								break;
							default:
								{
								fprintf( PG->out, "\nPARSE ERROR\n");
//								YYABORT;
								};
							}
						break;
					case JSP_USEBEAN:
						switch (PG->act_attribute)
							{
							case JSP_ID:
								strcpy((char *)&(PG->bs.bean.id), (char *)pi3lval);
								break;
							case JSP_SCOPE:
								if (!strcmp("page", (char *)pi3lval)) strcpy((char *)&(PG->bs.bean.scope), "PAGE_SCOPE");
								if (!strcmp("request", (char *)pi3lval)) strcpy((char *)&(PG->bs.bean.scope), "REQUEST_SCOPE");
								if (!strcmp("session", (char *)pi3lval)) strcpy((char *)&(PG->bs.bean.scope), "SESSION_SCOPE");
								if (!strcmp("application", (char *)pi3lval)) strcpy((char *)&(PG->bs.bean.scope), "APPLICATION_SCOPE");
								break;
							case JSP_CLASS:
								strcpy((char *)&(PG->bs.bean.className), (char *)pi3lval);
								break;
							case JSP_TYPE:
								strcpy((char *)&(PG->bs.bean.type), (char *)pi3lval);
								break;
							case JSP_BEANNAME:
								strcpy((char *)&(PG->bs.bean.beanName), (char *)pi3lval);
								break;
							default:
								{
								fprintf( PG->out, "\nPARSE ERROR\n");
//								YYABORT;
								};
							}
						break;
					case JSP_INCLUDE:
						switch (PG->act_attribute)
							{
							case JSP_PAGE:
								{
								fprintf( PG->out, "\"%s\"", (char *)pi3lval);
								break;
								};
							case JSP_FLUSH:
								break; // we do nothing because flush=true is mandatory in JSP 1.0
							default:
								{
								fprintf( PG->out, "\nPARSE ERROR\n");
//								YYABORT;
								};
							};
						break;
					case JSP_FORWARD:
						switch (PG->act_attribute)
							{
							case JSP_PAGE:
								{
								fprintf( PG->out, "\"%s\"", (char *)pi3lval);
								break;
								};
							default:
								{
								fprintf( PG->out, "\nPARSE ERROR\n");
//								YYABORT;
								};
							};
						break;
					default:
						{
						fprintf( PG->out, "\nPARSE ERROR\n");
//						YYABORT;
						};
					};
				};
			}
		;

jsp_end :	ACTION_END
			{
			if (PG->pass)
				{
				switch (PG->action)
					{
					case JSP_GETPROPERTY:
						fprintf( PG->out, "\t\t\tout.print(BeanProperty.get(%s,\"%s\"));\n",
							PG->bs.prop.name, PG->bs.prop.property );
						memset(&(PG->bs.bean), 0, sizeof(PG->bs.bean));
						break;
					case JSP_SETPROPERTY:
						if ( PG->bs.prop.property[0] ) 
							{
							if (!strcmp(PG->bs.prop.property, "*"))
								{
								if ( PG->bs.prop.param[0] || PG->bs.prop.value[0] )
									{
									fprintf( PG->out, "\n/**\n* ERROR: Illegal options 'param' or 'value' \
in action jsp:setproperty.\n**/\n");
									YYABORT;
									};

								}
							else
								{
								if ( PG->bs.prop.param[0] && PG->bs.prop.value[0] )
									{
									fprintf( PG->out, "\n/**\n* ERROR: Illegal usage of 'param' and 'value' \
together in action jsp:setproperty.\n**/\n");
									YYABORT;
									};									
								};
							pi3write_setproperty( PG );
							memset(&(PG->bs.bean), 0, sizeof(PG->bs.bean));
							}
						else
							{
							fprintf( PG->out, "\n/**\n* ERROR: Missing mandatory parameter 'property' \
in action jsp:setproperty.\n**/\n");

//							YYABORT;
							};
						break;
					case JSP_USEBEAN:
						if ( !PG->bs.bean.id[0] )
							{
							if (PG->pass) fprintf( PG->out, "\n/*\n* ERROR: Attribute 'id' is mandatory \
in JSP action 'usebean'.\n**/\n");
							YYABORT;
							};

						if ( !PG->bs.bean.className[0] && !PG->bs.bean.type[0] )
							{
							if (PG->pass) fprintf( PG->out, "\n/*\n* ERROR: Either 'class' or 'type' \
has to be specified in JSP action 'usebean' with id '%s'.\n**/\n", PG->bs.bean.id);
							YYABORT;
							};

						if ( PG->bs.bean.className[0] && PG->bs.bean.beanName[0] )
							{
							if (PG->pass) fprintf( PG->out, "\n/*\n* ERROR: Only either 'class' or 'beanName' \
allowed in JSP action 'usebean' with id '%s'.\n**/\n", PG->bs.bean.id);
							YYABORT;
							};

						pi3write_bean(PG);
						memset(&(PG->bs.bean), 0, sizeof(PG->bs.bean));
						break;
					case JSP_INCLUDE:
						switch (PG->act_attribute)
							{
							case JSP_PAGE:
								{
								fprintf( PG->out, ").include(request, response);\n\t\t\tout.flush();\n\n");
								break;
								};
							case JSP_FLUSH:
								fprintf( PG->out, ").include(request, response);\n\t\t\tout.flush();\n\n");
								break;
							default:
								{
								break;
//								YYABORT;
								};
							};
						break;
					case JSP_FORWARD:
						switch (PG->act_attribute)
							{
							case JSP_PAGE:
								{
								fprintf( PG->out, ").forward(request, response);\n\n", (char *)pi3lval);
								break;
								};
							default:
								{
								fprintf( PG->out, "\nPARSE ERROR\n");
//								YYABORT;
								};
							};
						break;
					default:
						{
						fprintf( PG->out, "\nPARSE ERROR\n");
//						YYABORT;
						};
					};
				};
			}
		;
%%


int pi3servlet_head(void *YYPARSE_PARAM) {

	const char *pValString;

	assert(PG->out);
	if (!PG->out) YYABORT;

	fprintf( PG->out, "\
/**\n\
* Pi3Web Server 2.0\n\
* JSP version 1.0 bison parser $Revision: 1.16 $\n\
* $Date: 2001/09/09 12:20:00 $\n\
* Automatically generated code. Don't modify!\n\
* \n\
* Generated page header follows >>>>>\n\
**/\n"
	);

	fprintf( PG->out, "\
import java.io.*;\n\
import javax.servlet.*;\n\
import javax.servlet.http.*;\n\
import javax.servlet.jsp.*;\n\
import org.pi3.jsp.core.Pi3HttpJspPage;\n\
import org.pi3.jsp.runtime.*;\n\
\n\
// imports from page directives follow ...\n"
	);

	for (PG->import_ptr; PG->import_ptr;) fprintf( PG->out, "\
import %s;\n", PG->imports[--PG->import_ptr]
	);

fprintf( PG->out, "\n\
public class %s extends %s", PG->classname, PG->extends);

	pValString = (PG->isThreadSafe) ? "" : " implements SingleThreadModel";

	fprintf( PG->out, pValString);

	fprintf( PG->out, " {\n\
	public void _jspService(HttpServletRequest request, HttpServletResponse response)\n\
		throws IOException, ServletException {\n\
\n\
		JspFactory  factory     = JspFactory.getDefaultFactory();\n\
		PageContext pageContext = factory.getPageContext(\n\
					this,\n\
					request,\n\
					response,\n"
	);
	
	fprintf( PG->out, "\
					\"%s\", // errorPageURL\n",	PG->errorPage);

	pValString = (PG->session) ? "true" : "false";
	
	fprintf( PG->out, "\
					%s, // needsSession\n", pValString);

	switch (PG->buffer) {
		case -1:
			fprintf( PG->out, "\
					JspWriter.NO_BUFFER,\n"); break;
		case 8:
			fprintf( PG->out, "\
					JspWriter.DEFAULT_BUFFER,\n"); break;
		default:
			fprintf( PG->out, "\
					%d,\n", PG->buffer*1024);
		};
	
	pValString = (PG->autoFlush) ? "true" : "false";
	
	fprintf( PG->out, "\
					%s   // autoFlush\n\
					);\n\n", pValString);

	fprintf( PG->out, "\
		// initialize implicit variables for scripting env ...\n\
\n\
		HttpSession session = pageContext.getSession();\n\
		JspWriter   out     = pageContext.getOut();\n\
		Object      page    = this;\n\
		ServletContext application = pageContext.getServletContext();\n\
		ServletConfig config = pageContext.getServletConfig();\n\
\n\
		try {\n");
	fprintf( PG->out, "/**\n* >>>>> End of generated page header\n");
	fprintf( PG->out, "* Generated page contents follows >>>>>\n**/\n");
	YYACCEPT;
}

int pi3servlet_foot(void *YYPARSE_PARAM) {

	assert(PG->out);
	if (!PG->out) YYABORT;

	fprintf( PG->out, "/**\n* >>>>> End of generated page contents\n");
	fprintf( PG->out, "* Generated page footer follows >>>>>\n**/\n");
	fprintf( PG->out, "\
			out.flush();\n\
		} catch (Exception e) {\n\
			out.clear();\n\
			pageContext.handlePageException(e);\n\
		} finally {\n\
			out.close();\n\
			factory.releasePageContext(pageContext);\n\
		}\n\
	}\n\
}\n"
	);
	YYACCEPT;
}

int pi3write_bean(void *YYPARSE_PARAM) {
	
	if (PG->bs.bean.className[0])
		{
		fprintf( PG->out, "\t\t\t%s %s;\n", PG->bs.bean.className, PG->bs.bean.id );
		}
	else
		{
		fprintf( PG->out, "\t\t\t%s %s;\n", PG->bs.bean.type, PG->bs.bean.id );
		}

	if (!PG->bs.bean.scope[0] || (!strcmp(PG->bs.bean.scope, "page")))		
		{
		fprintf( PG->out, "\t\t\t%s = (%s)pageContext.getAttribute(\"%s\");\n",
			PG->bs.bean.id, PG->bs.bean.type, PG->bs.bean.id );
		}
	else
		{
		fprintf( PG->out, "\t\t\t%s = (%s)pageContext.getAttribute(\"%s\", pageContext.%s);\n",
			PG->bs.bean.id, PG->bs.bean.type, PG->bs.bean.id, PG->bs.bean.scope );
		}

	fprintf( PG->out, "\t\t\tif (%s == null)\n", PG->bs.bean.id );
	fprintf( PG->out, "\t\t\t\t{\n");

	if (!PG->bs.bean.className[0] && !PG->bs.bean.beanName[0])
		{
		fprintf( PG->out, "\t\t\t\tClassCastException e = new ClassCastException(\
\"No class to instantiate bean with id %s\");\n", PG->bs.bean.id );
		fprintf( PG->out, "\t\t\t\tthrow e;\n");
		}

	if (PG->bs.bean.className[0])
		{
		// T.B.D.: Check class and return InstantiationException, if
		// class is abstract - class.forName().newInstance() ?
		// class is interface - class.isInterface()
		// class hasn't non-args constructor - class.getConstructor()
		fprintf( PG->out, "\t\t\t\t%s = new %s();\n", PG->bs.bean.id, PG->bs.bean.className );
		}

	if (PG->bs.bean.beanName[0])
		{
		fprintf( PG->out, "\t\t\t\t%s = java.beans.Beans.instantiate(this.getClass().getClassLoader(), %s);\n",
			PG->bs.bean.id, PG->bs.bean.beanName );
		}
	fprintf( PG->out, "\t\t\t\tif (%s != null)\n", PG->bs.bean.id );
	fprintf( PG->out, "\t\t\t\t\t{\n");
	fprintf( PG->out, "\t\t\t\t\tpageContext.setAttribute(\"%s\", %s, pageContext.%s);\n",
			PG->bs.bean.id, PG->bs.bean.id, PG->bs.bean.scope );
	fprintf( PG->out, "\t\t\t\t\t}\n");
	fprintf( PG->out, "\t\t\t\t}\n\n");

	YYACCEPT;
}


int pi3write_setproperty(void *YYPARSE_PARAM) {

	if (!strcmp(PG->bs.prop.property, "*"))
		{
		fprintf( PG->out, "\t\t\tBeanProperty.set(%s,\"%s\",request);\n",
			PG->bs.prop.name, PG->bs.prop.property );
		}
	else
		{
		if ( !PG->bs.prop.value[0] )
			{
			if ( PG->bs.prop.param[0] )
				{
				fprintf( PG->out, "\t\t\tBeanProperty.set(%s,\"%s\",request.getParameter(\"%s\"));\n",
					PG->bs.prop.name, PG->bs.prop.property, PG->bs.prop.param );
				}
			else
				{
				fprintf( PG->out, "\t\t\tBeanProperty.set(%s,\"%s\",request.getParameter(\"%s\"));\n",
					PG->bs.prop.name, PG->bs.prop.property, PG->bs.prop.property );
				};
			}
		else
			{
			fprintf( PG->out, "\t\t\tBeanProperty.set(%s,\"%s\",\"%s\");\n",
				PG->bs.prop.name, PG->bs.prop.property, PG->bs.prop.value );
			}												
		};
	YYACCEPT;
}


int pi3write_plain(void *YYPARSE_PARAM, const char *s) {

	unsigned int i;

	assert(PG->out&&s);
	if (!PG->out || !s) YYABORT;
	fputs( "\t\t\tout.print(\"", PG->out );
	for ( i=0; i<strlen(s); i++ )
		{
		if ( s[i] == 0x22 )
			{
			if ( i==0 || ((i>0) && (s[i-1] != 0x5c)))
				{
				fputs( "\\\"", PG->out );
				}
			else
				{
				fputc( s[i], PG->out );
				}
			}
		else
			{
			fputc( s[i], PG->out );
			}
		}
	fputs( "\");\n", PG->out );
	YYACCEPT;
}