/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2001 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    and Holger Zimmermann (http://pi3web.sourceforge.net/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Servlet/JspParser/jsp_scanner.l,v $
 * $Date: 2001/09/09 12:01:49 $
 *
 Description:
\*____________________________________________________________________________*/

%{
#include <fstream.h>
#include <string.h>
#include <stdio.h>
#include "jsp_parser.h"
#include "jsp_scanner.h"

%}

%x ST_IN_TAG
%x ST_IN_SCRP
%x ST_IN_COMM
%x ST_IN_DECL
%x ST_IN_EXPR
%x ST_IN_DIRE
%x ST_INCLUDE
%x ST_IN_JSP
%x ST_IN_ACTION
%x ST_NESTED_EXPR
%x ST_QUOTED
%x ST_QUOTED_2
%x ST_QUOTED_3

%option c++
%option stack
%option yylineno
%option noyywrap
/* %option yyclass="JspFlexLexer" */

NO_WS_AND_TAG		[^ \n\r\t%]*
NO_END_TAG			[^ \n\r\t>]*
NO_END_COMT			[^ \n\r\t\-%>]*
WHITESPACE			[ \n\r\t]+
QUOTED_TEXT			[^\"0-9]*
QUOTED_NUMBER		[0-9]*
QUOTED_FNAME		[^\"]*

%%
<INITIAL>[^<\n]*					*pi3lval = (int)pi3text; return PLAIN;
<INITIAL>"<"[^<%j\/\n]*				*pi3lval = (int)pi3text; return PLAIN;
<INITIAL>"</"[^j\n]*				*pi3lval = (int)pi3text; return PLAIN;
<INITIAL>\n							*pi3lval = (int)pi3text; return NL;
<INITIAL>"<%"						*pi3lval = (int)pi3text; BEGIN(ST_IN_TAG); yyless(0); return START_TAG;
<INITIAL>"<jsp:"					*pi3lval = (int)pi3text; BEGIN(ST_IN_JSP); yyless(0); return START_JSP;

<INITIAL>"</jsp:useBean>"			*pi3lval = (int)pi3text; return END_JSP;
<INITIAL>"</jsp:plugin>"			*pi3lval = (int)pi3text; return END_JSP;
<INITIAL>"</jsp:params>"			*pi3lval = (int)pi3text; return END_JSP;
<INITIAL>"</jsp:fallback>"			*pi3lval = (int)pi3text; return END_JSP;

<ST_IN_TAG>{WHITESPACE}"%>"			*pi3lval = (int)pi3text; BEGIN(INITIAL); return END_TAG;
<ST_IN_TAG>{WHITESPACE}"--%>"		*pi3lval = (int)pi3text; BEGIN(INITIAL); return END_TAG;

<ST_IN_JSP>"/>"						*pi3lval = (int)pi3text; BEGIN(INITIAL); return END_JSP;
<ST_IN_JSP>">"						*pi3lval = (int)pi3text; BEGIN(INITIAL); // xml end tag follows later

<ST_IN_TAG>"<%"{WHITESPACE}			*pi3lval = (int)pi3text; BEGIN(ST_IN_SCRP); return SCRIPTLET_START;
<ST_IN_SCRP>{WHITESPACE}			*pi3lval = (int)pi3text; return WHITESPACE;
<ST_IN_SCRP>{NO_WS_AND_TAG}			*pi3lval = (int)pi3text; return SCRIPTLET;
<ST_IN_SCRP>"%"{NO_END_TAG}			*pi3lval = (int)pi3text; return SCRIPTLET;
<ST_IN_SCRP>{WHITESPACE}"%>"		*pi3lval = (int)pi3text; BEGIN(ST_IN_TAG); PI3LESS(pi3text,3); return SCRIPTLET_END;

<ST_IN_TAG>"<%@"{WHITESPACE}		*pi3lval = (int)pi3text; BEGIN(ST_IN_DIRE); return DIRECTIVE_START;
<ST_IN_DIRE>"page"					*pi3lval = (int)pi3text; return DIR_PAGE;
<ST_IN_DIRE>"include"				*pi3lval = (int)pi3text; return DIR_INCLUDE;
<ST_IN_DIRE>"taglib"				*pi3lval = (int)pi3text; return DIR_TAGLIB;

<ST_IN_DIRE>{WHITESPACE}			*pi3lval = (int)pi3text; return WHITESPACE;
<ST_IN_DIRE>"language"				*pi3lval = (int)pi3text; return DIR_LANGUAGE;
<ST_IN_DIRE>"extends"				*pi3lval = (int)pi3text; return DIR_EXTENDS;
<ST_IN_DIRE>"import"				*pi3lval = (int)pi3text; return DIR_IMPORT;
<ST_IN_DIRE>"session"				*pi3lval = (int)pi3text; return DIR_SESSION;
<ST_IN_DIRE>"buffer"				*pi3lval = (int)pi3text; return DIR_BUFFER;
<ST_IN_DIRE>"autoFlush"				*pi3lval = (int)pi3text; return DIR_AUTOFLUSH;
<ST_IN_DIRE>"isThreadSafe"			*pi3lval = (int)pi3text; return DIR_ISTHREADSAFE;
<ST_IN_DIRE>"info"					*pi3lval = (int)pi3text; return DIR_INFO;
<ST_IN_DIRE>"errorPage"				*pi3lval = (int)pi3text; return DIR_ERRORPAGE;
<ST_IN_DIRE>"contentType"			*pi3lval = (int)pi3text; return DIR_CONTENTTYPE;
<ST_IN_DIRE>"isErrorPage"			*pi3lval = (int)pi3text; return DIR_ISERRORPAGE;
<ST_IN_DIRE>"uri"					*pi3lval = (int)pi3text; return DIR_URI;
<ST_IN_DIRE>"prefix"				*pi3lval = (int)pi3text; return DIR_PREFIX;
<ST_IN_DIRE>"file"					*pi3lval = (int)pi3text; BEGIN(ST_INCLUDE); return DIR_INCFILE;
<ST_INCLUDE>{WHITESPACE}			*pi3lval = (int)pi3text; return WHITESPACE;
<ST_INCLUDE>"="						return pi3text[0];
<ST_INCLUDE>\"						return pi3text[0];
<ST_INCLUDE>{QUOTED_FNAME} {

    if ( include_stack_ptr >= MAX_INCLUDE_DEPTH )
		{
		*pi3lval = (int)pi3text;
		BEGIN(ST_QUOTED);
		return ERROR_INCLUDES_TOO_DEEPLY;
		}

	ifstream *pi3in = new ifstream(pi3text, ios::nocreate);
    if ( !pi3in || !pi3in->is_open() )
		{
		*pi3lval = (int)pi3text;
		BEGIN(ST_QUOTED);
	    return ERROR_CANT_OPEN_INCLUDE;
		};

	strcpy((fname_buf), pi3text);
	include_stack[include_stack_ptr++] = YY_CURRENT_BUFFER;
	yy_switch_to_buffer( yy_create_buffer( pi3in, YY_BUF_SIZE ) );
    BEGIN(INITIAL);
	return PUSH_INCLUDE;
    }

<ST_IN_DIRE>"="						return pi3text[0];
<ST_IN_DIRE>\"						BEGIN(ST_QUOTED); return pi3text[0];
<ST_QUOTED>\"						BEGIN(ST_IN_DIRE); return pi3text[0];
<ST_QUOTED>"true"					*pi3lval = 1; return BOOL; 
<ST_QUOTED>"false"					*pi3lval = 0; return BOOL;
<ST_QUOTED>{QUOTED_FNAME}			*pi3lval = (int)pi3text; return STRING;

<ST_IN_DIRE>{WHITESPACE}"%>"		*pi3lval = (int)pi3text; BEGIN(ST_IN_TAG); PI3LESS(pi3text,3); return DIRECTIVE_END;

<ST_IN_TAG>"<%!"{WHITESPACE}		*pi3lval = (int)pi3text; BEGIN(ST_IN_DECL); return DECLARATION_START;
<ST_IN_DECL>{WHITESPACE}			*pi3lval = (int)pi3text; return WHITESPACE;
<ST_IN_DECL>{NO_WS_AND_TAG}			*pi3lval = (int)pi3text; return DECLARATION;
<ST_IN_DECL>"%"{NO_END_TAG}			*pi3lval = (int)pi3text; return DECLARATION;
<ST_IN_DECL>{WHITESPACE}%>			*pi3lval = (int)pi3text; BEGIN(ST_IN_TAG); PI3LESS(pi3text,3); return DECLARATION_END;

<ST_IN_TAG>"<%="{WHITESPACE}		*pi3lval = (int)pi3text; BEGIN(ST_IN_EXPR); return EXPRESSION_START;
<ST_IN_EXPR>{WHITESPACE}			*pi3lval = (int)pi3text; return WHITESPACE;
<ST_IN_EXPR>{NO_WS_AND_TAG}			*pi3lval = (int)pi3text; return EXPRESSION;
<ST_IN_EXPR>"%"{NO_END_TAG}			*pi3lval = (int)pi3text; return EXPRESSION;
<ST_IN_EXPR>{WHITESPACE}"%>"		*pi3lval = (int)pi3text; BEGIN(ST_IN_TAG); PI3LESS(pi3text,3); return EXPRESSION_END;

<ST_IN_TAG>"<%--"{WHITESPACE}		*pi3lval = (int)pi3text; BEGIN(ST_IN_COMM); return COMMENT_START;
<ST_IN_COMM>{WHITESPACE}			*pi3lval = (int)pi3text; return WHITESPACE;
<ST_IN_COMM>{NO_END_COMT}			*pi3lval = (int)pi3text; return COMMENT;
<ST_IN_COMM>"-"{NO_END_COMT}		*pi3lval = (int)pi3text; return COMMENT;
<ST_IN_COMM>"--%"{NO_END_COMT}		*pi3lval = (int)pi3text; return COMMENT;
<ST_IN_COMM>{WHITESPACE}"--%>"		*pi3lval = (int)pi3text; BEGIN(ST_IN_TAG); PI3LESS(pi3text,5); return COMMENT_END;

<ST_IN_JSP>"<jsp:"					*pi3lval = (int)pi3text; BEGIN(ST_IN_ACTION); return ACTION_START;
<ST_IN_ACTION>"useBean"				*pi3lval = (int)pi3text; return JSP_USEBEAN;
<ST_IN_ACTION>"setProperty"			*pi3lval = (int)pi3text; return JSP_SETPROPERTY;
<ST_IN_ACTION>"getProperty"			*pi3lval = (int)pi3text; return JSP_GETPROPERTY;
<ST_IN_ACTION>"include"				*pi3lval = (int)pi3text; return JSP_INCLUDE;
<ST_IN_ACTION>"forward"				*pi3lval = (int)pi3text; return JSP_FORWARD;
<ST_IN_ACTION>"plugin"				*pi3lval = (int)pi3text; return JSP_PLUGIN;
<ST_IN_ACTION>"params"				*pi3lval = (int)pi3text; return JSP_PARAMS;
<ST_IN_ACTION>"fallback"			*pi3lval = (int)pi3text; return JSP_FALLBACK;

<ST_IN_ACTION>{WHITESPACE}			*pi3lval = (int)pi3text; return WHITESPACE;
<ST_IN_ACTION>"id"					*pi3lval = (int)pi3text; return JSP_ID;
<ST_IN_ACTION>"scope"				*pi3lval = (int)pi3text; return JSP_SCOPE;
<ST_IN_ACTION>"class"				*pi3lval = (int)pi3text; return JSP_CLASS;
<ST_IN_ACTION>"type"				*pi3lval = (int)pi3text; return JSP_TYPE;
<ST_IN_ACTION>"beanName"			*pi3lval = (int)pi3text; return JSP_BEANNAME;
<ST_IN_ACTION>"name"				*pi3lval = (int)pi3text; return JSP_NAME;
<ST_IN_ACTION>"property"			*pi3lval = (int)pi3text; return JSP_PROPERTY;
<ST_IN_ACTION>"param"				*pi3lval = (int)pi3text; return JSP_PARAM;
<ST_IN_ACTION>"value"				*pi3lval = (int)pi3text; return JSP_VALUE;
<ST_IN_ACTION>"page"				*pi3lval = (int)pi3text; return JSP_PAGE;
<ST_IN_ACTION>"flush"				*pi3lval = (int)pi3text; return JSP_FLUSH;
<ST_IN_ACTION>"code"				*pi3lval = (int)pi3text; return JSP_CODE;
<ST_IN_ACTION>"codebase"			*pi3lval = (int)pi3text; return JSP_CODEBASE;
<ST_IN_ACTION>"align"				*pi3lval = (int)pi3text; return JSP_ALIGN;
<ST_IN_ACTION>"archive"				*pi3lval = (int)pi3text; return JSP_ARCHIVE;
<ST_IN_ACTION>"height"				*pi3lval = (int)pi3text; return JSP_HEIGHT;
<ST_IN_ACTION>"hspace"				*pi3lval = (int)pi3text; return JSP_HSPACE;
<ST_IN_ACTION>"jreversion"			*pi3lval = (int)pi3text; return JSP_JREVERSION;
<ST_IN_ACTION>"vspace"				*pi3lval = (int)pi3text; return JSP_VSPACE;
<ST_IN_ACTION>"width"				*pi3lval = (int)pi3text; return JSP_WIDTH;
<ST_IN_ACTION>"nspluginurl"			*pi3lval = (int)pi3text; return JSP_NSPLUGINURL;
<ST_IN_ACTION>"iepluginurl"			*pi3lval = (int)pi3text; return JSP_IEPLUGINURL;
<ST_IN_ACTION>"="					return pi3text[0];
<ST_IN_ACTION>\"					BEGIN(ST_QUOTED_2); return pi3text[0];
<ST_QUOTED_2>\"						BEGIN(ST_IN_ACTION); return pi3text[0];
<ST_IN_ACTION>'						BEGIN(ST_QUOTED_3); return pi3text[0];
<ST_QUOTED_3>'						BEGIN(ST_IN_ACTION); return pi3text[0];
<ST_QUOTED_2>{QUOTED_FNAME}			*pi3lval = (int)pi3text; return STRING;

<ST_IN_ACTION>"/>"					*pi3lval = (int)pi3text; BEGIN(ST_IN_JSP); PI3LESS(pi3text,2); return ACTION_END;
<ST_IN_ACTION>">"					*pi3lval = (int)pi3text; BEGIN(ST_IN_JSP); PI3LESS(pi3text,1); return ACTION_END;

<ST_QUOTED_3>"<%="{WHITESPACE}		*pi3lval = (int)pi3text; BEGIN(ST_NESTED_EXPR); return EXPRESSION_START;
<ST_NESTED_EXPR>{WHITESPACE}		*pi3lval = (int)pi3text; return WHITESPACE;
<ST_NESTED_EXPR>{NO_WS_AND_TAG}		*pi3lval = (int)pi3text; return JSP_EXPRESSION;
<ST_NESTED_EXPR>"%"{NO_END_TAG}		*pi3lval = (int)pi3text; return JSP_EXPRESSION;

<ST_NESTED_EXPR>{WHITESPACE}"%>"	*pi3lval = (int)pi3text; BEGIN(ST_QUOTED_3); return EXPRESSION_END;

<ST_IN_TAG,ST_IN_JSP,ST_IN_ACTION,ST_IN_SCRP,ST_IN_DIRE,ST_IN_EXPR,ST_IN_DECL,ST_IN_COMM,ST_INCLUDE,ST_QUOTED><<EOF>>	 BEGIN(INITIAL);

<INITIAL><<EOF>> {
	if ( --include_stack_ptr < 0 )
		{
		yyterminate();
		}
	else
		{
		yy_delete_buffer( YY_CURRENT_BUFFER );
		yy_switch_to_buffer( include_stack[include_stack_ptr] );
		BEGIN(ST_QUOTED);
		return POP_INCLUDE;
		}
	}
%%


/* test main */
int main(int argc, char *argv[ ]) {

	if (argc < 2) return 1;
//	ios::sync_with_stdio();

	JspFlexLexer scanner(argv[1], argv[2], "JSPtest", "Pi3HttpJspPage");
    scanner.exec();

	return 0;
}


/**
* Scanner error callback
**/
int pi3error(char *msg) {
//	cerr << msg << endl;
	return 1;
}


/**
* Wrapper functions for indirect call
* of class members by C-style parser
**/
int pi3lex(int *pi3lval, void *YYLEX_PARAM) {
	return SG->yylex(pi3lval);
}

const char *get_incl_name(void *YYLEX_PARAM) {
	return SG->get_incl_name();
}

int get_incl_level(void *YYLEX_PARAM) {
	return SG->get_incl_level();
}

int get_lineno(void *YYLEX_PARAM) {
	return SG->lineno();
}


/**
* Jsp Scanner class member functions
**/
JspFlexLexer::JspFlexLexer( const char *ifname, const char *ofname, const char *c, const char *e ) :
	inFileName( ifname ),
	i( 0 ),
	pp( new p_globals(c, e, ofname, this) ),
	include_stack_ptr( 0 )

{
	istream *arg_yyin = (ifname) ? (istream *)new ifstream(ifname) : &cin;
	yyFlexLexer(arg_yyin, 0);
	switch_streams(arg_yyin, 0);

	yylineno = 1;
}

JspFlexLexer::~JspFlexLexer() {
	((ifstream *)yyin)->close();
    delete( pp );
}

int JspFlexLexer::prepare_next_pass() {
	include_stack_ptr = 0;
	yy_push_state(INITIAL);
	((ifstream *)yyin)->close();
	((ifstream *)yyin)->open(inFileName);
	switch_streams(yyin, 0);
	yylineno = 1;
	return ++(pp->pass);
}

int JspFlexLexer::exec()
{
	/* execute pass 1 and print header*/
	if ( !pi3parse((pPG)pp) && !pi3servlet_head((pPG)pp) )
		{

		/* prepare and execute pass 2 */
		prepare_next_pass();
		int iRet = pi3parse((pPG)pp);

		/* print footer*/
		iRet = pi3servlet_foot((pPG)pp);

		return iRet;
		}
	else
		{
		return -1;
		}
}

/**
* Never call the yylex() inherited from the base class
**/
int yyFlexLexer::yylex()
{
	pi3error("Error: yyFlexLexer::yylex() called.");
	return -1;
}