/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2001 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in
	the documentation and/or other materials provided with the
	distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
	mentioning features must contain the following
	acknowledgment:
	"This product includes software developed by John Roy
	and Holger Zimmermann (http://pi3web.sourceforge.net/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Servlet/Pi3Reader.cpp,v $
 * $Date: 2001/06/03 19:11:24 $
 *
 Description:
	Execute an servlet extension.

\*____________________________________________________________________________*/
//$SourceTop:$

#include <jni.h>                            /* where everything is defined */
#include "org_pi3_servlet_core_Pi3Reader.h" /* automatically created JNI header */
#include "PiAPI.h"
#include "Pi3API.h"
#include "ServletHandler.h"
#include <wchar.h>

/*
 * Class:     org_pi3_servlet_core_Pi3Reader
 * Method:    internalRead
 * Signature: ([CII)I
 */
JNIEXPORT jint JNICALL Java_org_pi3_servlet_core_Pi3Reader_internalRead
  (JNIEnv *env, jobject rdr, jcharArray arr, jint off, jint len) {

	jclass cls = env->GetObjectClass(rdr);
	jmethodID getPiHttp;
	getPiHttp = env->GetMethodID( cls, "getPiHttp", "()I");
	// Errorhandling!
	if (!getPiHttp) { return -1; };
	env->ExceptionClear();

	int PiHttp = env->CallIntMethod(rdr, getPiHttp);
	if ( env->ExceptionOccurred() )
		{
		env->ExceptionDescribe();
		env->ExceptionClear();
		// Errorhandling!
		return -1;
		};

	/* ---
	read len bytes from PIIOBuffer
	--- */
	PIHTTP *pPIHTTP = (PIHTTP *)PiHttp;
	char *szBuf;
	int iRes = PIIOBuffer_pollBeforeRead( pPIHTTP->pBuffer );
	if (iRes <= 0) { goto read_error; };

	szBuf = (char *)PIUtil_malloc(len);
	iRes = PIIOBuffer_readToBuffer( pPIHTTP->pBuffer, szBuf, len );

	/* ---
	copy char buffer to java array
	--- */
	if (iRes > 0) {
		
		wchar_t wc;
		for (int i=0; i < iRes; i++) {
			if (mbtowc(&wc, szBuf + i, 1) > 0) {
				// off + len <= array size is already checked on the java side!
				env->SetCharArrayRegion(arr, off + i, 1, (jchar *)&wc); 
				if ( env->ExceptionOccurred() ) { break; };
			};
		};

	PIUtil_free(szBuf);

	if ( env->ExceptionOccurred() )
		{
		env->ExceptionDescribe();
		env->ExceptionClear();
		// Errorhandling!
		return -1;
		};
	};

read_error:
	switch (iRes) {
		case  0: return -1; // stream closed
		case -1: {
			cls = env->FindClass("java/io/IOException");
			if (cls) { env->ThrowNew(cls, "Error in PIIOBuffer_readToBuffer."); };
			// fall through default (return piRead)
		};
        default:  return iRes;
	};
};