/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2001 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    and Holger Zimmermann (http://pi3web.sourceforge.net/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/TestClnt/StressClient.h,v $
 * $Date: 2001/08/04 13:55:07 $
 *
 Description:
\*____________________________________________________________________________*/
//$SourceTop:$

#ifndef STRESSCLIENT_H_
#define STRESSCLIENT_H_

#include <assert.h>

#include "Pi2API.h"
#include "PIString.h"
#include "PIIOBuf.h"


/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
#define CONFIG_ERR(this, msg)   \
	{ PILog_addMessage( PIObject_getDB(this), \
	PIObject_getConfigurationDB(this), PILOG_ERROR, "%s", (msg) ); }

#define KEY_CONF_USERAGENT			"UserAgent"
#define KEY_CONF_PROTOCOL			"Protocol"
#define KEY_CONF_VIRTUALHOST		"VirtualHost"
#define KEY_CONF_KEEPOPEN			"KeepOpen"
#define KEY_CONF_KEEPOPENCOUNT		"KeepOpenCount"
#define KEY_CONF_KEEPOPENTIMEOUT	"KeepOpenTimeout"
#define KEY_CONF_RANDOM				"Random"
#define KEY_CONF_FILEMASK			"FileMask"
#define KEY_CONF_TESTFILE			"TestFile"
#define KEY_CONF_TESTCASE			"TestCase"
#define KEY_CONF_LOGFILE			"Logfile"
#define KEY_CONF_OPENMODE			"OpenMode"

#define KEY_INT_TESTFILE			"Testfile"
#define KEY_INT_STATUS				"Status"

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class StressClient{

private:
	/* ---
	Configuration data
	--- */
	PIObject *pObject;				/* the object */
	int iOK;						/* status */
	unsigned int iKeepOpenCount;
	int iProtocol;
	int iKeepAlive;	
	int iAliveTimeout;
	const char *pFKRecvTimeout;
	const char *pFKTestfile;
	const char *pFKHost;
	unsigned int iRandom;
	unsigned int iCurrentTest;
	int iMaxTest;
	PIString sUserAgent;
	PIString sVirtualHost;
	PIPLATFORM_FD logfile;
	PIString sLogfile;
	PIString sOpenMode;
	PIDB *pTestcaseDB;

	int ReadRequestLine( PIDB *pDB, const char *pLine );
	int ReadFileMask( const char *pValue, int iStatus );
	int SendRequestDB(PIIOBuffer *pBuf, PIDB *pC, PIDB *pQ, int iKeepOpen );
	int SendRequestFile(PIIOBuffer *pBuf, const char *pPath );
protected:
	StressClient( const StressClient & ) { assert( 0 ); };

	/* ---
	Derived classes override this to handle parameters
	--- */
	int Parameter( const char *pVariable, const char *pValue,
		const char *pWhere );
	inline void SetOK( int iNewValue ) { iOK = iNewValue; };
	static int ParameterFn( void *pData, const char *pVar, const char *pVal,
		const char *pWhere )
		{
		assert( pData );
		return ((StressClient *)pData)->Parameter( pVar, pVal,
			pWhere ? pWhere : "" );
		};

public:
	StressClient( PIObject *pObject, int iArgc, const char *ppArgv[] );
	virtual ~StressClient();
	int Execute( int iArgc, const char *ppArgv[] );
	inline int IsOK()				{ return iOK; };
	inline PIObject *Object()		{ return pObject; };
};


/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Function prototypes
\*____________________________________________________________________________*/
PUBLIC_PIAPI int StressClient_execute( PIObject *pObj, int iArgc, const char *ppArgv[] );
PUBLIC_PIAPI int StressClient_onClassLoad( PIClass_LoadAction eLoad, void * );
PUBLIC_PIAPI int StressClient_constructor( PIObject *pObj, int, const char *[] );
PUBLIC_PIAPI int StressClient_copyConstructor( PIObject *pObj, int, const char *[] );
PUBLIC_PIAPI int StressClient_destructor( PIObject *pObj, int, const char *[] );

#endif /* STRESSCLIENT_H_ */

